/*
 * ConversionHandler.cpp
 *
 *  Created on: 23. 2. 2014
 *	  Author: Tomas Pecka
 */

#include <measure>
#include "ConversionHandler.h"

#include "automaton/convert/ToRegExpStateElimination.h"
#include "automaton/convert/ToRegExpAlgebraic.h"
#include "automaton/convert/ToGrammarLeftRG.h"
#include "automaton/convert/ToGrammarRightRG.h"

#include "regexp/convert/ToAutomatonGlushkov.h"
#include "regexp/convert/ToAutomatonThompson.h"
#include "regexp/convert/ToAutomatonDerivation.h"

#include "grammar/convert/ToAutomaton.h"
#include "grammar/convert/ToAutomatonBottomUp.h"
#include "grammar/convert/ToRegExpAlgebraic.h"

#include "regexp/convert/ToGrammarRightRGGlushkov.h"
#include "regexp/convert/ToGrammarRightRGDerivation.h"

#include "rte/convert/ToPostfixPushdownAutomatonGlushkov.h"

#include <factory/XmlDataFactory.hpp>
#include <exception/CommonException.h>

ConversionHandler::ConversionHandler( std::deque<sax::Token> && tokens, const std::string & target, const std::string & algorithm ) :
	m_tokens( std::move( tokens ) ),
	m_source( parseFormalismFromTokens( ) ),
	m_target( parseFormalismFromString( target ) ),
	m_algorithm( parseAlgorithmFromString( algorithm ) )
{
}

void ConversionHandler::convert ( void ) {
	if ( m_source == FINITE_AUTOMATON )
		convertFA ( );
	else if ( m_source == REGULAR_EXPRESSION )
		convertRE ( );
	else if ( m_source == REGULAR_GRAMMAR )
		convertRG ( );
	else if ( m_source == PUSHDOWN_AUTOMATON )
		convertPDA ( );
	else if ( m_source == CONTEXTFREE_GRAMMAR )
		convertCFG ( );
	else if ( m_source == REGULAR_TREE_EXPRESSION )
		convertRTE ( );
	else
		throw exception::CommonException( "ConversionHandler: Unknown source formalism." );
}

// ----------------------------------------------------------------------------

void ConversionHandler::convertFA ( void ) {
	if ( m_target == REGULAR_EXPRESSION )
		convertFAtoRE ( );
	else if ( m_target == REGULAR_GRAMMAR )
		convertFAtoRG ( );
	else
		throw exception::CommonException( "ConversionHandler: Unknown target formalism." );
}

void ConversionHandler::convertRE ( void ) {
	if ( m_target == FINITE_AUTOMATON )
		convertREtoFA ( );
	else if ( m_target == REGULAR_GRAMMAR )
		convertREtoRG ( );
	else
		throw exception::CommonException( "ConversionHandler: Invalid target formalism." );
}

void ConversionHandler::convertRG ( void ) {
	if ( m_target == FINITE_AUTOMATON )
		convertRGtoFA ( );
	else if ( m_target == REGULAR_EXPRESSION )
		convertRGtoRE ( );
	else
		throw exception::CommonException( "ConversionHandler: Invalid target formalism." );
}

void ConversionHandler::convertPDA ( void ) {
	throw exception::CommonException ( "ConversionHandler: Invalid source formalism." );
}

void ConversionHandler::convertCFG ( void ) {
	if ( m_target == PUSHDOWN_AUTOMATON )
		convertCFGtoPDA ( );
	else
		throw exception::CommonException ( "ConversionHandler: Invalid target formalism." );
}

void ConversionHandler::convertRTE ( void ) {
	if ( m_target == PUSHDOWN_AUTOMATON )
		convertRTEtoPDA ( );
	else
		throw exception::CommonException( "ConversionHandler: Invalid target formalism." );
}

// ----------------------------------------------------------------------------

void ConversionHandler::convertFAtoRE ( void ) {
	const automaton::Automaton automaton = alib::XmlDataFactory::fromTokens < automaton::Automaton > ( std::move ( m_tokens ) );

	measurements::end ( );
	measurements::start ( "Algorithm", measurements::Type::MAIN );

	switch ( m_algorithm ) {
	case BRZOZOWSKI_ALGEBRAIC: {
		   regexp::RegExp regexp = automaton::convert::ToRegExpAlgebraic::convert ( automaton );

		   measurements::end ( );
		   measurements::start ( "Output write", measurements::Type::AUXILIARY );

		   alib::XmlDataFactory::toStdout ( regexp );
		   break;
	   }

	case STATE_ELIMINATION:
	case DEFAULT: {
		   regexp::RegExp regexp = automaton::convert::ToRegExpStateElimination::convert ( automaton );

		   measurements::end ( );
		   measurements::start ( "Output write", measurements::Type::AUXILIARY );

		   alib::XmlDataFactory::toStdout ( regexp );
		   break;
	   }

	default:
		throw exception::CommonException ( "ConversionHandler: Invalid algorithm." );
	}
}

void ConversionHandler::convertFAtoRG ( void ) {
	const automaton::Automaton fsm = alib::XmlDataFactory::fromTokens < automaton::Automaton > ( std::move ( m_tokens ) );

	measurements::end ( );
	measurements::start ( "Algorithm", measurements::Type::MAIN );

	switch ( m_algorithm ) {
	case INCOMING_TRANSITIONS: {
		   grammar::LeftRG grammar = automaton::convert::ToGrammarLeftRG::convert ( fsm );

		   measurements::end ( );
		   measurements::start ( "Output write", measurements::Type::AUXILIARY );

		   alib::XmlDataFactory::toStdout ( grammar );
		   break;
	   }

	case OUTGOING_TRANSITIONS:
	case DEFAULT: {
		   grammar::RightRG grammar = automaton::convert::ToGrammarRightRG::convert ( fsm );

		   measurements::end ( );
		   measurements::start ( "Output write", measurements::Type::AUXILIARY );

		   alib::XmlDataFactory::toStdout ( grammar );
		   break;
	   }

	default:
		throw exception::CommonException( "ConversionHandler: Invalid algorithm." );
	}
}

// ----------------------------------------------------------------------------

void ConversionHandler::convertREtoFA ( void ) {
	const regexp::RegExp regexp = alib::XmlDataFactory::fromTokens < regexp::RegExp > ( std::move ( m_tokens ) );

	measurements::end ( );
	measurements::start ( "Algorithm", measurements::Type::MAIN );

	switch ( m_algorithm ) {
	case BRZOZOWSKI_DERIVATION: {
		   automaton::DFA automaton = regexp::convert::ToAutomatonDerivation::convert ( regexp );

		   measurements::end ( );
		   measurements::start ( "Output write", measurements::Type::AUXILIARY );

		   alib::XmlDataFactory::toStdout ( automaton );
		   break;
	   }

	case THOMPSON_NFA: {
		   automaton::EpsilonNFA automaton = regexp::convert::ToAutomatonThompson::convert ( regexp );

		   measurements::end ( );
		   measurements::start ( "Output write", measurements::Type::AUXILIARY );

		   alib::XmlDataFactory::toStdout ( automaton );
		   break;
	   }

	case GLUSHKOV_NFA:
	case DEFAULT: {
		   automaton::NFA automaton = regexp::convert::ToAutomatonGlushkov::convert ( regexp );

		   measurements::end ( );
		   measurements::start ( "Output write", measurements::Type::AUXILIARY );

		   alib::XmlDataFactory::toStdout ( automaton );
		   break;
	   }

	default:
		throw exception::CommonException( "ConversionHandler: Invalid algorithm." );
	}
}

void ConversionHandler::convertREtoRG ( void ) {
	const regexp::RegExp regexp = alib::XmlDataFactory::fromTokens < regexp::RegExp > ( std::move ( m_tokens ) );

	measurements::end ( );
	measurements::start ( "Algorithm", measurements::Type::MAIN );

	switch ( m_algorithm ) {
	case BRZOZOWSKI_DERIVATION: {
		   grammar::RightRG grammar = regexp::convert::ToGrammarRightRGDerivation::convert ( regexp );

		   measurements::end ( );
		   measurements::start ( "Output write", measurements::Type::AUXILIARY );

		   alib::XmlDataFactory::toStdout ( grammar );
		   break;
	   }

	case GLUSHKOV_NFA:
	case DEFAULT: {
		   grammar::RightRG grammar = regexp::convert::ToGrammarRightRGGlushkov::convert ( regexp );

		   measurements::end ( );
		   measurements::start ( "Output write", measurements::Type::AUXILIARY );

		   alib::XmlDataFactory::toStdout ( grammar );
		   break;
	   }

	default:
		throw exception::CommonException( "ConversionHandler: Invalid algorithm." );
	}
}

// ----------------------------------------------------------------------------

void ConversionHandler::convertRGtoFA( void ) {
	const grammar::Grammar grammar = alib::XmlDataFactory::fromTokens<grammar::Grammar>(std::move(m_tokens));

	measurements::end ( );
	measurements::start ( "Algorithm", measurements::Type::MAIN );

	switch ( m_algorithm ) {
	case DEFAULT: {
		   automaton::Automaton automaton = grammar::convert::ToAutomaton::convert ( grammar );

		   measurements::end ( );
		   measurements::start ( "Output write", measurements::Type::AUXILIARY );

		   alib::XmlDataFactory::toStdout ( automaton );
		   break;
	   }

	default:
		throw exception::CommonException( "ConversionHandler: Invalid algorithm." );
	}
}

void ConversionHandler::convertRGtoRE ( void ) {
	const grammar::Grammar grammar = alib::XmlDataFactory::fromTokens < grammar::Grammar > ( std::move ( m_tokens ) );

	measurements::end ( );
	measurements::start ( "Algorithm", measurements::Type::MAIN );

	switch ( m_algorithm ) {
	case BRZOZOWSKI_ALGEBRAIC:
	case DEFAULT: {
		   regexp::RegExp regexp = grammar::convert::ToRegExpAlgebraic::convert ( grammar );

		   measurements::end ( );
		   measurements::start ( "Output write", measurements::Type::AUXILIARY );

		   alib::XmlDataFactory::toStdout ( regexp );
		   break;
	   }

	default:
		throw exception::CommonException( "ConversionHandler: Invalid algorithm." );
	}
}

void ConversionHandler::convertCFGtoPDA ( void ) {
	const grammar::Grammar grammar = alib::XmlDataFactory::fromTokens < grammar::Grammar > ( std::move ( m_tokens ) );

	measurements::end ( );
	measurements::start ( "Algorithm", measurements::Type::MAIN );

	switch ( m_algorithm ) {
	case TOP_DOWN: {
		   automaton::Automaton automaton = grammar::convert::ToAutomaton::convert ( grammar );

		   measurements::end ( );
		   measurements::start ( "Output write", measurements::Type::AUXILIARY );

		   alib::XmlDataFactory::toStdout ( automaton );
		   break;
	   }

	case BOTTOM_UP: {
		   automaton::Automaton automaton = grammar::convert::ToAutomatonBottomUp::convert ( grammar );

		   measurements::end ( );
		   measurements::start ( "Output write", measurements::Type::AUXILIARY );

		   alib::XmlDataFactory::toStdout ( automaton );
		   break;
	   }

	default:
		throw exception::CommonException ( "ConversionHandler: Invalid algorithm." );
	}
}

void ConversionHandler::convertRTEtoPDA ( void ) {
	const rte::RTE rte = alib::XmlDataFactory::fromTokens < rte::RTE > ( std::move ( m_tokens ) );

	measurements::end ( );
	measurements::start ( "Algorithm", measurements::Type::MAIN );

	switch ( m_algorithm ) {
	case GLUSHKOV_RTE: {
		   automaton::NPDA automaton = rte::convert::ToPostfixPushdownAutomatonGlushkov::convert ( rte );

		   measurements::end ( );
		   measurements::start ( "Output write", measurements::Type::AUXILIARY );

		   alib::XmlDataFactory::toStdout ( automaton );
		   break;
	   }

	default:
		throw exception::CommonException ( "ConversionHandler: Invalid algorithm." );
	}
}

// ----------------------------------------------------------------------------

ConversionHandler::TFormalism ConversionHandler::parseFormalismFromString ( const std::string & target ) const {
	if ( target == "fa" ) return FINITE_AUTOMATON;

	if ( target == "re" ) return REGULAR_EXPRESSION;

	if ( target == "rg" ) return REGULAR_GRAMMAR;

	if ( target == "pda" ) return PUSHDOWN_AUTOMATON;

	if ( target == "cfg" ) return CONTEXTFREE_GRAMMAR;

	if ( target == "rte" ) return REGULAR_TREE_EXPRESSION;

	throw exception::CommonException ( "ConversionHandler: Unknown target type." );
}

ConversionHandler::TFormalism ConversionHandler::parseFormalismFromTokens ( void ) const {
	if ( alib::XmlDataFactory::first < automaton::Automaton > ( m_tokens ) )
		return FINITE_AUTOMATON;

	if ( alib::XmlDataFactory::first < regexp::RegExp > ( m_tokens ) )
		return REGULAR_EXPRESSION;

	if ( alib::XmlDataFactory::first < grammar::CFG > ( m_tokens ) )
		return CONTEXTFREE_GRAMMAR;

	if ( alib::XmlDataFactory::first < grammar::Grammar > ( m_tokens ) )
		return REGULAR_GRAMMAR;

	if ( alib::XmlDataFactory::first < rte::RTE > ( m_tokens ) )
		return REGULAR_TREE_EXPRESSION;

	throw exception::CommonException ( "ConversionHandler: Invalid input formalism." );
}

ConversionHandler::TAlgorithm ConversionHandler::parseAlgorithmFromString ( const std::string & algorithm ) const {
	if ( algorithm == "elimination" ) return STATE_ELIMINATION;

	if ( algorithm == "algebraic" ) return BRZOZOWSKI_ALGEBRAIC;

	if ( algorithm == "brzozowski" ) return BRZOZOWSKI_DERIVATION;

	if ( algorithm == "glushkov" ) return GLUSHKOV_NFA;

	if ( algorithm == "thompson" ) return THOMPSON_NFA;

	if ( algorithm == "incoming" ) return INCOMING_TRANSITIONS;

	if ( algorithm == "outgoing" ) return OUTGOING_TRANSITIONS;

	if ( algorithm == "topdown" ) return TOP_DOWN;

	if ( algorithm == "bottomup" ) return BOTTOM_UP;

	if ( algorithm == "glushkovrte" ) return GLUSHKOV_RTE;

	if ( ( algorithm == "" ) || ( algorithm == "default" ) ) return DEFAULT;

	throw exception::CommonException( "ConversionHandler:: Invalid algorithm. See help." );
}
