/*
 * BoyerMooreHorspool.cpp
 *
 *  Created on: 5. 11. 2014
 *      Author: Jan Travnicek
 */

#include "BoyerMooreHorspool.h"
#include "BadCharacterShiftTable.h"
#include "SubtreeJumpTable.h"
#include "ExactSubtreeRepeatsNaive.h"

#include <tree/Tree.h>
#include <tree/ranked/PrefixRankedBarTree.h>
#include <tree/ranked/PrefixRankedBarPattern.h>
#include <tree/ranked/PrefixRankedBarNonlinearPattern.h>
#include <alphabet/RankedSymbol.h>

#include <map>

namespace arbology {

namespace exact {

std::set < unsigned > BoyerMooreHorspool::match ( const tree::Tree & subject, const tree::Tree & pattern ) {
	return dispatch ( subject.getData ( ), pattern.getData ( ) );
}

std::set < unsigned > BoyerMooreHorspool::match ( const tree::PrefixRankedBarTree & subject, const tree::PrefixRankedBarTree & pattern ) {
	return match ( subject, tree::PrefixRankedBarPattern ( pattern ) );
}

auto BoyerMooreHorspoolPrefixRankedBarTreePrefixRankedBarTree = BoyerMooreHorspool::RegistratorWrapper < std::set < unsigned >, tree::PrefixRankedBarTree, tree::PrefixRankedBarTree > ( BoyerMooreHorspool::match );

std::set < unsigned > BoyerMooreHorspool::match ( const tree::PrefixRankedBarTree & subject, const tree::PrefixRankedBarPattern & pattern ) {
	std::set < unsigned > occ;
	std::map < alphabet::RankedSymbol, size_t > bcs = BadCharacterShiftTable::bcs ( pattern ); //NOTE: the subjects alphabet must be a subset or equal to the pattern
	std::vector < int > subjectSubtreeJumpTable = SubtreeJumpTable::compute ( subject );

	 // index to the subject
	unsigned i = 0;

	 // main loop of the algorithm over all possible indexes where the pattern can start
	while ( i + pattern.getContent ( ).size ( ) <= subject.getContent ( ).size ( ) ) {

		 // index to the pattern
		int j = pattern.getContent ( ).size ( ) - 1;

		 // offset to the subject
		int offset = i + j;

		while ( ( j >= 0 ) && ( offset >= 0 ) ) {
			if ( subject.getContent ( )[offset] == pattern.getContent ( )[j] ) {
				 // match of symbol
				offset = offset - 1;
				j = j - 1;
			} else if ( ( pattern.getContent ( )[j] == pattern.getVariablesBar ( ) ) && ( pattern.getBars ( ).count ( subject.getContent ( )[offset] )) ) {
				 // match of variable with subtree
				offset = subjectSubtreeJumpTable[offset];
				j = j - 2;
			} else {
				break;
			}
		}

		 // match was found
		if ( j == -1 ) occ.insert ( offset );

		 // shift heuristics
		i += bcs[subject.getContent ( )[i + pattern.getContent ( ).size ( ) - 1]];
	}

	return occ;
}

auto BoyerMooreHorspoolPrefixRankedBarTreePrefixRankedBarPattern = BoyerMooreHorspool::RegistratorWrapper < std::set < unsigned >, tree::PrefixRankedBarTree, tree::PrefixRankedBarPattern > ( BoyerMooreHorspool::match );

std::set < unsigned > BoyerMooreHorspool::match ( const tree::PrefixRankedBarTree & subject, const tree::PrefixRankedBarNonlinearPattern & pattern ) {
	std::set < unsigned > occ;
	std::map < alphabet::RankedSymbol, size_t > bcs = BadCharacterShiftTable::bcs ( pattern ); //NOTE: the subjects alphabet must be a subset or equal to the pattern
	std::vector < int > subjectSubtreeJumpTable = SubtreeJumpTable::compute ( subject );
	std::map < alphabet::RankedSymbol, alphabet::Symbol > variablesSetting;

	tree::PrefixRankedBarTree repeats = ExactSubtreeRepeatsNaive::repeats ( subject );

	 // index to the subject
	unsigned i = 0;

	 // main loop of the algorithm over all possible indexes where the pattern can start
	while ( i + pattern.getContent ( ).size ( ) <= subject.getContent ( ).size ( ) ) {
		 // clear the current state of variable to subtree repeat
		variablesSetting.clear();

		 // index to the pattern
		int j = pattern.getContent ( ).size ( ) - 1;

		 // offset to the subject
		int offset = i + j;

		while ( ( j >= 0 ) && ( offset >= 0 ) ) {
			if ( subject.getContent ( )[offset] == pattern.getContent ( )[j] ) {
				 // match of symbol
				offset = offset - 1;
				j = j - 1;
			} else if ( ( pattern.getContent ( )[j] == pattern.getVariablesBar ( ) ) && ( pattern.getBars ( ).count ( subject.getContent ( )[offset] )) ) {
				 // else match of variable with subtree
				offset = subjectSubtreeJumpTable[offset];
				j = j - 2;

				 // check nonlinear variable
				if ( pattern.getNonlinearVariables ( ).count ( pattern.getContent ( )[ j + 1 ] ) ) {
					auto setting = variablesSetting.find ( pattern.getContent ( )[ j + 1 ] );

					if ( setting != variablesSetting.end ( ) && repeats.getContent ( )[ offset + 1 ].getSymbol ( ) != setting->second )
						break;

					variablesSetting.insert ( std::make_pair ( pattern.getContent ( )[ j + 1 ], repeats.getContent( )[ offset + 1 ].getSymbol ( ) ) );
				}
			} else {
				break;
			}
		}

		 // match was found
		if ( j == -1 ) occ.insert ( offset );

		 // shift heuristics
		i += bcs[subject.getContent ( )[i + pattern.getContent ( ).size ( ) - 1]];
	}

	return occ;
}

auto BoyerMooreHorspoolPrefixRankedBarTreePrefixRankedBarNonlinearPattern = BoyerMooreHorspool::RegistratorWrapper < std::set < unsigned >, tree::PrefixRankedBarTree, tree::PrefixRankedBarNonlinearPattern > ( BoyerMooreHorspool::match );

} /* namespace exact */

} /* namespace arbology */
