#include "RHDPDACommon.h"
#include "label/LabelPairLabel.h"
#include "label/LabelSetLabel.h"
#include "label/ObjectLabel.h"
#include "object/Object.h"
#include "alphabet/LabeledSymbol.h"

#include "automaton/PDA/VisiblyPushdownDPDA.h"
#include "automaton/PDA/VisiblyPushdownNPDA.h"
#include "automaton/PDA/RealTimeHeightDeterministicDPDA.h"
#include "automaton/PDA/RealTimeHeightDeterministicNPDA.h"

namespace automaton {

namespace determinize {

label::Label packToStateLabel(std::set<std::pair<label::Label, label::Label>>&& data) {
	std::set<label::Label> res;
	for(auto&& subData : std::move(data)) {
		res.insert(label::Label(label::LabelPairLabel(std::move(subData))));
	}
	return label::Label(label::LabelSetLabel(std::move(res)));
}

std::set<std::pair<label::Label, label::Label>> unpackFromStateLabel(const label::Label& data) {
	std::set<std::pair<label::Label, label::Label>> res;
	const std::set<label::Label>& labelSet = static_cast<const label::LabelSetLabel&>(data.getData()).getData();
	for (const auto& subData : labelSet) {
		const std::pair<label::Label, label::Label>& labelPair = static_cast<const label::LabelPairLabel&>(subData.getData()).getData();
		res.insert(labelPair);
	}
	return res;
}

label::Label packToStackSymbolLabel(std::pair<std::set<std::pair<label::Label, label::Label>>, alphabet::Symbol>&& data) {
	label::Label res1 = packToStateLabel(std::move(data.first));
	label::Label res2 = label::Label(label::ObjectLabel(alib::Object(std::move(data.second.getData()))));
	return label::Label(label::LabelPairLabel(std::make_pair(std::move(res1), std::move(res2))));
}

std::pair<std::set<std::pair<label::Label, label::Label>>, alphabet::Symbol> unpackFromDVPAStackSymbol(const alphabet::Symbol& symbol) {
	const label::Label& data = static_cast<const alphabet::LabeledSymbol&>(symbol.getData()).getLabel();
	const label::Label& labelPairFirst = static_cast<const label::LabelPairLabel&>(data.getData()).getData().first;
	std::set<std::pair<label::Label, label::Label>> res = unpackFromStateLabel(labelPairFirst);

	const label::Label& labelPairSecond = static_cast<const label::LabelPairLabel&>(data.getData()).getData().second;
	const alib::Object& object = static_cast<const label::ObjectLabel&>(labelPairSecond.getData()).getData();
	return std::make_pair(std::move(res), alphabet::Symbol(static_cast<const alphabet::SymbolBase&>(object.getData())));
}

label::Label packToStackSymbolLabel(std::pair<std::set<std::pair<label::Label, label::Label>>, std::variant<string::Epsilon, alphabet::Symbol>>&& data) {
	label::Label res1 = packToStateLabel(std::move(data.first));
	if(data.second.is<string::Epsilon>()) {
		label::Label res2 = label::Label(label::ObjectLabel(alib::Object(std::move(data.second.get<string::Epsilon>()))));
		return label::Label(label::LabelPairLabel(std::make_pair(std::move(res1), std::move(res2))));
	} else {
		label::Label res2 = label::Label(label::ObjectLabel(alib::Object(std::move(data.second.get<alphabet::Symbol>().getData()))));
		return label::Label(label::LabelPairLabel(std::make_pair(std::move(res1), std::move(res2))));
	}
}

std::pair<std::set<std::pair<label::Label, label::Label>>, std::variant<string::Epsilon, alphabet::Symbol>> unpackFromDRHDPDAStackSymbol(const alphabet::Symbol& symbol) {
	const label::Label& data = static_cast<const alphabet::LabeledSymbol&>(symbol.getData()).getLabel();
	const label::Label& labelPairFirst = static_cast<const label::LabelPairLabel&>(data.getData()).getData().first;
	std::set<std::pair<label::Label, label::Label>> res = unpackFromStateLabel(labelPairFirst);

	const label::Label& labelPairSecond = static_cast<const label::LabelPairLabel&>(data.getData()).getData().second;
	const alib::Object& object = static_cast<const label::ObjectLabel&>(labelPairSecond.getData()).getData();
	if(object.getData() == string::Epsilon::EPSILON) {
		return std::make_pair(std::move(res), std::variant<string::Epsilon, alphabet::Symbol>(static_cast<const string::Epsilon&>(object.getData())));
	} else {
		return std::make_pair(std::move(res), std::variant<string::Epsilon, alphabet::Symbol>(alphabet::Symbol(static_cast<const alphabet::SymbolBase&>(object.getData()))));
	}
}

std::set<label::Label> retrieveDSubSet(const std::set<std::pair<label::Label, label::Label>>& localOperation) {
	std::set<label::Label> id;
	for(const auto& entry : localOperation) {
		id.insert(entry.second);
	}
	return id;
}

std::set<std::pair<label::Label, label::Label>> createIdentity(std::set<label::Label>&& states) {
	std::set<std::pair<label::Label, label::Label>> id;
	for(const label::Label& state : states) {
		id.insert(std::make_pair(state, state));
	}
	return id;
}

template<class T, class S, class R>
std::set<automaton::State> existsDirtyState(const T& d, const std::set<automaton::State>& rubbishStates, const std::map<std::pair<State, S>, std::pair<State, alphabet::Symbol> >& rubbishCallTransitions, const std::map<std::pair<State, S>, State>& rubbishLocalTransitions, const R& n) {
	std::set<automaton::State> dirtyStates;

	std::set<automaton::State> states = d.getStates();
	states.insert(rubbishStates.begin(), rubbishStates.end());

	auto callTransitions = d.getCallTransitions();
	callTransitions.insert(rubbishCallTransitions.begin(), rubbishCallTransitions.end());

	auto localTransitions = d.getLocalTransitions();
	localTransitions.insert(rubbishLocalTransitions.begin(), rubbishLocalTransitions.end());

	for(const automaton::State& state : states) {
		const label::Label& stateLabel = state.getName();
		const std::set<label::Label> dSubSet = retrieveDSubSet(unpackFromStateLabel(stateLabel));

		bool originalCallsLocals = false;
		for(const auto& transition : n.getCallTransitions()) {
			if(dSubSet.count(transition.first.first.getName())) {
				originalCallsLocals = true;
				break;
			}
		}
		if(!originalCallsLocals) for(const auto& transition : n.getLocalTransitions()) {
			if(dSubSet.count(transition.first.first.getName())) {
				originalCallsLocals = true;
				break;
			}
		}

		if(!originalCallsLocals) continue;
		bool deterministicCallLocals = false;

		for(const auto& transition : callTransitions) {
			if(stateLabel == transition.first.first.getName()) {
				deterministicCallLocals = true;
				break;
			}
		}
		if(deterministicCallLocals) for(const auto& transition : localTransitions) {
			if(stateLabel == transition.first.first.getName()) {
				deterministicCallLocals = true;
				break;
			}
		}

		if(originalCallsLocals && !deterministicCallLocals) dirtyStates.insert(state);

	}
	return dirtyStates;
}

template std::set<automaton::State> existsDirtyState(const automaton::VisiblyPushdownDPDA& d, const std::set<automaton::State>& rubbishStates, const std::map<std::pair<State, alphabet::Symbol>, std::pair<State, alphabet::Symbol> >& rubbishCallTransitions, const std::map<std::pair<State, alphabet::Symbol>, State>& rubbishLocalTransitions, const automaton::VisiblyPushdownNPDA& n);
template std::set<automaton::State> existsDirtyState(const automaton::RealTimeHeightDeterministicDPDA& d, const std::set<automaton::State>& rubbishStates, const std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>>, std::pair<State, alphabet::Symbol> >& rubbishCallTransitions, const std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>>, State>& rubbishLocalTransitions, const automaton::RealTimeHeightDeterministicNPDA& n);

template<class T>
void localClosure(std::set<label::Label>& states, const std::set<label::Label>& oldStates, const T& d) {
	std::set<label::Label> newStates;

	for(const label::Label& state : oldStates) {
		for(const auto& transition : d.getLocalTransitions()) {
			if(transition.second.getName() != state) continue;

			if(!states.count(transition.first.first.getName())) {
				states.insert(transition.first.first.getName());
				newStates.insert(transition.first.first.getName());
			}
		}

		for(const auto& transition : d.getReturnTransitions()) {
			if(transition.second.getName() != state) continue;

			const alphabet::Symbol& popSymbol = std::get<2>(transition.first);
			if(popSymbol == d.getBottomOfTheStackSymbol()) continue;
			const label::Label& popSymbolLabel = static_cast<const alphabet::LabeledSymbol&>(popSymbol.getData()).getLabel();
			const label::Label& statePart = static_cast<const label::LabelPairLabel&>(popSymbolLabel.getData()).getData().first;

			if(!states.count(statePart)) {
				states.insert(statePart);
				newStates.insert(statePart);
			}
		}
	}

	if(newStates.empty()) return;
	localClosure(states, newStates, d);
}

template void localClosure(std::set<label::Label>& states, const std::set<label::Label>& oldStates, const automaton::VisiblyPushdownDPDA& d);
template void localClosure(std::set<label::Label>& states, const std::set<label::Label>& oldStates, const automaton::RealTimeHeightDeterministicDPDA& d);

template<class T, class S, class R>
std::set<std::pair<automaton::State, alphabet::Symbol>> existsDirtyStateSymbol(const T& d, const std::set<automaton::State>& rubbishStates, const std::map<std::pair<State, S>, std::pair<State, alphabet::Symbol> >& rubbishCallTransitions, const std::map<std::tuple<State, S, alphabet::Symbol>, State>& rubbishReturnTransitions,  const R& n) {
	std::set<std::pair<automaton::State, alphabet::Symbol>> dirtyStateSymbols;

	std::set<automaton::State> states = d.getStates();
	states.insert(rubbishStates.begin(), rubbishStates.end());

	auto callTransitions = d.getCallTransitions();
	callTransitions.insert(rubbishCallTransitions.begin(), rubbishCallTransitions.end());

	auto returnTransitions = d.getReturnTransitions();
	returnTransitions.insert(rubbishReturnTransitions.begin(), rubbishReturnTransitions.end());

	for(const automaton::State& state : states) {
		const label::Label& stateLabel = state.getName();

		bool originalPops = false;
		const std::set<label::Label> dSubSet = retrieveDSubSet(unpackFromStateLabel(stateLabel));

		for(const auto& transition : n.getReturnTransitions()) {
			if(dSubSet.count(std::get<0>(transition.first).getName())) {
				originalPops = true;
				goto break2;
			}
		}
break2:		if(!originalPops) continue;

		std::set<label::Label> lc { stateLabel };
		localClosure(lc, std::set<label::Label>{stateLabel}, d); //intentional copy

		std::set<alphabet::Symbol> topSymbols;
		for(const label::Label& localState : lc) {
			for(const auto& transition : callTransitions) {
				const auto& to = transition.second;
				if(localState != to.first.getName()) continue;

				topSymbols.insert(to.second);
			}

			if(d.getInitialState().getName() == localState) {
				topSymbols.insert(d.getBottomOfTheStackSymbol());
			}
		}

		for(const auto& transition : returnTransitions) {
			if(stateLabel != std::get<0>(transition.first).getName()) continue;

			topSymbols.erase(std::get<2>(transition.first));
		}

		for(const alphabet::Symbol& topSymbol : topSymbols)
			dirtyStateSymbols.insert(std::make_pair(state, topSymbol));
	}
	return dirtyStateSymbols;
}

template std::set<std::pair<automaton::State, alphabet::Symbol>> existsDirtyStateSymbol(const automaton::VisiblyPushdownDPDA& d, const std::set<automaton::State>& rubbishStates, const std::map<std::pair<State, alphabet::Symbol>, std::pair<State, alphabet::Symbol> >& rubbishCallTransitions, const std::map<std::tuple<State, alphabet::Symbol, alphabet::Symbol>, State>& rubbishReturnTransitions, const automaton::VisiblyPushdownNPDA& n);
template std::set<std::pair<automaton::State, alphabet::Symbol>> existsDirtyStateSymbol(const automaton::RealTimeHeightDeterministicDPDA& d, const std::set<automaton::State>& rubbishStates, const std::map<std::pair<State, std::variant<string::Epsilon, alphabet::Symbol>>, std::pair<State, alphabet::Symbol> >& rubbishCallTransitions, const std::map<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, alphabet::Symbol>, State>& rubbishReturnTransitions, const automaton::RealTimeHeightDeterministicNPDA& n);

std::set<label::Label> retrieveLabels(const std::set<automaton::State>& states) {
	std::set<label::Label> labels;
	for(const automaton::State& state : states) {
		labels.insert(state.getName());
	}
	return labels;
}

} /* namespace automaton */

} /* namespace determinize */
