/*
 * RegularEquationSolver.cpp
 *
 *  Created on: 4. 3. 2014
 *	  Author: Tomas Pecka
 */

#include "RegularEquationSolver.h"
#include <regexp/unbounded/UnboundedRegExpElements.h>

#include <exception/CommonException.h>
#include <algorithm>

namespace equations {

void RegularEquationSolver::setSymbols(const std::set<alphabet::Symbol>& newSymbols) {
	std::set<alphabet::Symbol> removed, added;
	
	std::set_difference(nonterminalSymbols.begin(), nonterminalSymbols.end(), newSymbols.begin(), newSymbols.end(), std::inserter(removed, removed.end()));
	std::set_difference(newSymbols.begin(), newSymbols.end(), nonterminalSymbols.begin(), nonterminalSymbols.end(), std::inserter(added, added.end()));

	for(const auto& symb : removed) {
		removeSymbol(symb);
	}

	for(const auto& symb : added) {
		addSymbol(symb);
	}
}

void RegularEquationSolver::removeSymbol(const alphabet::Symbol& symb) {
	for(const auto& kv : equationTransition) {
		const alphabet::Symbol& from = kv.first.first;
		const alphabet::Symbol& to = kv.first.second;
		const regexp::UnboundedRegExpAlternation& alt = kv.second;

		if((from == symb || to == symb) && alt.getElements().size() != 0) {
			throw exception::CommonException("Symbol '" + (std::string) symb + "' is in use.");
		}
	}

	for(const auto& kv : equationFinal) {
		const alphabet::Symbol& from = kv.first;
		const regexp::UnboundedRegExpAlternation& alt = kv.second;

		if(from == symb && alt.getElements().size() != 0) {
			throw exception::CommonException("Symbol '" + (std::string) from + "' is in use.");
		}
	}


	nonterminalSymbols.erase(nonterminalSymbols.find(symb));
	equationFinal.erase(equationFinal.find(symb));

	for(const auto& s : nonterminalSymbols) {
		equationTransition.erase(equationTransition.find(std::make_pair(s, symb)));
		equationTransition.erase(equationTransition.find(std::make_pair(symb, s)));
	}
	equationTransition.erase(equationTransition.find(std::make_pair(symb, symb)));
}

void RegularEquationSolver::addSymbol(const alphabet::Symbol& symb) {
	for(const auto& s : nonterminalSymbols) {
		equationTransition.insert(std::make_pair(std::make_pair(symb, s), regexp::UnboundedRegExpAlternation { }));
		equationTransition.insert(std::make_pair(std::make_pair(s, symb), regexp::UnboundedRegExpAlternation { }));
	}
	equationTransition.insert(std::make_pair(std::make_pair(symb, symb), regexp::UnboundedRegExpAlternation { }));

	nonterminalSymbols.insert(symb);
	equationFinal.insert(std::make_pair(symb, regexp::UnboundedRegExpAlternation { }));
}

void RegularEquationSolver::addEquation(const alphabet::Symbol& from, const alphabet::Symbol& to, const regexp::UnboundedRegExpElement& eq) {
	if(nonterminalSymbols.count(from) == 0) {
		throw exception::CommonException("Symbol from ('" + (std::string) from + "') is not in equation system.");
	}

	if(nonterminalSymbols.count(to) == 0) {
		throw exception::CommonException("Symbol to ('" + (std::string) to + "') is not in equation system.");
	}

	equationTransition.find(std::make_pair(from, to))->second.appendElement(eq);
}

void RegularEquationSolver::addEquation(const alphabet::Symbol& from, const regexp::UnboundedRegExpElement& eq) {
	if(nonterminalSymbols.count(from) == 0) {
		throw exception::CommonException("Symbol from ('" + (std::string) from + "') is not in equation system.");
	}

	equationFinal.find(from)->second.appendElement(eq);
}

void RegularEquationSolver::sortSymbolsByDepth(const alphabet::Symbol& solveFor) {
	std::map<alphabet::Symbol, bool> visited;
	std::queue<alphabet::Symbol> queue;

	for(const auto& symbol : nonterminalSymbols) {
		visited[symbol] = false;
	}

	visited[solveFor] = true;
	queue.push(solveFor);

	while(! queue.empty()) {
		alphabet::Symbol s = queue.front();
		queue.pop();

		nonterminalSymbolsByDepth.push_back(s);

		for(const auto& kv : equationTransition) {
			// find all transitions from current symbol that are non-empty, enqueue transition's target symbol
			if(kv.first.first == s && visited[ kv.first.second ] == false && kv.second.getElements().size() > 0) {
				visited[kv.first.second] = true;
				queue.push(kv.first.second);
			}
		}
	}
}

regexp::UnboundedRegExp RegularEquationSolver::solve(const alphabet::Symbol& solveFor) {
	if(nonterminalSymbols.count(solveFor) == 0) {
		throw exception::CommonException("Symbol solveFor ('" + (std::string) solveFor + "') is not in equation system.");
	}

	/*
	 * Firstly, organize states by depth so we can output better looking
	 * expressions. We need to solve equation system for automaton's initial state,
	 * so lets start with the deepest ones and walk towards the initial one.
	 */
	sortSymbolsByDepth(solveFor);
	return eliminate();
}

} /* namespace equations */
