/*
 * Follow.cpp
 *
 *  Created on: 9. 6. 2015
 *	  Author: Tomas Pecka
 */

#include "Follow.h"
#include "First.h"
#include <algorithm>
#include <string/Epsilon.h>
#include <iterator>

#include <grammar/ContextFree/CFG.h>
#include <grammar/ContextFree/EpsilonFreeCFG.h>
#include <grammar/ContextFree/CNF.h>
#include <grammar/ContextFree/GNF.h>
#include <grammar/ContextFree/LG.h>
#include <grammar/Regular/LeftLG.h>
#include <grammar/Regular/LeftRG.h>
#include <grammar/Regular/RightLG.h>
#include <grammar/Regular/RightRG.h>
#include <exception/CommonException.h>

namespace grammar {

namespace parsing {

template < class T >
void Follow::follow ( const T & grammar, std::map < alphabet::Symbol, std::set < std::variant < alphabet::Symbol, string::Epsilon > > > & followSet ) {
	for ( const std::pair < const alphabet::Symbol, std::set < std::vector < alphabet::Symbol > > > & rule : grammar.getRawRules ( ) ) {
		const alphabet::Symbol & X = rule.first;

		for ( const std::vector < alphabet::Symbol > & rhs : rule.second )
			 // every nt in rhs is Y
			for ( std::vector < alphabet::Symbol >::const_iterator it = rhs.begin ( ); it != rhs.end ( ); it++ ) {
				const alphabet::Symbol & Y = * it;

				if ( !grammar.getNonterminalAlphabet ( ).count ( Y ) ) continue;

				std::set < std::variant < alphabet::Symbol, string::Epsilon > > firstBeta = First::first ( grammar, std::vector < alphabet::Symbol > ( std::next ( it ), rhs.end ( ) ) );

				if ( firstBeta.count ( string::Epsilon::EPSILON ) ) {
					firstBeta.erase ( string::Epsilon::EPSILON );
					followSet[Y].insert ( followSet[X].begin ( ), followSet[X].end ( ) );
				}

				followSet[Y].insert ( firstBeta.begin ( ), firstBeta.end ( ) );
			}

	}
}

template < class T >
std::map < alphabet::Symbol, std::set < std::variant < alphabet::Symbol, string::Epsilon > > > Follow::follow ( const T & grammar ) {
	/*
	 * 1. Follow(S) = { \varepsilon }
	 *    Follow(A) = {} forall A \in N, A \neq S
	 * 2. Forall p \in P:
	 *      if p == X -> \alpha Y \beta
	 *		Follow(Y) = Follow(Y) \cup (First(\beta) \setminus { \varepsilon})
	 *	if p == X -> \alpha Y \beta \wedge \varepsilon \in First(\beta)
	 *		Follow(Y) = Follow(Y) \cup Follow(X)
	 * 3. goto 2 if any follow set was changed in prev step.
	 */

	std::map < alphabet::Symbol, std::set < std::variant < alphabet::Symbol, string::Epsilon > > > followSet1;

	for ( const alphabet::Symbol & symb : grammar.getNonterminalAlphabet ( ) )
		followSet1[symb];

	followSet1[grammar.getInitialSymbol ( )] = { string::Epsilon::EPSILON };

	std::map < alphabet::Symbol, std::set < std::variant < alphabet::Symbol, string::Epsilon > > > followSet2 = followSet1;

	do {
		follow ( grammar, followSet2 );

		if ( followSet1 == followSet2 ) break;

		followSet1 = followSet2;
	} while ( true );

	return followSet1;
}

template < class T >
std::set < std::variant < alphabet::Symbol, string::Epsilon > > Follow::follow ( const T & grammar, const alphabet::Symbol & nt ) {
	if ( !grammar.getNonterminalAlphabet ( ).count ( nt ) )
		throw exception::CommonException ( "Follow: Given symbol is not nonterminal." );

	return follow ( grammar )[nt];
}

auto FollowCFG = FollowBase1::RegistratorWrapper < FollowResult1, grammar::CFG > ( Follow::follow );
auto FollowEpsilonFreeCFG = FollowBase1::RegistratorWrapper < FollowResult1, grammar::EpsilonFreeCFG > ( Follow::follow );
auto FollowGNF = FollowBase1::RegistratorWrapper < FollowResult1, grammar::GNF > ( Follow::follow );
auto FollowCNF = FollowBase1::RegistratorWrapper < FollowResult1, grammar::CNF > ( Follow::follow );
auto FollowLG  = FollowBase1::RegistratorWrapper < FollowResult1, grammar::LG > ( Follow::follow );
auto FollowLeftLG  = FollowBase1::RegistratorWrapper < FollowResult1, grammar::LeftLG > ( Follow::follow );
auto FollowLeftRG  = FollowBase1::RegistratorWrapper < FollowResult1, grammar::LeftRG > ( Follow::follow );
auto FollowRightLG = FollowBase1::RegistratorWrapper < FollowResult1, grammar::RightLG > ( Follow::follow );
auto FollowRightRG = FollowBase1::RegistratorWrapper < FollowResult1, grammar::RightRG > ( Follow::follow );

std::map < alphabet::Symbol, std::set < std::variant < alphabet::Symbol, string::Epsilon > > > Follow::follow ( const grammar::Grammar & grammar ) {
	return FollowBase1::dispatch ( grammar.getData ( ) );
}

auto FollowCFG2 = FollowBase2::RegistratorWrapper < FollowResult2, grammar::CFG > ( Follow::follow );
auto FollowEpsilonFreeCFG2 = FollowBase2::RegistratorWrapper < FollowResult2, grammar::EpsilonFreeCFG > ( Follow::follow );
auto FollowGNF2 = FollowBase2::RegistratorWrapper < FollowResult2, grammar::GNF > ( Follow::follow );
auto FollowCNF2 = FollowBase2::RegistratorWrapper < FollowResult2, grammar::CNF > ( Follow::follow );
auto FollowLG2	= FollowBase2::RegistratorWrapper < FollowResult2, grammar::LG > ( Follow::follow );
auto FollowLeftLG2	= FollowBase2::RegistratorWrapper < FollowResult2, grammar::LeftLG > ( Follow::follow );
auto FollowLeftRG2	= FollowBase2::RegistratorWrapper < FollowResult2, grammar::LeftRG > ( Follow::follow );
auto FollowRightLG2 = FollowBase2::RegistratorWrapper < FollowResult2, grammar::RightLG > ( Follow::follow );
auto FollowRightRG2 = FollowBase2::RegistratorWrapper < FollowResult2, grammar::RightRG > ( Follow::follow );

std::set < std::variant < alphabet::Symbol, string::Epsilon > > Follow::follow ( const grammar::Grammar & grammar, const alphabet::Symbol & nt ) {
	return FollowBase2::dispatch ( grammar.getData ( ), nt );
}

} /* namespace parsing */

} /* namespace grammar */
