/*
 * RegExpOptimize.cpp
 *
 *  Created on: 20. 1. 2014
 *	  Author: Jan Travnicek
 */

namespace regexp {

namespace simplify {

void RegExpOptimize::optimize( FormalRegExpElement & element ) {
	FormalRegExpElement* optimized = optimizeInner( element );

	FormalRegExpAlternation * alternation = dynamic_cast<FormalRegExpAlternation *>( & element );
	if( alternation ) {
		FormalRegExpAlternation * alternationOptimized = dynamic_cast<FormalRegExpAlternation *>( optimized );
		if( alternationOptimized ) {
			* alternation = std::move( * alternationOptimized );
			delete alternationOptimized;
		} else {
			* alternation = FormalRegExpAlternation { std::manage_move ( optimized ), FormalRegExpEmpty { } };
		}
		return;
	}

	FormalRegExpConcatenation * concatenation = dynamic_cast<FormalRegExpConcatenation *>( & element );
	if( concatenation ) {
		FormalRegExpConcatenation * concatenationOptimized = dynamic_cast<FormalRegExpConcatenation *>( optimized );
		if( concatenationOptimized ) {
			* concatenation = std::move( * concatenationOptimized );
			delete concatenationOptimized;
		} else {
			* concatenation = FormalRegExpConcatenation { std::manage_move ( optimized ), FormalRegExpEpsilon { } };
		}
		return;
	}

	FormalRegExpIteration * iteration = dynamic_cast<FormalRegExpIteration *>( & element );
	if( iteration ) {
		FormalRegExpIteration * iterationOptimized = dynamic_cast<FormalRegExpIteration *>( optimized );
		if( iterationOptimized ) {
			* iteration = std::move( * iterationOptimized );
			delete iterationOptimized;
		} else {
			* iteration = FormalRegExpIteration { std::manage_move ( optimized ) };
		}
		return;
	}

	// Nothing to optimize original element was FormalRegExpSymbol, FormalRegExpEpsilon, or FormalRegExpEmpty
	return;
}

FormalRegExpElement* RegExpOptimize::optimizeInner( const FormalRegExpElement & node ) {
	FormalRegExpElement* elem = node.clone();

	// optimize while you can
	while(    A1( elem ) || A2( elem ) || A3( elem ) || A4( elem ) || A10( elem ) || V2( elem ) || V5( elem ) || V6( elem ) || X1( elem )
	       || A5( elem ) || A6( elem ) || A7( elem ) || A8( elem ) || A9( elem ) || V8( elem ) //|| V9( elem )
	       || A11( elem ) || V1( elem ) || V3( elem ) || V4( elem ) || V10( elem ) || S(elem) );

	return elem;
}

bool RegExpOptimize::S( FormalRegExpElement * & node ) {
	bool optimized = false;
	FormalRegExpAlternation * alternation = dynamic_cast<FormalRegExpAlternation*>( node );
	if( alternation ) {
		FormalRegExpElement * tmp = optimizeInner ( alternation->getLeftElement ( ) );
		if(* tmp != alternation->getLeftElement ( ) ) {
			optimized = true;
			alternation->setLeftElement ( * std::smart_ptr < FormalRegExpElement > ( tmp ) );
		}

		tmp = optimizeInner ( alternation->getRightElement ( ) );
		if(* tmp != alternation->getRightElement ( ) ) {
			optimized = true;
			alternation->setRightElement ( * std::smart_ptr < FormalRegExpElement > ( tmp ) );
		}

		return optimized;
	}

	FormalRegExpConcatenation * concatenation = dynamic_cast<FormalRegExpConcatenation*>( node );
	if( concatenation ) {
		FormalRegExpElement* tmp = optimizeInner ( concatenation->getLeftElement() );
		if(* tmp != concatenation->getLeftElement ( ) ) {
			optimized = true;
			concatenation->setLeftElement ( * std::smart_ptr < FormalRegExpElement > ( tmp ) );
		}

		tmp = optimizeInner ( concatenation->getRightElement ( ));
		if(* tmp != concatenation->getRightElement ( )) {
			optimized = true;
			concatenation->setRightElement ( * std::smart_ptr < FormalRegExpElement > ( tmp ) );
		}

		return optimized;
	}

	FormalRegExpIteration * iteration = dynamic_cast<FormalRegExpIteration*>( node );
	if( iteration ) {
		FormalRegExpElement* tmp = optimizeInner ( iteration->getElement() );

		if(* tmp != iteration->getElement ( ) ) {
			optimized = true;
			iteration->setElement ( * std::smart_ptr < FormalRegExpElement > ( tmp ) );
		}
		return optimized;
	}

	return optimized;
}


/**
  * optimization A1: ( x + y ) + z = x + ( y + z )
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::A1( FormalRegExpElement * & n ) {
	FormalRegExpAlternation * node = dynamic_cast<FormalRegExpAlternation *>( n );
	if( ! node ) return false;

	if( dynamic_cast<FormalRegExpAlternation *>( node->getLeft().get() ) ) {
		std::smart_ptr < FormalRegExpAlternation > leftAlt ( static_cast<FormalRegExpAlternation *>( node->getLeft().release() ) );

		std::smart_ptr < FormalRegExpElement > x = std::move ( leftAlt->getLeft() );
		std::smart_ptr < FormalRegExpElement > y = std::move ( leftAlt->getRight() );
		std::smart_ptr < FormalRegExpElement > z = std::move ( node->getRight() );

		leftAlt->setLeft ( std::move ( y ) );
		leftAlt->setRight ( std::move ( z ) );
		node->setLeft ( std::move ( x ) );
		node->setRight ( std::move ( leftAlt ) );

		return true;
	}

	return false;
}

/**
  * optimization A2: x + y = y + x (sort)
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::A2( FormalRegExpElement * & n ) {
	FormalRegExpAlternation * node = dynamic_cast<FormalRegExpAlternation *>( n );
	if( ! node ) return false;

	if( dynamic_cast<FormalRegExpAlternation *>( node->getRight().get() ) ) {
		FormalRegExpAlternation * rightAlt = static_cast < FormalRegExpAlternation * > ( node->getRight ( ).get ( ) );

		std::smart_ptr < FormalRegExpElement > x = std::move ( node->getLeft ( ) );
		std::smart_ptr < FormalRegExpElement > y = std::move ( rightAlt->getLeft ( ) );

		if(*x > *y) {
			node->setLeft ( std::move ( y ) );
			rightAlt->setLeft ( std::move ( x ) );
			return true;
		} else {
			node->setLeft ( std::move ( x ) );
			rightAlt->setLeft ( std::move ( y ) );
			return false;
		}
	}

	return false;
}

/**
  * optimization A3: x + \0 = x
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::A3( FormalRegExpElement * & n ) {
	FormalRegExpAlternation * node = dynamic_cast<FormalRegExpAlternation *>( n );
	if( ! node ) return false;

	// input can be \0 + \0, so at least one element must be preserved

	if( dynamic_cast<FormalRegExpEmpty *>( node->getRight().get() ) ) {
		n = node->getLeft().release();
		delete node;
		return true;
	}

	if( dynamic_cast<FormalRegExpEmpty *>( node->getLeft().get() ) ) {
		n = node->getRight().release();
		delete node;
		return true;
	}

	return false;
}

/**
  * optimization A4: x + x = x
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::A4( FormalRegExpElement * & n ) {
	/*
	 * two ways of implementing this opitimization:
	 * - sort and call std::unique ( O(n lg n) + O(n) ), but it also sorts...
	 * - check every element against other ( O(n*n) )
	 *
	 * As we always sort in optimization, we can use the first version, but A4 must be __always__ called __after__ A2
	 */

	FormalRegExpAlternation * node = dynamic_cast<FormalRegExpAlternation *>( n );
	if( ! node ) return false;

	if( node->getLeftElement() == node->getRightElement() ) {
		n = node->getRight().release();
		delete node;
		return true;
	}

	return false;
}

/**
  * optimization A5: x.(y.z) = (x.y).z = x.y.z
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::A5( FormalRegExpElement * & n ) {
	FormalRegExpConcatenation * node = dynamic_cast<FormalRegExpConcatenation *>( n );
	if( ! node ) return false;

	if( dynamic_cast<FormalRegExpConcatenation *>( node->getLeft().get() ) ) {
		std::smart_ptr < FormalRegExpConcatenation > leftCon ( static_cast<FormalRegExpConcatenation *>( node->getLeft().release() ) );

		std::smart_ptr < FormalRegExpElement > x = std::move ( leftCon->getLeft() );
		std::smart_ptr < FormalRegExpElement > y = std::move ( leftCon->getRight() );
		std::smart_ptr < FormalRegExpElement > z = std::move ( node->getRight() );

		leftCon->setLeft ( std::move ( y ) );
		leftCon->setRight ( std::move ( z ) );
		node->setLeft ( std::move ( x ) );
		node->setRight ( std::move ( leftCon ) );

		return true;
	}

	return false;
}

/**
  * optimization A6: \e.x = x.\e = x
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::A6( FormalRegExpElement * & n ) {
	FormalRegExpConcatenation * node = dynamic_cast<FormalRegExpConcatenation *>( n );
	if( ! node ) return false;

	// input can be \e + \e, so at least one element must be preserved

	if( dynamic_cast<FormalRegExpEpsilon *>( node->getRight().get() ) ) {
		n = node->getLeft().release();
		delete node;
		return true;
	}

	if( dynamic_cast<FormalRegExpEpsilon *>( node->getLeft().get() ) ) {
		n = node->getRight().release();
		delete node;
		return true;
	}

	return false;
}

/**
  * optimization A7: \0.x = x.\0 = \0
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::A7( FormalRegExpElement * & n ) {
	FormalRegExpConcatenation * node = dynamic_cast<FormalRegExpConcatenation *>( n );
	if( ! node ) return false;

	if( dynamic_cast<FormalRegExpEmpty *>( node->getRight().get() ) || dynamic_cast<FormalRegExpEmpty *>( node->getLeft().get() ) ) {
		delete node;
		n = new FormalRegExpEmpty { };
		return true;
	}

	return false;
}

/**
  * optimization A8: x.(y+z) = x.y + x.z
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::A8( FormalRegExpElement * & /* n */) {
	return false; //TODO
}

/**
  * optimization A9: (x+y).z = x.z + y.z
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::A9( FormalRegExpElement * & /* n */) {
	return false; //TODO
}

/**
  * optimization A10: x* = \e + x*x
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::A10( FormalRegExpElement * & n ) {
	/*
	 * problem:
	 * - \e + x*x = x*
	 * - but if we do not have the eps, but we do have iteration, then \e \in h(iter), therefore \e in h(node).
	 */

	FormalRegExpAlternation * node = dynamic_cast<FormalRegExpAlternation *>( n );
	if( ! node ) return false;

	if( dynamic_cast<FormalRegExpEpsilon *>( node->getLeft().get() ) ) {
		FormalRegExpConcatenation * rightCon = dynamic_cast<FormalRegExpConcatenation *>( node->getRight().get() );
		if( ! rightCon ) return false;

		FormalRegExpIteration * rightLeftIte = dynamic_cast<FormalRegExpIteration *>( rightCon->getLeft().get() );
		if( rightLeftIte ) {
			if(rightLeftIte->getElement() == rightCon->getRightElement()) {
				n = rightCon->getLeft().release();
				delete node;
				return true;
			}
		}

		FormalRegExpIteration * rightRightIte = dynamic_cast<FormalRegExpIteration *>( rightCon->getRight().get() );
		if( rightRightIte ) {
			if(rightRightIte->getElement() == rightCon->getLeftElement()) {
				n = rightCon->getRight().release();
				delete node;
				return true;
			}
		}
	}

	if( dynamic_cast<FormalRegExpEpsilon *>( node->getRight().get() ) ) {
		FormalRegExpConcatenation * leftCon = dynamic_cast<FormalRegExpConcatenation *>( node->getLeft().get() );
		if( ! leftCon ) return false;

		FormalRegExpIteration * leftLeftIte = dynamic_cast<FormalRegExpIteration *>( leftCon->getLeft().get() );
		if( leftLeftIte ) {
			if(leftLeftIte->getElement() == leftCon->getRightElement()) {
				n = leftCon->getLeft().release();
				delete node;
				return true;
			}
		}

		FormalRegExpIteration * leftRightIte = dynamic_cast<FormalRegExpIteration *>( leftCon->getRight().get() );
		if( leftRightIte ) {
			if(leftRightIte->getElement() == leftCon->getLeftElement()) {
				n = leftCon->getRight().release();
				delete node;
				return true;
			}
		}
	}

	return false;
}

/**
  * optimization A11: x* = (\e + x)*
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::A11( FormalRegExpElement * & n ) {
	FormalRegExpIteration * node = dynamic_cast<FormalRegExpIteration *>( n );
	if( ! node ) return false;

	FormalRegExpAlternation * childAlt = dynamic_cast<FormalRegExpAlternation *>( node->getChild().get() );
	if( childAlt ) {
		if( dynamic_cast < FormalRegExpEpsilon * > ( childAlt->getLeft ( ).get ( ) ) ) {
			node->setChild ( std::move ( childAlt->getRight ( ) ) );
			return true;
		}
		if( dynamic_cast < FormalRegExpEpsilon * > ( childAlt->getRight ( ).get ( ) ) ) {
			node->setChild ( std::move ( childAlt->getLeft ( ) ) );
			return true;
		}
	}

	return false;
}

/**
  * optimization V1: \0* = \e
  * optimization T1: \e* = \e
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::V1( FormalRegExpElement * & n ) {
	FormalRegExpIteration * node = dynamic_cast<FormalRegExpIteration *>( n );
	if( ! node ) return false;

	if( dynamic_cast<FormalRegExpEmpty*>( node->getChild ( ).get() ) ) {
		delete node;
		n = new FormalRegExpEpsilon( );
		return true;
	}
	if( dynamic_cast<FormalRegExpEpsilon*>( node->getChild ( ).get() ) ) {
		delete node;
		n = new FormalRegExpEpsilon( );
		return true;
	}
	return false;
}

/**
  * optimization V2: x* + x = x*
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::V2( FormalRegExpElement * & n ) {
	FormalRegExpAlternation * node = dynamic_cast<FormalRegExpAlternation *>( n );
	if( ! node ) return false;

	FormalRegExpIteration * leftIte = dynamic_cast<FormalRegExpIteration *>( node->getLeft().get() );
	if( leftIte ) {
		if(leftIte->getElement() == node->getRightElement()) {
			n = node->getLeft().release();
			delete node;
			return true;
		}
	}

	FormalRegExpIteration * rightIte = dynamic_cast<FormalRegExpIteration *>( node->getRight().get() );
	if( rightIte ) {
		if(rightIte->getElement() == node->getLeftElement()) {
			n = node->getRight().release();
			delete node;
			return true;
		}
	}

	return false;
}

/**
  * optimization V3: x** = x*
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::V3( FormalRegExpElement * & n ) {
	FormalRegExpIteration * node = dynamic_cast<FormalRegExpIteration *>( n );
	if( ! node ) return false;

	FormalRegExpIteration* childIter = dynamic_cast<FormalRegExpIteration*>( node->getChild().get() );
	if( childIter ) {
		node->setChild ( std::move ( childIter->getChild ( ) ) );
		return true;
	}

	return false;
}

/**
  * optimization V4: (x+y)* = (x*y*)*
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::V4( FormalRegExpElement * & n ) {
	FormalRegExpIteration * node = dynamic_cast<FormalRegExpIteration *>( n );
	if( ! node ) return false;

	FormalRegExpConcatenation * child = dynamic_cast<FormalRegExpConcatenation *>( node->getChild().get() );
	if( ! child ) return false;

	FormalRegExpIteration * leftIte = dynamic_cast<FormalRegExpIteration *>( child->getLeft().get() );
	if( ! leftIte ) return false;

	FormalRegExpIteration * rightIte = dynamic_cast<FormalRegExpIteration *>( child->getRight().get() );
	if( ! rightIte ) return false;

	n = new FormalRegExpIteration( FormalRegExpAlternation(std::move( leftIte->getElement ( ) ), std::move( rightIte->getElement ( ) ) ) );

	delete node;
	return true;
}

/**
  * optimization V5: x*y = y + x*xy
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::V5( FormalRegExpElement * & /* n */) {
	return false; //TODO
}

/**
  * optimization V6: x*y = y + xx*y
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::V6( FormalRegExpElement * & /* n */) {
	return false; //TODO
}

/**
  * optimization V8: \e in h(x) => xx*=x*
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::V8( FormalRegExpElement * & /* n */) {
	return false; //TODO
}

/**
  * optimization V9: (xy)*x = x(yx)*
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::V9( FormalRegExpElement * & /* n */) {
	return false; //TODO
}

/**
  * optimization V10: (x+y)* = (x*+y*)*
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::V10( FormalRegExpElement * & n ) {
	FormalRegExpAlternation * node = dynamic_cast<FormalRegExpAlternation *>( n );
	if( ! node ) return false;

	FormalRegExpIteration * leftIte = dynamic_cast<FormalRegExpIteration *>( node->getLeft().get() );
	if( ! leftIte ) return false;

	FormalRegExpIteration * rightIte = dynamic_cast<FormalRegExpIteration *>( node->getRight().get() );
	if( ! rightIte ) return false;

	n = new FormalRegExpConcatenation(std::move( leftIte->getElement() ), std::move( rightIte->getElement() ) );

	delete node;
	return true;
}

/**
  * optimization X1: a* + \e = a*
  * @param node FormalRegExpElement node
  * @return bool true if optimization applied else false
  */
bool RegExpOptimize::X1( FormalRegExpElement * & n ) {
	FormalRegExpAlternation * node = dynamic_cast<FormalRegExpAlternation *>( n );
	if( ! node ) return false;

	FormalRegExpIteration * leftIte = dynamic_cast<FormalRegExpIteration *>( node->getLeft().get() );
	if( leftIte ) {
		if(dynamic_cast<FormalRegExpEpsilon*>(node->getRight().get())) {
			n = node->getLeft().release();
			delete node;
			return true;
		}
	}

	FormalRegExpIteration * rightIte = dynamic_cast<FormalRegExpIteration *>( node->getRight().get() );
	if( rightIte ) {
		if(dynamic_cast<FormalRegExpEpsilon*>(node->getLeft().get())) {
			n = node->getRight().release();
			delete node;
			return true;
		}
	}

	return false;

}

} /* namespace simplify */

} /* namespace regexp */
