/*
 * GlushkovTraversal.cpp
 *
 *  Created on: 14. 4. 2016
 *	  Author: Tomas Pecka
 */

#include "GlushkovTraversal.h"
#include <iterator>
#include <vector>

namespace rte {

const GlushkovSymbol & GlushkovTraversal::findSymbol ( rte::FormalRTESymbol const * const symbol, const std::set < GlushkovSymbol > & symbolSet ) {
	auto it = find_if ( symbolSet.begin ( ), symbolSet.end ( ), [symbol] ( GlushkovSymbol const & gs ) -> bool {
			return gs.getSymbolPtr ( ) == symbol;
		} );

	if ( it == symbolSet.end ( ) )
		throw exception::CommonException ( "GlushkovTraversal RTE - Can not find GlushkovSymbol for regexp node. Probably symbol from constant alphabet?" );

	return * it;
}

bool GlushkovTraversal::pos ( GlushkovSymbol const & symbol, rte::FormalRTE const * const & node ) {
	return pos ( & node->getRTE ( ), symbol.getSymbolPtr ( ) );
}

std::set < GlushkovSymbol > GlushkovTraversal::first ( rte::FormalRTE const & re ) {
	std::set < GlushkovSymbol > firstSet, symbolSet = getSymbols ( re );

	for ( auto const & s : first ( & re.getRTE ( ) ) )
		 // std::cerr << "In first set: " << s -> getSymbol ( ) << std::endl;
		firstSet.insert ( findSymbol ( s, symbolSet ) );

	return firstSet;
}

/*
 * std::set<GlushkovSymbol> GlushkovTraversal::last( rte::FormalRTE const& re )
 * {
 *  std::set<GlushkovSymbol> lastSet, symbolSet = getSymbols( re );
 *
 *  for( auto const& s : last( & re.getRTE() ) )
 *      lastSet.insert( findSymbol( s, symbolSet ) );
 *
 *  return lastSet;
 * }
 */

std::set < std::vector < GlushkovSymbol > > GlushkovTraversal::follow ( rte::FormalRTE const & re, GlushkovSymbol const & symbol ) {
	std::set < GlushkovSymbol > symbolSet = getSymbols ( re );
	std::set < std::vector < GlushkovSymbol > > followSet;
	std::set < alphabet::RankedSymbol > alphabetK = re.getConstantAlphabet ( );
	std::map < alphabet::RankedSymbol, std::set < rte::FormalRTESymbol const * > > symbMap;

	for ( const auto & csymbol : alphabetK )
		symbMap.insert ( std::make_pair ( csymbol, std::set < rte::FormalRTESymbol const * > ( ) ) );

	for ( const auto & f : follow ( & re.getRTE ( ), symbol.getSymbolPtr ( ), alphabetK, symbMap ) ) {
		std::vector < GlushkovSymbol > curr;

		for ( const auto & s : f )
			curr.push_back ( findSymbol ( s, symbolSet ) );

		followSet.insert ( curr );
	}

	return followSet;
}

// -----------------------------------------------------------------------------

std::set < rte::FormalRTESymbol const * > GlushkovTraversal::first ( rte::FormalRTEElement const * const & node ) {
	rte::FormalRTEAlternation const * const alternation = dynamic_cast < rte::FormalRTEAlternation const * const > ( node );
	rte::FormalRTESubstitution const * const concatenation = dynamic_cast < rte::FormalRTESubstitution const * const > ( node );
	rte::FormalRTEIteration const * const iteration = dynamic_cast < rte::FormalRTEIteration const * const > ( node );
	rte::FormalRTESymbol const * const symbol = dynamic_cast < rte::FormalRTESymbol const * const > ( node );
	rte::FormalRTEEmpty const * const empty = dynamic_cast < rte::FormalRTEEmpty const * const > ( node );

	if ( alternation )
		return first ( alternation );
	else if ( concatenation )
		return first ( concatenation );
	else if ( iteration )
		return first ( iteration );
	else if ( empty )
		return first ( empty );
	else if ( symbol )
		return first ( symbol );

	throw exception::CommonException ( "GlushkovTraversal::first - invalid RegExpElement node" );
}

std::set < rte::FormalRTESymbol const * > GlushkovTraversal::first ( rte::FormalRTEAlternation const * const & node ) {
	std::set < rte::FormalRTESymbol const * > ret, tmp;

	tmp = first ( & node->getLeftElement ( ) );
	ret.insert ( tmp.begin ( ), tmp.end ( ) );

	tmp = first ( & node->getRightElement ( ) );
	ret.insert ( tmp.begin ( ), tmp.end ( ) );

	return ret;
}

std::set < rte::FormalRTESymbol const * > GlushkovTraversal::first ( rte::FormalRTESubstitution const * const & node ) {
	std::set < rte::FormalRTESymbol const * > ret, tmp;

	tmp = first ( & node->getLeftElement ( ) );
	ret.insert ( tmp.begin ( ), tmp.end ( ) );

	auto it = std::find_if ( ret.begin ( ), ret.end ( ), [node] ( rte::FormalRTESymbol const * a ) {
			return * a == node->getSubstitutionSymbol ( );
		} );

	if ( it != ret.end ( ) ) {
		ret.erase ( it );
		tmp = first ( & node->getRightElement ( ) );
		ret.insert ( tmp.begin ( ), tmp.end ( ) );
	}

	return ret;
}

std::set < rte::FormalRTESymbol const * > GlushkovTraversal::first ( rte::FormalRTEIteration const * const & node ) {
	std::set < rte::FormalRTESymbol const * > ret = first ( & node->getElement ( ) );

	ret.insert ( & node->getSubstitutionSymbol ( ) );
	return ret;
}

std::set < rte::FormalRTESymbol const * > GlushkovTraversal::first ( rte::FormalRTESymbol const * const & node ) {
	return std::set < rte::FormalRTESymbol const * > { node };
}

std::set < rte::FormalRTESymbol const * > GlushkovTraversal::first ( rte::FormalRTEEmpty const * const & /* node */ ) {
	return std::set < rte::FormalRTESymbol const * > ( );
}

// ----------------------------------------------------------------------------

/*
 * std::set<rte::FormalRTESymbol const *> GlushkovTraversal::last( rte::FormalRTEElement const * const & node )
 * {
 *  rte::FormalRTEAlternation const * const alternation = dynamic_cast<rte::FormalRTEAlternation const * const>( node );
 *  rte::FormalRTESubstitution const * const concatenation = dynamic_cast<rte::FormalRTESubstitution const * const>( node );
 *  rte::FormalRTEIteration const * const iteration = dynamic_cast<rte::FormalRTEIteration const * const>( node );
 *  rte::FormalRTESymbol const * const symbol = dynamic_cast<rte::FormalRTESymbol const * const>( node );
 *  rte::FormalRTEEmpty const * const empty = dynamic_cast<rte::FormalRTEEmpty const * const>( node );
 *
 *  if( symbol )
 *      return last( symbol );
 *  else if( alternation )
 *      return last( alternation );
 *  else if( concatenation )
 *      return last( concatenation );
 *  else if( iteration )
 *      return last( iteration );
 *  else if( empty )
 *      return last( empty );
 *
 *   throw exception::CommonException( "GlushkovTraversal::last - invalid RegExpElement node" );
 * }
 *
 * std::set<rte::FormalRTESymbol const *> GlushkovTraversal::last( rte::FormalRTEAlternation const * const & node )
 * {
 *  std::set<rte::FormalRTESymbol const *> ret;
 *
 *  for( auto const& element : node->getElements( ) )
 *  {
 *      std::set<rte::FormalRTESymbol const *> tmp = last( element );
 *      ret.insert( tmp.begin( ), tmp.end( ) );
 *  }
 *
 *  return ret;
 * }
 *
 * std::set<rte::FormalRTESymbol const *> GlushkovTraversal::last( rte::FormalRTESubstitution const * const & node )
 * {
 *  std::set<rte::FormalRTESymbol const *> ret, tmp;
 *
 *  for( auto it = node->getElements( ).rbegin( ); it != node->getElements( ).rend( ) ; it ++ )
 *  {
 *      tmp = last( *it );
 *      ret.insert( tmp.begin( ), tmp.end( ) );
 *
 *      if( ! rte::properties::RegExpEpsilon::languageContainsEpsilon(**it) )
 *          break;
 *  }
 *
 *  return ret;
 * }
 *
 * std::set<rte::FormalRTESymbol const *> GlushkovTraversal::last( rte::FormalRTEIteration const * const & node )
 * {
 *  return last( & node->getElement( ) );
 * }
 *
 * std::set<rte::FormalRTESymbol const *> GlushkovTraversal::last( rte::FormalRTESymbol const * const & node )
 * {
 *  return std::set<rte::FormalRTESymbol const *> { node };
 * }
 *
 * std::set<rte::FormalRTESymbol const *> GlushkovTraversal::last( rte::FormalRTEEmpty const * const & node )
 * {
 *  return std::set<rte::FormalRTESymbol const *>( );
 * }
 */

// ----------------------------------------------------------------------------

std::set < std::vector < rte::FormalRTESymbol const * > > GlushkovTraversal::follow ( rte::FormalRTEElement const * const & node, rte::FormalRTESymbol const * const & symbolptr, const std::set < alphabet::RankedSymbol > & alphabetK, std::map < alphabet::RankedSymbol, std::set < rte::FormalRTESymbol const * > > & subMap ) {
	rte::FormalRTEAlternation const * const alternation = dynamic_cast < rte::FormalRTEAlternation const * const > ( node );
	rte::FormalRTESubstitution const * const concatenation = dynamic_cast < rte::FormalRTESubstitution const * const > ( node );
	rte::FormalRTEIteration const * const iteration = dynamic_cast < rte::FormalRTEIteration const * const > ( node );
	rte::FormalRTESymbol const * const symbol = dynamic_cast < rte::FormalRTESymbol const * const > ( node );
	rte::FormalRTEEmpty const * const empty = dynamic_cast < rte::FormalRTEEmpty const * const > ( node );

	if ( alternation )
		return follow ( alternation, symbolptr, alphabetK, subMap );

	else if ( concatenation )
		return follow ( concatenation, symbolptr, alphabetK, subMap );

	else if ( iteration )
		return follow ( iteration, symbolptr, alphabetK, subMap );

	else if ( symbol )
		return follow ( symbol, symbolptr, alphabetK, subMap );

	else if ( empty )
		return follow ( empty, symbolptr, alphabetK, subMap );

	throw exception::CommonException ( "GlushkovTraversal::follow() - unknown RegExpElement node" );
}

#include <iostream>
std::ostream & operator <<( std::ostream & os, const std::set < std::vector < rte::FormalRTESymbol const * > > & x ) {
	os << "{";

	for ( const auto & f : x ) {
		os << "[";

		for ( const auto & ff : f )
			os << ff->getSymbol ( ).getSymbol ( ) << "(" << ff->getSymbol ( ).getRank ( ) << "), ";

		os << "], ";
	}

	os << "}";
	return os;
}

std::set < std::vector < rte::FormalRTESymbol const * > > GlushkovTraversal::follow ( rte::FormalRTEAlternation const * const & node, rte::FormalRTESymbol const * const & symbolptr, const std::set < alphabet::RankedSymbol > & alphabetK, std::map < alphabet::RankedSymbol, std::set < rte::FormalRTESymbol const * > > & subMap ) {
	std::set < std::vector < rte::FormalRTESymbol const * > > ret, tmp;

	tmp = follow ( & node->getLeftElement ( ), symbolptr, alphabetK, subMap );
	ret.insert ( tmp.begin ( ), tmp.end ( ) );

	tmp = follow ( & node->getRightElement ( ), symbolptr, alphabetK, subMap );
	ret.insert ( tmp.begin ( ), tmp.end ( ) );

	// std::cerr << "FollowAlt(" << symbolptr->getSymbol ( ) << "): " << ret << std::endl;

	return ret;
}

std::set < std::vector < rte::FormalRTESymbol const * > > GlushkovTraversal::follow ( rte::FormalRTESubstitution const * const & node, rte::FormalRTESymbol const * const & symbolptr, const std::set < alphabet::RankedSymbol > & alphabetK, std::map < alphabet::RankedSymbol, std::set < rte::FormalRTESymbol const * > > & subMap ) {

	std::map < alphabet::RankedSymbol, std::set < rte::FormalRTESymbol const * > > subMap2 ( subMap );
	auto itMap = subMap2.find ( node->getSubstitutionSymbol ( ).getSymbol ( ) );

	itMap->second.clear ( );

	for ( const auto & s : first ( & node->getRightElement ( ) ) )
		itMap->second.insert ( s );

	/*
	 * E sub F
	 *   1. if symbolptr in F subtree, then Follow(F, symbolptr);
	 *   2. if symbolptr in E subtree, then Follow(E, symbolptr);
	 */

	std::set < std::vector < rte::FormalRTESymbol const * > > ret;

	if ( pos ( & node->getLeftElement ( ), symbolptr ) )
		ret = follow ( & node->getLeftElement ( ), symbolptr, alphabetK, subMap2 );
	else
		ret = follow ( & node->getRightElement ( ), symbolptr, alphabetK, subMap );

	 // std::cerr << "FollowSub(" << symbolptr->getSymbol ( ) << "): " << ret << std::endl;
	return ret;
}

std::set < std::vector < rte::FormalRTESymbol const * > > GlushkovTraversal::follow ( rte::FormalRTEIteration const * const & node, rte::FormalRTESymbol const * const & symbolptr, const std::set < alphabet::RankedSymbol > & alphabetK, std::map < alphabet::RankedSymbol, std::set < rte::FormalRTESymbol const * > > & subMap ) {

	std::set < std::vector < rte::FormalRTESymbol const * > > ret;

	for ( const auto & s : first ( & node->getElement ( ) ) )
		subMap[node->getSubstitutionSymbol ( ).getSymbol ( )].insert ( s );

	ret = follow ( & node->getElement ( ), symbolptr, alphabetK, subMap );

	 // std::cerr << "FollowIter(" << symbolptr->getSymbol ( ) << "): " << ret << std::endl;
	return ret;
}

void preprocessSubMap ( const std::set < alphabet::RankedSymbol > & alphabetK, std::map < alphabet::RankedSymbol, std::set < rte::FormalRTESymbol const * > > & subMap ) {
	bool change = true;

	while ( change ) {
		change = false;

		for ( auto & kv : subMap ) {
			std::set < rte::FormalRTESymbol const * > & value = kv.second;

			for ( const auto & v : value ) {
				if ( alphabetK.find ( v->getSymbol ( ) ) == alphabetK.end ( ) ) continue;

				// std::cerr << "Gonna replace in submap: " << kv.first << ": " << v->getSymbol() << std::endl;

				auto it = subMap.find ( v->getSymbol ( ) );
				size_t vsize = value.size ( );
				value.insert ( it->second.begin ( ), it->second.end ( ) );
				change = ( vsize != value.size ( ) );
				value.erase ( v );

				/*
				 * for(const auto& x : it->second)
				 *  std::cerr << "\t" << x->getSymbol() << std::endl;
				 */
			}
		}
	}
}

std::set < std::vector < rte::FormalRTESymbol const * > > replaceConstants ( const std::set < alphabet::RankedSymbol > & alphabetK, const std::vector < std::smart_ptr < const rte::FormalRTESymbol > > & f, const std::map < alphabet::RankedSymbol, std::set < rte::FormalRTESymbol const * > > & subMap2 ) {
	// std::cerr << " replC for " << f << std::endl;

	std::set < std::vector < rte::FormalRTESymbol const * > > ret;

	if ( f.size ( ) == 0 ) return ret;

	std::map < alphabet::RankedSymbol, std::set < rte::FormalRTESymbol const * > > subMap ( subMap2 );
	preprocessSubMap ( alphabetK, subMap );

	auto subIt = subMap.find ( f[0]->getSymbol ( ) );

	if ( subIt == subMap.end ( ) ) {
		std::vector < rte::FormalRTESymbol const * > v;
		v.push_back ( f[0].get() );
		ret.insert ( v );
	} else {
		for ( const auto & subSymbol : subIt->second ) {
			std::vector < rte::FormalRTESymbol const * > v;
			v.push_back ( subSymbol );
			ret.insert ( v );
		}
	}

	for ( size_t i = 1; i < f.size ( ); i++ ) {
		std::set < std::vector < rte::FormalRTESymbol const * > > tmp;

		subIt = subMap.find ( f[i]->getSymbol ( ) );

		if ( subIt == subMap.end ( ) )
			for ( const auto & r : ret ) {
				std::vector < rte::FormalRTESymbol const * > v = r;
				v.push_back ( f[i].get() );
				tmp.insert ( v );
			}

		else

			for ( const auto & r : ret )
				for ( const auto & subSymbol : subIt->second ) {
					std::vector < rte::FormalRTESymbol const * > v = r;
					v.push_back ( subSymbol );
					tmp.insert ( v );
				}

		ret = tmp;
	}

	// std::cerr << "ReplC for " << f << " rets:" << std::endl;

	/*
	 * for ( const auto & r : ret ) {
	 *  std::cerr << "\t";
	 *
	 *  for ( const auto & x : r )
	 *      std::cerr << " " << * x;
	 *
	 *  std::cerr << std::endl;
	 * }
	 */

	return ret;
}

std::set < std::vector < rte::FormalRTESymbol const * > > GlushkovTraversal::follow ( rte::FormalRTESymbol const * const & node, rte::FormalRTESymbol const * const & symbolptr, const std::set < alphabet::RankedSymbol > & alphabetK, std::map < alphabet::RankedSymbol, std::set < rte::FormalRTESymbol const * > > & subMap ) {
	std::set < std::vector < rte::FormalRTESymbol const * > > ret, tmp;
	const std::vector < std::smart_ptr < const rte::FormalRTESymbol > >  & children = node->getElements ( );

	if ( symbolptr == node ) {
		ret = replaceConstants ( alphabetK, children, subMap );

		 // std::cerr << "FollowSymbol1(" << symbolptr->getSymbol ( ) << "): " << ret << std::endl;
		return ret;
	}

	for ( const auto & c : children ) {
		tmp = follow ( c.get(), symbolptr, alphabetK, subMap );
		ret.insert ( tmp.begin ( ), tmp.end ( ) );
	}

	return ret;
}

std::set < std::vector < rte::FormalRTESymbol const * > > GlushkovTraversal::follow ( rte::FormalRTEEmpty const * const & /* node */, rte::FormalRTESymbol const * const & /* symbolptr */, const std::set < alphabet::RankedSymbol > & /* alphabetK */, std::map < alphabet::RankedSymbol, std::set < rte::FormalRTESymbol const * > > & /* subMap */ ) {
	return std::set < std::vector < rte::FormalRTESymbol const * > > ( );
}

// ----------------------------------------------------------------------------

bool GlushkovTraversal::pos ( rte::FormalRTEElement const * const & node, rte::FormalRTESymbol const * const & symbolptr ) {
	rte::FormalRTEAlternation const * const alternation = dynamic_cast < rte::FormalRTEAlternation const * const > ( node );
	rte::FormalRTEIteration const * const iteration = dynamic_cast < rte::FormalRTEIteration const * const > ( node );
	rte::FormalRTESubstitution const * const concatenation = dynamic_cast < rte::FormalRTESubstitution const * const > ( node );
	rte::FormalRTESymbol const * const symbol = dynamic_cast < rte::FormalRTESymbol const * const > ( node );
	rte::FormalRTEEmpty const * const empty = dynamic_cast < rte::FormalRTEEmpty const * const > ( node );

	if ( alternation )
		return pos ( alternation, symbolptr );

	else if ( concatenation )
		return pos ( concatenation, symbolptr );

	else if ( iteration )
		return pos ( iteration, symbolptr );

	else if ( symbol )
		return pos ( symbol, symbolptr );

	else if ( empty )
		return pos ( empty, symbolptr );

	throw exception::CommonException ( "GlushkovTraversal::pos() - unknown RTE node" );
}

bool GlushkovTraversal::pos ( rte::FormalRTEAlternation const * const & node, rte::FormalRTESymbol const * const & symbolptr ) {
	return pos ( & node->getLeftElement ( ), symbolptr ) || pos ( & node->getRightElement ( ), symbolptr );
}

bool GlushkovTraversal::pos ( rte::FormalRTESubstitution const * const & node, rte::FormalRTESymbol const * const & symbolptr ) {
	return pos ( & node->getLeftElement ( ), symbolptr ) || pos ( & node->getRightElement ( ), symbolptr );
}

bool GlushkovTraversal::pos ( rte::FormalRTEIteration const * const & node, rte::FormalRTESymbol const * const & symbolptr ) {
	return pos ( & node->getElement ( ), symbolptr );
}

bool GlushkovTraversal::pos ( rte::FormalRTESymbol const * const & node, rte::FormalRTESymbol const * const & symbolptr ) {
	if ( symbolptr == node ) return true;

	for ( auto const & element : node->getElements ( ) )
		if ( pos ( element.get(), symbolptr ) )
			return true;

	return false;
}

bool GlushkovTraversal::pos ( rte::FormalRTEEmpty const * const & /* node */, rte::FormalRTESymbol const * const & /* symbolptr */ ) {
	return false;
}

// ----------------------------------------------------------------------------

std::set < GlushkovSymbol > GlushkovTraversal::getSymbols ( rte::FormalRTE const & re ) {
	std::set < GlushkovSymbol > posSet;
	int i = 1;

	std::set < alphabet::RankedSymbol > alphabetF = re.getAlphabet ( );

	getSymbols ( & re.getRTE ( ), alphabetF, posSet, i );

	return posSet;
}

void GlushkovTraversal::getSymbols ( rte::FormalRTEElement const * const & node, const std::set < alphabet::RankedSymbol > & alphabetF, std::set < GlushkovSymbol > & posSet, int & i ) {
	rte::FormalRTEAlternation const * const alternation = dynamic_cast < rte::FormalRTEAlternation const * const > ( node );
	rte::FormalRTEIteration const * const iteration = dynamic_cast < rte::FormalRTEIteration const * const > ( node );
	rte::FormalRTESubstitution const * const substitution = dynamic_cast < rte::FormalRTESubstitution const * const > ( node );
	rte::FormalRTESymbol const * const symbol = dynamic_cast < rte::FormalRTESymbol const * const > ( node );
	rte::FormalRTEEmpty const * const empty = dynamic_cast < rte::FormalRTEEmpty const * const > ( node );

	if ( symbol ) {
		if ( alphabetF.count ( symbol->getSymbol ( ) ) )
			posSet.insert ( GlushkovSymbol ( symbol, i++ ) );

		for ( const auto & c : symbol->getElements ( ) )
			getSymbols ( c.get(), alphabetF, posSet, i );

		return;
	} else if ( alternation ) {
		getSymbols ( & alternation->getLeftElement ( ), alphabetF, posSet, i );
		getSymbols ( & alternation->getRightElement ( ), alphabetF, posSet, i );
		return;
	} else if ( substitution ) {
		getSymbols ( & substitution->getLeftElement ( ), alphabetF, posSet, i );
		getSymbols ( & substitution->getRightElement ( ), alphabetF, posSet, i );
		return;
	} else if ( iteration ) {
		getSymbols ( & iteration->getElement ( ), alphabetF, posSet, i );
		return;
	} else if ( empty ) {
		return;
	}

	throw exception::CommonException ( "GlushkovTraversal::getSymbols() - unknown RegExpElement node" );
}

} /* namespace rte */
