/*
 * Author: Radovan Cerveny
 */

#include "SuffixAutomaton.h"
#include <string/Epsilon.h>
#include <label/LabelSetLabel.h>
#include "../../automaton/determinize/Determinize.h"
#include "../../automaton/simplify/Minimize.h"
#include "../../automaton/simplify/EpsilonRemoverIncoming.h"

namespace stringology {

namespace exact {

automaton::Automaton SuffixAutomaton::naiveConstruct ( const string::String & pattern ) {
    return dispatch ( pattern.getData ( ) );
}

automaton::DFA SuffixAutomaton::naiveConstruct ( const string::LinearString & pattern ) {
    automaton::EpsilonNFA nfaSuffixAutomaton ( automaton::State ( 0 ) );

    nfaSuffixAutomaton.setInputAlphabet ( pattern.getAlphabet ( ) );

    int i = 0;

    for ( const alphabet::Symbol & symbol : pattern.getContent ( ) ) {
        i++;
        nfaSuffixAutomaton.addState ( automaton::State ( i ) );
        nfaSuffixAutomaton.addTransition ( automaton::State ( i - 1 ), symbol, automaton::State ( i ) );
        nfaSuffixAutomaton.addTransition ( automaton::State ( 0 ), automaton::State ( i ) );
    }

    nfaSuffixAutomaton.addFinalState ( automaton::State ( i ) );

    automaton::DFA minimalSuffixAutomaton = automaton::simplify::Minimize::minimize ( automaton::determinize::Determinize::determinize ( automaton::simplify::EpsilonRemoverIncoming::remove ( nfaSuffixAutomaton ) ) );

    automaton::State failState = automaton::State ( label::Label ( label::LabelSetLabel ( { } ) ) );

    auto transitionsToFailState = minimalSuffixAutomaton.getTransitionsToState ( failState );

    for ( const auto & transition : transitionsToFailState )
        minimalSuffixAutomaton.removeTransition ( transition.first.first, transition.first.second, transition.second );

    minimalSuffixAutomaton.removeState ( failState );

    return minimalSuffixAutomaton;
}

automaton::Automaton SuffixAutomaton::construct ( const string::String & pattern ) {
    return dispatch ( pattern.getData ( ) );
}

automaton::DFA SuffixAutomaton::construct ( const string::LinearString & pattern ) {
    automaton::DFA suffixAutomaton ( automaton::State ( 0 ) );

    suffixAutomaton.setInputAlphabet ( pattern.getAlphabet ( ) );

    std::map < automaton::State, std::pair < automaton::State, int > > suffixLinks = { { automaton::State ( 0 ), { automaton::State ( -1 ), 0 } } };
    automaton::State lastState ( 0 );

    for ( const alphabet::Symbol & symbol : pattern.getContent ( ) )
        suffixAutomatonAddSymbol ( suffixAutomaton, symbol, suffixLinks, lastState );

    while ( lastState != automaton::State ( -1 ) ) {
        suffixAutomaton.addFinalState ( lastState );
        lastState = suffixLinks.find ( lastState )->second.first;
    }

    return suffixAutomaton;
}

void SuffixAutomaton::suffixAutomatonAddSymbol ( automaton::DFA & suffixAutomaton, const alphabet::Symbol & symbol, std::map < automaton::State, std::pair < automaton::State, int > > & suffixLinks, automaton::State & lastState ) {

    automaton::State newState ( ( int ) suffixAutomaton.getStates ( ).size ( ) );

    suffixAutomaton.addState ( newState );

    int lastSuffixLength = suffixLinks.find ( lastState )->second.second;

    suffixLinks.insert ( { newState, { automaton::State ( -1 ), lastSuffixLength + 1 } } );

    automaton::State kState = lastState;

    while ( kState != automaton::State ( -1 ) && suffixAutomaton.getTransitions ( ).find ( { kState, symbol } ) == suffixAutomaton.getTransitions ( ).end ( ) ) {
        suffixAutomaton.addTransition ( kState, symbol, newState );
        kState = suffixLinks.find ( kState )->second.first;
    }

    if ( kState == automaton::State ( -1 ) ) {
        suffixLinks.find ( newState )->second.first = automaton::State ( 0 );
    } else {
        automaton::State qState = suffixAutomaton.getTransitions ( ).find ( { kState, symbol } )->second;

        int kSuffixLength = suffixLinks.find ( kState )->second.second;
        int qSuffixLength = suffixLinks.find ( qState )->second.second;

        if ( kSuffixLength + 1 == qSuffixLength ) {
            suffixLinks.find ( newState )->second.first = qState;
        } else {

            automaton::State cloneState ( ( int ) suffixAutomaton.getStates ( ).size ( ) );
            suffixAutomaton.addState ( cloneState );

            suffixLinks.insert ( { cloneState, { suffixLinks.find ( qState )->second.first, kSuffixLength + 1 } } );

            for ( const auto & transition : suffixAutomaton.getTransitionsFromState ( qState ) )
                suffixAutomaton.addTransition ( cloneState, transition.first.second, transition.second );

            while ( kState != automaton::State ( -1 )
                && suffixAutomaton.getTransitions ( ).find ( { kState, symbol } ) != suffixAutomaton.getTransitions ( ).end ( )
                && suffixAutomaton.getTransitions ( ).find ( { kState, symbol } )->second == qState ) {
                suffixAutomaton.removeTransition ( kState, symbol, qState );
                suffixAutomaton.addTransition ( kState, symbol, cloneState );
                kState = suffixLinks.find ( kState )->second.first;
            }

            suffixLinks.find ( qState )->second.first = cloneState;
            suffixLinks.find ( newState )->second.first = cloneState;
        }
    }
    lastState = newState;
}

auto SuffixAutomatonLinearString = SuffixAutomaton::RegistratorWrapper < automaton::DFA, string::LinearString > ( SuffixAutomaton::construct );

} /* namespace exact */

} /* namespace stringology */
