#include "EdmondsTest.h"

#include <graph/spanningtree/Edmonds.h>

#include <graph/generate/RandomGraphFactory.h>
#include <exception/CommonException.h>
#include "../TestUtils.h"

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( GraphEdmondsTest, "graph" );
CPPUNIT_TEST_SUITE_REGISTRATION( GraphEdmondsTest );


#define GraphDef( edgeCount ) const struct { uint nodeCnt; uint edgeCnt; struct { uint node1; uint node2; uint weight; } edges [ edgeCount ]; }

GraphDef(1) graph1 = {
	2, 1,
	{
		{ 0, 1,   5 }
	}
};
GraphDef(1) spanningTree1 = {
	2, 1,
	{
		{ 0, 1,   5 }
	}
};

//----------------------------------
GraphDef(2) graph2 = {
	3, 2,
	{
		{ 0, 1,   5 },
		{ 1, 2,   6 }
	}
};
GraphDef(2) spanningTree2 = {
	3, 2,
	{
		{ 0, 1,   5 },
		{ 1, 2,   6 }
	}
};

//----------------------------------
GraphDef(2) graph3 = {
	3, 2,
	{
		{ 0, 1,   5 },
		{ 0, 2,   6 }
	}
};
GraphDef(2) spanningTree3 = {
	3, 2,
	{
		{ 0, 1,   5 },
		{ 0, 2,   6 }
	}
};

//----------------------------------
GraphDef(3) graph4 = {
	4, 3,
	{
		{ 0, 1,   5 },
		{ 0, 3,   6 },
		{ 2, 0,   7 }
	}
};
GraphDef(3) spanningTree4 = {
	4, 3,
	{
		{ 0, 1,   5 },
		{ 0, 3,   6 },
		{ 2, 0,   7 }
	}
};

//----------------------------------
GraphDef(3) graph5 = {
	3, 3,
	{
		{ 0, 1,   7 },
		{ 1, 2,   6 },
		{ 1, 2,   5 }
	}
};
GraphDef(2) spanningTree5 = {
	3, 2,
	{
		{ 0, 1,   7 },
		{ 1, 2,   6 },
	}
};

//----------------------------------
GraphDef(5) graph6 = {
	4, 5,
	{
		{ 0, 1,   4 },
		{ 0, 2,   1 },
		{ 0, 3,   2 },
		{ 1, 3,   3 },
		{ 3, 2,   5 }
	}
};
GraphDef(3) spanningTree6 = {
	4, 3,
	{
		{ 0, 2,   1 },
		{ 0, 1,   4 },
		{ 0, 3,   2 }
	}
};

//----------------------------------
GraphDef(10) graph7 = {
	6, 10,
	{
		{ 0, 1,   5 },
		{ 0, 2,   2 },
		{ 0, 4,   1 },
		{ 1, 3,   4 },
		{ 1, 5,   3 },
		{ 2, 3,   2 },
		{ 3, 4,  11 },
		{ 4, 2,   1 },
		{ 4, 1,   8 },
		{ 5, 0,   9 }
	}
};
GraphDef(5) spanningTree7 = {
	6, 5,
	{
		{ 0, 4,   1 },
		{ 4, 2,   1 },
		{ 2, 3,   2 },
		{ 0, 1,   5 },
		{ 1, 5,   3 }
	}
};


template< typename Type1, typename Type2 >
void testGraph( uint graphID, const Type1 * graphDef, const Type2 * spanTreeDef )
{
	std::vector<graph::Node> id2node( graphDef->nodeCnt );
	std::vector<graph::DirectedEdge> id2edge( graphDef->edgeCnt );
	std::unordered_map<graph::Node,uint> node2id;
	std::unordered_map<graph::DirectedEdge,uint> edge2id;

	graph::AdjacencyListDirectedGraph graph;

	for (uint i = 0; i < graphDef->nodeCnt; i++) {
		graph::Node node( std::string("n")+toString(i) );
		graph.addNode( node );
		id2node[ i ] = node;
		node2id[ node ] = i;
	}
	for (uint i = 0; i < graphDef->edgeCnt; i++) {
		graph::DirectedEdge edge( id2node[ graphDef->edges[i].node1 ], id2node[ graphDef->edges[i].node2 ], std::string("e")+toString(i) );
		graph.addEdge( edge, graphDef->edges[i].weight );
		id2edge[ i ] = edge;
		edge2id[ edge ] = i;
	}

	std::cout << "testing spanning tree #" << graphID << std::endl;

	graph::AdjacencyListDirectedGraph spanningTree = graph::spanningtree::Edmonds::edmonds( graph );

	CPPUNIT_ASSERT_EQUAL_MESSAGE( "invalid node count", spanTreeDef->nodeCnt, (uint)spanningTree.nodeCount() );
	CPPUNIT_ASSERT_EQUAL_MESSAGE( "invalid edge count", spanTreeDef->edgeCnt, (uint)spanningTree.edgeCount() );
	for (uint i = 0; i < spanTreeDef->edgeCnt; i++) {
		CPPUNIT_ASSERT_EQUAL_MESSAGE( "missing edge: "+toString(spanTreeDef->edges[i].node1)+","+toString(spanTreeDef->edges[i].node2),
			true, spanningTree.hasEdge( id2node[ spanTreeDef->edges[i].node1 ], id2node[ spanTreeDef->edges[i].node2 ] ) );
	}
}

void GraphEdmondsTest::runTests()
{
	testGraph( 1, &graph1, &spanningTree1 );
	testGraph( 2, &graph2, &spanningTree2 );
	testGraph( 3, &graph3, &spanningTree3 );
	testGraph( 4, &graph4, &spanningTree4 );
	testGraph( 5, &graph5, &spanningTree5 );
	testGraph( 6, &graph6, &spanningTree6 );
	testGraph( 7, &graph7, &spanningTree7 );
}
