/*
 * Component.hpp
 *
 *  Created on: Mar 16, 2016
 *      Author: Jan Travnicek
 */

#ifndef COMPONENT_HPP_
#define COMPONENT_HPP_

#include <set>
#include <algorithm>
#include <utility>
#include <tuple>
#include "../exception/CommonException.h"

namespace std {

/**
 * Represents a set of elements.
 * @param Derived class representing datatype using this set.
 * @param SymbolType underlying type of symbols in the set.
 * @param SetType arbitrary type used to distinguish different sets.
 */
template < class Derived, class SymbolType, class SetType >
class Component {
	/**
	 * The set.
	 */
	std::set < SymbolType > data;

	/**
	 * Checks whether symbol can be added to the set. Calls valid and available functions.
	 * @throws CommonException if symbol cannot be added.
	 */
	void checkAdd ( const SymbolType & symbol ) {
		valid ( symbol );

		if ( !available ( symbol ) )
			throw::exception::CommonException ( "Symbol " + ( std::string ) symbol + " is not available." );
	}

	/**
	 * Checks whether symbol can be removed from the set. Calls used function.
	 * @throws CommonException if symbol cannot be removed.
	 */
	void checkRemove ( const SymbolType & symbol ) {
		if ( used ( symbol ) )
			throw::exception::CommonException ( "Symbol " + ( std::string ) symbol + " is used." );
	}

protected:
	/** Checks the state of the set.
	 */
	void checkState ( ) {
		for ( const SymbolType & symbol : data )
			checkAdd ( symbol );
	}

public:
	/*
	 * Constructs an empty set.
	 */
	Component ( ) {
	}

	/**
	 * Constructs a set containing given elements.
	 * @throw CommonException if elements are not available in context of datatype where the set is used
	 */
	Component ( std::set < SymbolType > symbols ) : data ( std::move ( symbols ) ) {
	}

	/**
	 * Adds an elements to the set.
	 * @param element to add to the set
	 * @throw CommonException if element is not available in context of datatype where the set is used
	 * @return true if element was indeed added
	 *         false if element was present in the set
	 */
	bool add ( SymbolType symbol ) {
		checkAdd ( symbol );
		return data.insert ( std::move ( symbol ) ).second;
	}

	/**
	 * Adds a set of elements to the set.
	 * @param elements to add to the set
	 * @throw CommonException if one of the elements is not available in context of datatype where the set is used
	 */
	void add ( std::set < SymbolType > symbols ) {
		for ( SymbolType symbol : std::make_moveable_set ( symbols ) )
			add ( std::move ( symbol ) );
	}

	/**
	 * Changes the set.
	 * @param elements by which to replace those currently in the set
	 * @throw CommonException if one of the removed elements is used in context of datatype where the set is used
	 *        CommonException if one of the added elements is not available in context of datatype where the set is used
	 */
	void set ( std::set < SymbolType > symbols ) {
		std::set < SymbolType > removed;
		std::set_difference ( data.begin ( ), data.end ( ), symbols.begin ( ), symbols.end ( ), std::inserter ( removed, removed.end ( ) ) );

		for ( const SymbolType & symbol : removed )
			checkRemove ( symbol );

		for ( const SymbolType & symbol : symbols )
			checkAdd ( symbol );

		data = std::move ( symbols );
	}

	/**
	 * @return the set.
	 */
	std::set < SymbolType > & get ( ) {
		return data;
	}

	/**
	 * @return the set.
	 */
	const std::set < SymbolType > & get ( ) const {
		return data;
	}

	/**
	 * Removes an element from the set if not used.
	 * @throw CommonException if element is used in context of datatype where the set is used
	 * @return true if element was indeed removed
	 *         false if element was not present in the set
	 */
	bool remove ( const SymbolType & symbol ) {
		checkRemove ( symbol );
		return data.erase ( symbol );
	}

	/**
	 * Removes a set of elements from alphabet if not used.
	 * @throw CommonException if element is used in context of datatype where the set is used
	 */
	void remove ( const std::set < SymbolType > & symbols ) {
		for ( const SymbolType & symbol : symbols )
			remove ( symbol );
	}

	/**
	 * Component emptiness checker.
	 * @return true if set is an empty
	 */
	bool empty ( ) const {
		return data.empty ( );
	}

	/**
	 * Checks whether a concrete element is used in context of the datatype where the set is used
	 *
	 * To be implemented by all template instantiations explicitly
	 *
	 * @param element to check
	 * @return true if element is used
	 *         false if element is not used
	 */
	bool used ( const SymbolType & symbol ) const;

	/**
	 * Checks whether a concrete element is available in context of the datatype where the set is used
	 *
	 * To be implemented by all template instantiations explicitly
	 *
	 * @param element to check
	 * @return true if element is available
	 *         false if element is not available
	 */
	bool available ( const SymbolType & symbol ) const;

	/**
	 * Checks whether a concrete element is valid in context of the datatype where the set is used
	 *
	 * To be implemented by all template instantiations explicitly
	 *
	 * @param element to check
	 * @throw CommonException if the element in any way invalid
	 */
	void valid ( const SymbolType & symbol ) const;
};

/**
 * Represents an notable element.
 * @param Derived class representing datatype using this notable element.
 * @param SymbolType underlying type of element.
 * @param ElementType arbitrary type used to distinguish different notable elements.
 */
template < class Derived, class SymbolType, class ElementType >
class Element {
	/**
	 * The element.
	 */
	SymbolType data;

	/**
	 * Checks whether element can be set. Calls valid and available functions.
	 * @throws CommonException if element cannot be added.
	 */
	void checkSet ( const SymbolType & symbol ) {
		valid ( symbol );

		if ( !available ( symbol ) )
			throw::exception::CommonException ( "Symbol " + ( std::string ) symbol + " is not available." );
	}

protected:
	/** Checks the state of the element.
	 */
	void checkState ( ) {
		checkSet ( data );
	}

public:
	/**
	 * Constructs a notable element.
	 * @throw CommonException if element is not available in context of datatype where the class is used
	 */
	Element ( SymbolType symbol ) : data ( std::move ( symbol ) ) {
	}

	/**
	 * Changes the notable element.
	 * @param new value of the element
	 * @return bool true if the element was set
	 *              false if the element was the same as already present
	 * @throw CommonException if the new element is not available in context of datatype where the class is used
	 */
	bool set ( SymbolType symbol ) {
		checkSet ( symbol );

		if ( data == symbol ) return false;

		data = std::move ( symbol );
		return true;
	}

	/**
	 * Returns the current notable element of ElementType.
	 * @return the notable element
	 */
	SymbolType & get ( ) {
		return data;
	}

	/**
	 * Returns the current notable element of ElementType.
	 * @return the notable element
	 */
	const SymbolType & get ( ) const {
		return data;
	}

	/**
	 * Checks whether a concrete symbol is available in context of the datatype instance using this class
	 *
	 * To be implemented by all template instantiations explicitly
	 *
	 * @param element to check
	 * @return true if element is available
	 *         false if element is not available
	 */
	bool available ( const SymbolType & symbol ) const;

	/**
	 * Checks whether a concrete symbol is valid in context of the datatype instance using this class
	 *
	 * To be implemented by all template instantiations explicitly
	 *
	 * @param element to check
	 * @throw CommonException if the element in any way invalid
	 */
	void valid ( const SymbolType & symbol ) const;
};

/**
 * Auxiliary base handling all sets of elements from components
 */
template < class Derived, class SymbolType, class ... SetTypes >
struct ComponentAux;

/**
 * Specialisation for tuple.
 */
template < class Derived, class SymbolType, class ... SetTypes >
struct ComponentAux < Derived, SymbolType, tuple < SetTypes ... > > : public Component < Derived, SymbolType, SetTypes > ... {

	/**
	 * Constructor
	 */
	template < class ... RealSetTypes, size_t ... Indexes >
	ComponentAux ( tuple < RealSetTypes ... > params, std::index_sequence < Indexes ... > ) : Component < Derived, SymbolType, SetTypes > ( std::move ( get < Indexes > ( params ) ) ) ... {
		( void ) params; // No-op
	}

	/**
	 * Allows access to sub-component using its type.
	 * @param SetType alphabet type used to distinguish different sub alphabets
	 * @return sub-component
	 */
	template < class SetType, typename std::enable_if < std::is_in < SetType, SetTypes ... >::value >::type * = nullptr >
	const Component < Derived, SymbolType, SetType > & accessComponent ( ) const {
		return static_cast < const Component < Derived, SymbolType, SetType > & > ( * this );
	}

	/**
	 * Allows access to sub-component using its type.
	 * @param SetType alphabet type used to distinguish different sub alphabets
	 * @return sub-component
	 */
	template < class SetType, typename std::enable_if < std::is_in < SetType, SetTypes ... >::value >::type * = nullptr >
	Component < Derived, SymbolType, SetType > & accessComponent ( ) {
		return static_cast < Component < Derived, SymbolType, SetType > & > ( * this );
	}

protected:
	/**
	 * postponed checker function
	 */
	void checkState ( ) {
		( void ) std::initializer_list < int > { ( Component < Derived, SymbolType, SetTypes >::checkState ( ), 0 ) ... };
	}

};

/**
 * Auxiliary base handling all notable elements from components
 */
template < class Derived, class SymbolType, class ... ElementTypes >
struct ElementAux;

/**
 * Specialisation for tuple.
 */
template < class Derived, class SymbolType, class ... ElementTypes >
struct ElementAux < Derived, SymbolType, tuple < ElementTypes ... > > : public Element < Derived, SymbolType, ElementTypes > ... {

	/**
	 * Constructor
	 */
	template < class ... RealElementTypes, size_t ... Indexes >
	ElementAux ( tuple < RealElementTypes ... > params, std::index_sequence < Indexes ... > ) : Element < Derived, SymbolType, ElementTypes > ( std::move ( get < Indexes > ( params ) ) ) ... {
		( void ) params; // No-op
	}

	/**
	 * Allows access to sub-component using its type.
	 * @param ElementType alphabet type used to distinguish different sub alphabets
	 * @return sub-component
	 */
	template < class ElementType, typename std::enable_if < std::is_in < ElementType, ElementTypes ... >::value >::type * = nullptr >
	const Element < Derived, SymbolType, ElementType > & accessElement ( ) const {
		return static_cast < const Element < Derived, SymbolType, ElementType > & > ( * this );
	}

	/**
	 * Allows access to sub-component using its type.
	 * @param ElementType alphabet type used to distinguish different sub alphabets
	 * @return sub-component
	 */
	template < class ElementType, typename std::enable_if < std::is_in < ElementType, ElementTypes ... >::value >::type * = nullptr >
	Element < Derived, SymbolType, ElementType > & accessElement ( ) {
		return static_cast < Element < Derived, SymbolType, ElementType > & > ( * this );
	}

protected:
	/**
	 * postponed checker function
	 */
	void checkState ( ) {
		( void ) std::initializer_list < int > { ( Element < Derived, SymbolType, ElementTypes >::checkState ( ), 0 ) ... };
	}

};

template < class ... Types >
class Components;

/**
 * Auxiliary class allowing simple access to the alphabets.
 */
template < class Derived, class SymbolType, class SetTypesPack, class ElementTypesPack >
class Components < Derived, SymbolType, SetTypesPack, ElementTypesPack > : public ComponentAux < Derived, SymbolType, SetTypesPack >, public ElementAux < Derived, SymbolType, ElementTypesPack > {
public:
	/**
	 * Construct an alphabet pack from two alphabets.
	 */
	template < class RealSetTypes, class RealElementTypes >
	Components ( RealSetTypes params1, RealElementTypes params2 ) : ComponentAux < Derived, SymbolType, SetTypesPack > ( std::move ( params1 ), std::make_index_sequence < std::tuple_size < SetTypesPack >::value > { } ), ElementAux < Derived, SymbolType, ElementTypesPack > ( std::move ( params2 ), std::make_index_sequence < std::tuple_size < ElementTypesPack >::value > { } ) {
		ComponentAux < Derived, SymbolType, SetTypesPack >::checkState ( );

		ElementAux < Derived, SymbolType, ElementTypesPack >::checkState ( );
	}
};

/**
 * Auxiliary class allowing simple access to the alphabets.
 */
template < class Derived, class SymbolType, class SetTypesPack, class ElementTypesPack, class ... NextGroup >
class Components < Derived, SymbolType, SetTypesPack, ElementTypesPack, NextGroup ... > : public ComponentAux < Derived, SymbolType, SetTypesPack >, public ElementAux < Derived, SymbolType, ElementTypesPack >, public Components < Derived, NextGroup ... > {
public:
	/**
	 * Construct an alphabet pack from two alphabets.
	 */
	template < class RealSetTypes, class RealElementTypes, class ... NextRealTypes >
	Components ( RealSetTypes params1, RealElementTypes params2, NextRealTypes ... nextParams ) : ComponentAux < Derived, SymbolType, SetTypesPack > ( std::move ( params1 ), std::make_index_sequence < std::tuple_size < SetTypesPack >::value > { } ), ElementAux < Derived, SymbolType, ElementTypesPack > ( std::move ( params2 ), std::make_index_sequence < std::tuple_size < ElementTypesPack >::value > { } ), Components < Derived, NextGroup ... > ( std::move ( nextParams ) ... ) {
		ComponentAux < Derived, SymbolType, SetTypesPack >::checkState ( );

		ElementAux < Derived, SymbolType, ElementTypesPack >::checkState ( );
	}

	using ComponentAux < Derived, SymbolType, SetTypesPack >::accessComponent;
	using ElementAux < Derived, SymbolType, ElementTypesPack >::accessElement;

	using Components < Derived, NextGroup ... >::accessComponent;
	using Components < Derived, NextGroup ... >::accessElement;
};

} /* namespace std */

#endif /* COMPONENT_HPP_ */
