/*
 * Components2.hpp
 *
 *  Created on: Mar 16, 2016
 *      Author: Jan Travnicek
 */

#ifndef COMPONENTS2_HPP_
#define COMPONENTS2_HPP_

#include <set>
#include <algorithm>
#include <utility>
#include <tuple>
#include "../exception/CommonException.h"

namespace std {

template < class Derived, class DataType, class SetType >
class ComponentConstraint2 {
public:
	/**
	 * Checks whether a concrete element is used in context of the datatype where the set is used
	 *
	 * To be implemented by all template instantiations explicitly
	 *
	 * @param element to check
	 * @return true if element is used
	 *         false if element is not used
	 */
	static bool used ( const Derived & object, const DataType & symbol );

	/**
	 * Checks whether a concrete element is available in context of the datatype where the set is used
	 *
	 * To be implemented by all template instantiations explicitly
	 *
	 * @param element to check
	 * @return true if element is available
	 *         false if element is not available
	 */
	static bool available ( const Derived & object, const DataType & symbol );

	/**
	 * Checks whether a concrete element is valid in context of the datatype where the set is used
	 *
	 * To be implemented by all template instantiations explicitly
	 *
	 * @param element to check
	 * @throw CommonException if the element in any way invalid
	 */
	static void valid ( const Derived & object, const DataType & symbol );
};

/**
 * Represents a set of elements.
 * @param Derived class representing datatype using this set.
 * @param DataType underlying type of symbols in the set.
 * @param SetType arbitrary type used to distinguish different sets.
 */
template < class Derived, class DataType, class SetType >
class Component2 {
	/**
	 * The set.
	 */
	std::set < DataType > data;

	/**
	 * Checks whether symbol can be added to the set. Calls valid and available functions.
	 * @throws CommonException if symbol cannot be added.
	 */
	void checkAdd ( const DataType & symbol ) {
		ComponentConstraint2 < Derived, DataType, SetType >::valid ( static_cast < const Derived & > ( * this ), symbol );

		if ( !ComponentConstraint2 < Derived, DataType, SetType >::available ( static_cast < const Derived & > ( * this ), symbol ) )
			throw::exception::CommonException ( "Symbol " + ( std::string ) symbol + " is not available." );
	}

	/**
	 * Checks whether symbol can be removed from the set. Calls used function.
	 * @throws CommonException if symbol cannot be removed.
	 */
	void checkRemove ( const DataType & symbol ) {
		if ( ComponentConstraint2 < Derived, DataType, SetType >::used ( static_cast < const Derived & > ( * this ), symbol ) )
			throw::exception::CommonException ( "Symbol " + ( std::string ) symbol + " is used." );
	}

protected:
	/** Checks the state of the set.
	 */
	void checkState ( ) {
		for ( const DataType & symbol : data )
			checkAdd ( symbol );
	}

public:
	/**
	 * Constructs a set containing given elements.
	 * @throw CommonException if elements are not available in context of datatype where the set is used
	 */
	Component2 ( std::set < DataType > symbols ) : data ( std::move ( symbols ) ) {
	}

	/**
	 * Adds an elements to the set.
	 * @param element to add to the set
	 * @throw CommonException if element is not available in context of datatype where the set is used
	 * @return true if element was indeed added
	 *         false if element was present in the set
	 */
	bool add ( DataType symbol ) {
		checkAdd ( symbol );
		return data.insert ( std::move ( symbol ) ).second;
	}

	/**
	 * Adds a set of elements to the set.
	 * @param elements to add to the set
	 * @throw CommonException if one of the elements is not available in context of datatype where the set is used
	 */
	void add ( std::set < DataType > symbols ) {
		for ( DataType symbol : std::make_moveable_set ( symbols ) )
			add ( std::move ( symbol ) );
	}

	/**
	 * Changes the set.
	 * @param elements by which to replace those currently in the set
	 * @throw CommonException if one of the removed elements is used in context of datatype where the set is used
	 *        CommonException if one of the added elements is not available in context of datatype where the set is used
	 */
	void set ( std::set < DataType > symbols ) {
		std::set < DataType > removed;
		std::set_difference ( data.begin ( ), data.end ( ), symbols.begin ( ), symbols.end ( ), std::inserter ( removed, removed.end ( ) ) );

		for ( const DataType & symbol : removed )
			checkRemove ( symbol );

		for ( const DataType & symbol : symbols )
			checkAdd ( symbol );

		data = std::move ( symbols );
	}

	/**
	 * @return the set.
	 */
	std::set < DataType > & get ( ) {
		return data;
	}

	/**
	 * @return the set.
	 */
	const std::set < DataType > & get ( ) const {
		return data;
	}

	/**
	 * Removes an element from the set if not used.
	 * @throw CommonException if element is used in context of datatype where the set is used
	 * @return true if element was indeed removed
	 *         false if element was not present in the set
	 */
	bool remove ( const DataType & symbol ) {
		checkRemove ( symbol );
		return data.erase ( symbol );
	}

	/**
	 * Removes a set of elements from alphabet if not used.
	 * @throw CommonException if element is used in context of datatype where the set is used
	 */
	void remove ( const std::set < DataType > & symbols ) {
		for ( const DataType & symbol : symbols )
			remove ( symbol );
	}

	/**
	 * Component2 emptiness checker.
	 * @return true if set is an empty
	 */
	bool empty ( ) const {
		return data.empty ( );
	}

};

template < class Derived, class DataType, class SetType >
class ElementConstraint2 {
public:
	/**
	 * Checks whether a concrete symbol is available in context of the datatype instance using this class
	 *
	 * To be implemented by all template instantiations explicitly
	 *
	 * @param element to check
	 * @return true if element is available
	 *         false if element is not available
	 */
	static bool available ( const Derived & object, const DataType & symbol );

	/**
	 * Checks whether a concrete symbol is valid in context of the datatype instance using this class
	 *
	 * To be implemented by all template instantiations explicitly
	 *
	 * @param element to check
	 * @throw CommonException if the element in any way invalid
	 */
	static void valid ( const Derived & object, const DataType & symbol );
};

/**
 * Represents an notable element.
 * @param Derived class representing datatype using this notable element.
 * @param DataType underlying type of element.
 * @param ElementType arbitrary type used to distinguish different notable elements.
 */
template < class Derived, class DataType, class ElementType >
class Element2 {
	/**
	 * The element.
	 */
	DataType data;

	/**
	 * Checks whether element can be set. Calls valid and available functions.
	 * @throws CommonException if element cannot be added.
	 */
	void checkSet ( const DataType & symbol ) {
		ElementConstraint2 < Derived, DataType, ElementType >::valid ( static_cast < const Derived & > ( * this ), symbol );

		if ( !ElementConstraint2 < Derived, DataType, ElementType >::available ( static_cast < const Derived & > ( * this ), symbol ) )
			throw::exception::CommonException ( "Symbol " + ( std::string ) symbol + " is not available." );
	}

protected:
	/** Checks the state of the element.
	 */
	void checkState ( ) {
		checkSet ( data );
	}

public:
	/**
	 * Constructs a notable element.
	 * @throw CommonException if element is not available in context of datatype where the class is used
	 */
	Element2 ( DataType symbol ) : data ( std::move ( symbol ) ) {
	}

	/**
	 * Changes the notable element.
	 * @param new value of the element
	 * @return bool true if the element was set
	 *              false if the element was the same as already present
	 * @throw CommonException if the new element is not available in context of datatype where the class is used
	 */
	bool set ( DataType symbol ) {
		checkSet ( symbol );

		if ( data == symbol ) return false;

		data = std::move ( symbol );
		return true;
	}

	/**
	 * Returns the current notable element of ElementType.
	 * @return the notable element
	 */
	DataType & get ( ) {
		return data;
	}

	/**
	 * Returns the current notable element of ElementType.
	 * @return the notable element
	 */
	const DataType & get ( ) const {
		return data;
	}

};

/**
 * Auxiliary base handling all sets of elements from components
 */
template < class Derived, class DataType, class ... SetTypes >
struct ComponentAux2;

/**
 * Specialisation for tuple.
 */
template < class Derived, class DataType, class ... SetTypes >
struct ComponentAux2 < Derived, DataType, tuple < SetTypes ... > > : public Component2 < Derived, DataType, SetTypes > ... {

	/**
	 * Constructor
	 */
	template < class ... RealSetTypes, size_t ... Indexes >
	ComponentAux2 ( tuple < RealSetTypes ... > params, std::index_sequence < Indexes ... > ) : Component2 < Derived, DataType, SetTypes > ( std::move ( get < Indexes > ( params ) ) ) ... {
		( void ) params; // No-op
	}

	/**
	 * Allows access to sub-component using its type.
	 * @param SetType alphabet type used to distinguish different sub alphabets
	 * @return sub-component
	 */
	template < class SetType, typename std::enable_if < std::is_in < SetType, SetTypes ... >::value >::type * = nullptr >
	const Component2 < Derived, DataType, SetType > & accessComponent ( ) const {
		return static_cast < const Component2 < Derived, DataType, SetType > & > ( * this );
	}

	/**
	 * Allows access to sub-component using its type.
	 * @param SetType alphabet type used to distinguish different sub alphabets
	 * @return sub-component
	 */
	template < class SetType, typename std::enable_if < std::is_in < SetType, SetTypes ... >::value >::type * = nullptr >
	Component2 < Derived, DataType, SetType > & accessComponent ( ) {
		return static_cast < Component2 < Derived, DataType, SetType > & > ( * this );
	}

protected:
	/**
	 * postponed checker function
	 */
	void checkState ( ) {
		( void ) std::initializer_list < int > { ( Component2 < Derived, DataType, SetTypes >::checkState ( ), 0 ) ... };
	}

};

/**
 * Auxiliary base handling all notable elements from components
 */
template < class Derived, class DataType, class ... ElementTypes >
struct ElementAux2;

/**
 * Specialisation for tuple.
 */
template < class Derived, class DataType, class ... ElementTypes >
struct ElementAux2 < Derived, DataType, tuple < ElementTypes ... > > : public Element2 < Derived, DataType, ElementTypes > ... {

	/**
	 * Constructor
	 */
	template < class ... RealElementTypes, size_t ... Indexes >
	ElementAux2 ( tuple < RealElementTypes ... > params, std::index_sequence < Indexes ... > ) : Element2 < Derived, DataType, ElementTypes > ( std::move ( get < Indexes > ( params ) ) ) ... {
		( void ) params; // No-op
	}

	/**
	 * Allows access to sub-component using its type.
	 * @param ElementType alphabet type used to distinguish different sub alphabets
	 * @return sub-component
	 */
	template < class ElementType, typename std::enable_if < std::is_in < ElementType, ElementTypes ... >::value >::type * = nullptr >
	const Element2 < Derived, DataType, ElementType > & accessElement ( ) const {
		return static_cast < const Element2 < Derived, DataType, ElementType > & > ( * this );
	}

	/**
	 * Allows access to sub-component using its type.
	 * @param ElementType alphabet type used to distinguish different sub alphabets
	 * @return sub-component
	 */
	template < class ElementType, typename std::enable_if < std::is_in < ElementType, ElementTypes ... >::value >::type * = nullptr >
	Element2 < Derived, DataType, ElementType > & accessElement ( ) {
		return static_cast < Element2 < Derived, DataType, ElementType > & > ( * this );
	}

protected:
	/**
	 * postponed checker function
	 */
	void checkState ( ) {
		( void ) std::initializer_list < int > { ( Element2 < Derived, DataType, ElementTypes >::checkState ( ), 0 ) ... };
	}

};

template < class ... Types >
class Components2;

/**
 * Auxiliary class allowing simple access to the alphabets.
 */
template < class Derived, class DataType, class SetTypesPack, class ElementTypesPack >
class Components2 < Derived, DataType, SetTypesPack, ElementTypesPack > : public ComponentAux2 < Derived, DataType, SetTypesPack >, public ElementAux2 < Derived, DataType, ElementTypesPack > {
public:
	/**
	 * Construct an alphabet pack from two alphabets.
	 */
	template < class RealSetTypes, class RealElementTypes >
	Components2 ( RealSetTypes params1, RealElementTypes params2 ) : ComponentAux2 < Derived, DataType, SetTypesPack > ( std::move ( params1 ), std::make_index_sequence < std::tuple_size < SetTypesPack >::value > { } ), ElementAux2 < Derived, DataType, ElementTypesPack > ( std::move ( params2 ), std::make_index_sequence < std::tuple_size < ElementTypesPack >::value > { } ) {
		ComponentAux2 < Derived, DataType, SetTypesPack >::checkState ( );

		ElementAux2 < Derived, DataType, ElementTypesPack >::checkState ( );
	}
};

/**
 * Auxiliary class allowing simple access to the alphabets.
 */
template < class Derived, class DataType, class SetTypesPack, class ElementTypesPack, class ... NextGroup >
class Components2 < Derived, DataType, SetTypesPack, ElementTypesPack, NextGroup ... > : public ComponentAux2 < Derived, DataType, SetTypesPack >, public ElementAux2 < Derived, DataType, ElementTypesPack >, public Components2 < Derived, NextGroup ... > {
public:
	/**
	 * Construct an alphabet pack from two alphabets.
	 */
	template < class RealSetTypes, class RealElementTypes, class ... NextRealTypes >
	Components2 ( RealSetTypes params1, RealElementTypes params2, NextRealTypes ... nextParams ) : ComponentAux2 < Derived, DataType, SetTypesPack > ( std::move ( params1 ), std::make_index_sequence < std::tuple_size < SetTypesPack >::value > { } ), ElementAux2 < Derived, DataType, ElementTypesPack > ( std::move ( params2 ), std::make_index_sequence < std::tuple_size < ElementTypesPack >::value > { } ), Components2 < Derived, NextGroup ... > ( std::move ( nextParams ) ... ) {
		ComponentAux2 < Derived, DataType, SetTypesPack >::checkState ( );

		ElementAux2 < Derived, DataType, ElementTypesPack >::checkState ( );
	}

	using ComponentAux2 < Derived, DataType, SetTypesPack >::accessComponent;
	using ElementAux2 < Derived, DataType, ElementTypesPack >::accessElement;

	using Components2 < Derived, NextGroup ... >::accessComponent;
	using Components2 < Derived, NextGroup ... >::accessElement;
};

} /* namespace std */

#endif /* COMPONENTS2_HPP_ */
