/*
 * multipleDispatch.hpp
 *
 *  Created on: Apr 05, 2014
 *	  Author: Jan Travnicek
 */

#ifndef MULTIPLE_DISPATCH_H_
#define MULTIPLE_DISPATCH_H_

#include <stdexcept>
#include <functional>
#include <map>
#include <iostream>

#include "../cast/CastApi.hpp"
#include "../exception/CommonException.h"

namespace std {

template < class Algorithm, class ReturnType, class FirstParameterType >
class SingleDispatch {
public:
	class RegistratorWrapperBase {
	public:
		virtual ReturnType eval ( const FirstParameterType & ) = 0;
	};

private:
	std::map < std::type_index, RegistratorWrapperBase * > registeredFunctions;

	static SingleDispatch < Algorithm, ReturnType, FirstParameterType > & getInstance ( ) {
		static SingleDispatch < Algorithm, ReturnType, FirstParameterType > res;

		return res;
	}

public:
	template < class RealReturnType, class RealFirstParameterType >
	class RegistratorWrapper : public RegistratorWrapperBase {
		std::function < RealReturnType ( const RealFirstParameterType & ) > callback;

	public:
		ReturnType eval ( const FirstParameterType & first ) {
			return ReturnType ( callback ( ( const RealFirstParameterType & ) first ) );
		}

		RegistratorWrapper ( RealReturnType ( * callback ) ( const RealFirstParameterType & ) ) : callback ( callback ) {
			if ( !getInstance ( ).registeredFunctions.insert ( std::make_pair ( std::type_index ( typeid ( RealFirstParameterType ) ), this ) ).second ) {
				std::string firstType = std::cstringToString ( std::type_name < RealFirstParameterType > ( ) );

				std::string classType = std::cstringToString ( std::type_name < Algorithm > ( ) );

				throw::exception::CommonException ( "Callback for " + firstType + " already registered on " + classType + "." );
			}
		}

	};

	static ReturnType dispatch ( const FirstParameterType & first ) {
		typename std::map < std::type_index, RegistratorWrapperBase * >::iterator callback = getInstance ( ).registeredFunctions.find ( std::type_index ( typeid ( first ) ) );

		if ( callback == getInstance ( ).registeredFunctions.end ( ) ) {
			std::string firstType = std::cstringToString ( std::type_name ( typeid ( first ) ) );

			std::string classType = std::cstringToString ( std::type_name < Algorithm > ( ) );

			throw::exception::CommonException ( "Callback for " + firstType + " not registered on " + classType + "." );
		}

		return callback->second->eval ( first );
	}

};

template < class Algorithm, class ReturnType, class StaticParamType, class FirstParameterType >
class SingleDispatchFirstStaticParam {
public:
	class RegistratorWrapperBase {
	public:
		virtual ReturnType eval ( StaticParamType, const FirstParameterType & ) = 0;
	};

private:
	std::map < std::type_index, RegistratorWrapperBase * > registeredFunctions;

	static SingleDispatchFirstStaticParam < Algorithm, ReturnType, StaticParamType, FirstParameterType > & getInstance ( ) {
		static SingleDispatchFirstStaticParam < Algorithm, ReturnType, StaticParamType, FirstParameterType > res;

		return res;
	}

public:
	template < class RealReturnType, class RealFirstParameterType >
	class RegistratorWrapper : public RegistratorWrapperBase {
		std::function < RealReturnType ( StaticParamType, const RealFirstParameterType & ) > callback;

	public:
		ReturnType eval ( StaticParamType res, const FirstParameterType & first ) {
			return ReturnType ( callback ( res, ( const RealFirstParameterType & ) first ) );
		}

		RegistratorWrapper ( RealReturnType ( * callback ) ( StaticParamType, const RealFirstParameterType & ) ) : callback ( callback ) {
			if ( !getInstance ( ).registeredFunctions.insert ( std::make_pair ( std::type_index ( typeid ( RealFirstParameterType ) ), this ) ).second ) {
				std::string firstType = std::cstringToString ( std::type_name < RealFirstParameterType > ( ) );

				std::string classType = std::cstringToString ( std::type_name < Algorithm > ( ) );

				throw::exception::CommonException ( "Callback for " + firstType + " already registered on " + classType + "." );
			}
		}

	};

	static ReturnType dispatch ( StaticParamType res, const FirstParameterType & first ) {
		typename std::map < std::type_index, RegistratorWrapperBase * >::iterator callback = getInstance ( ).registeredFunctions.find ( std::type_index ( typeid ( first ) ) );

		if ( callback == getInstance ( ).registeredFunctions.end ( ) ) {
			std::string firstType = std::cstringToString ( std::type_name ( typeid ( first ) ) );

			std::string classType = std::cstringToString ( std::type_name < Algorithm > ( ) );

			throw::exception::CommonException ( "Callback for " + firstType + " not registered on " + classType + "." );
		}

		return callback->second->eval ( res, first );
	}

};

template < class Algorithm, class ReturnType, class FirstParameterType, class StaticParamType >
class SingleDispatchLastStaticParam {
public:
	class RegistratorWrapperBase {
	public:
		virtual ReturnType eval ( const FirstParameterType &, StaticParamType ) = 0;
	};

private:
	std::map < std::type_index, RegistratorWrapperBase * > registeredFunctions;

	static SingleDispatchLastStaticParam < Algorithm, ReturnType, FirstParameterType, StaticParamType > & getInstance ( ) {
		static SingleDispatchLastStaticParam < Algorithm, ReturnType, FirstParameterType, StaticParamType > res;

		return res;
	}

public:
	template < class RealReturnType, class RealFirstParameterType >
	class RegistratorWrapper : public RegistratorWrapperBase {
		std::function < RealReturnType ( const RealFirstParameterType &, StaticParamType ) > callback;

	public:
		ReturnType eval ( const FirstParameterType & first, StaticParamType res ) {
			return ReturnType ( callback ( ( const RealFirstParameterType & ) first, res ) );
		}

		RegistratorWrapper ( RealReturnType ( * callback ) ( const RealFirstParameterType &, StaticParamType ) ) : callback ( callback ) {
			if ( !getInstance ( ).registeredFunctions.insert ( std::make_pair ( std::type_index ( typeid ( RealFirstParameterType ) ), this ) ).second ) {
				std::string firstType = std::cstringToString ( std::type_name < RealFirstParameterType > ( ) );

				std::string classType = std::cstringToString ( std::type_name < Algorithm > ( ) );

				throw::exception::CommonException ( "Callback for " + firstType + " already registered on " + classType + "." );
			}
		}

	};

	static ReturnType dispatch ( const FirstParameterType & first, StaticParamType res ) {
		typename std::map < std::type_index, RegistratorWrapperBase * >::iterator callback = getInstance ( ).registeredFunctions.find ( std::type_index ( typeid ( first ) ) );

		if ( callback == getInstance ( ).registeredFunctions.end ( ) ) {
			std::string firstType = std::cstringToString ( std::type_name ( typeid ( first ) ) );

			std::string classType = std::cstringToString ( std::type_name < Algorithm > ( ) );

			throw::exception::CommonException ( "Callback for " + firstType + " not registered on " + classType + "." );
		}

		return callback->second->eval ( first, res );
	}

};

template < class Algorithm, class ReturnType, class FirstParameterType, class FirstStaticParamType, class SecondStaticParamType >
class SingleDispatch2LastStaticParams {
public:
	class RegistratorWrapperBase {
	public:
		virtual ReturnType eval ( const FirstParameterType &, FirstStaticParamType, SecondStaticParamType ) = 0;
	};

private:
	std::map < std::type_index, RegistratorWrapperBase * > registeredFunctions;

	static SingleDispatch2LastStaticParams < Algorithm, ReturnType, FirstParameterType, FirstStaticParamType, SecondStaticParamType > & getInstance ( ) {
		static SingleDispatch2LastStaticParams < Algorithm, ReturnType, FirstParameterType, FirstStaticParamType, SecondStaticParamType > res;

		return res;
	}

public:
	template < class RealReturnType, class RealFirstParameterType >
	class RegistratorWrapper : public RegistratorWrapperBase {
		std::function < RealReturnType ( const RealFirstParameterType &, FirstStaticParamType, SecondStaticParamType ) > callback;

	public:
		ReturnType eval ( const FirstParameterType & first, FirstStaticParamType res, SecondStaticParamType res2 ) {
			return ReturnType ( callback ( ( const RealFirstParameterType & ) first, res, res2 ) );
		}

		RegistratorWrapper ( RealReturnType ( * callback ) ( const RealFirstParameterType &, FirstStaticParamType, SecondStaticParamType ) ) : callback ( callback ) {
			if ( !getInstance ( ).registeredFunctions.insert ( std::make_pair ( std::type_index ( typeid ( RealFirstParameterType ) ), this ) ).second ) {
				std::string firstType = std::cstringToString ( std::type_name < RealFirstParameterType > ( ) );

				std::string classType = std::cstringToString ( std::type_name < Algorithm > ( ) );

				throw::exception::CommonException ( "Callback for " + firstType + " already registered on " + classType + "." );
			}
		}

	};

	static ReturnType dispatch ( const FirstParameterType & first, FirstStaticParamType res, SecondStaticParamType res2 ) {
		typename std::map < std::type_index, RegistratorWrapperBase * >::iterator callback = getInstance ( ).registeredFunctions.find ( std::type_index ( typeid ( first ) ) );

		if ( callback == getInstance ( ).registeredFunctions.end ( ) ) {
			std::string firstType = std::cstringToString ( std::type_name ( typeid ( first ) ) );

			std::string classType = std::cstringToString ( std::type_name < Algorithm > ( ) );

			throw::exception::CommonException ( "Callback for " + firstType + " not registered on " + classType + "." );
		}

		return callback->second->eval ( first, res, res2 );
	}

};

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

template < class Algorithm, class ReturnType, class ParametersType >
class PromotingDoubleDispatch {
public:
	class RegistratorWrapperBase {
	public:
		virtual ReturnType eval ( bool firstAttempt, const ParametersType &, const ParametersType & ) = 0;
		virtual bool available ( bool firstAttempt, std::type_index first, std::type_index second ) = 0;
	};

private:
	std::map < std::type_index, RegistratorWrapperBase * > registeredFunctions;

	static PromotingDoubleDispatch < Algorithm, ReturnType, ParametersType > & getInstance ( ) {
		static PromotingDoubleDispatch < Algorithm, ReturnType, ParametersType > res;

		return res;
	}

public:
	template < class RealReturnType, class RealParametersType >
	class RegistratorWrapper : public RegistratorWrapperBase {
		std::function < RealReturnType ( const RealParametersType &, const RealParametersType & ) > callback;

	public:
		ReturnType eval ( bool firstAttempt, const ParametersType & first, const ParametersType & second ) {
			if ( std::type_index ( typeid ( first ) ) == std::type_index ( typeid ( second ) ) )
				return ReturnType ( callback ( ( const RealParametersType & ) first, ( const RealParametersType & ) second ) );

			if ( firstAttempt ) {
				alib::Object casted = alib::castApi::getCastPool ( std::type_index ( typeid ( first ) ) ).cast ( second );
				return ReturnType ( callback ( ( const RealParametersType & ) first, ( const RealParametersType & ) casted.getData ( ) ) );
			} else {
				alib::Object casted = alib::castApi::getCastPool ( std::type_index ( typeid ( second ) ) ).cast ( first );
				return ReturnType ( callback ( ( const RealParametersType & ) casted.getData ( ), ( const RealParametersType & ) second ) );
			}
		}

		bool available ( bool firstAttempt, std::type_index first, std::type_index second ) {
			if ( first == second ) return true;

			if ( firstAttempt )
				return alib::castApi::castAvailable ( first, second );
			else
				return alib::castApi::castAvailable ( second, first );
		}

		RegistratorWrapper ( RealReturnType ( * callback ) ( const RealParametersType &, const RealParametersType & ) ) : callback ( callback ) {
			if ( !getInstance ( ).registeredFunctions.insert ( std::make_pair ( std::type_index ( typeid ( RealParametersType ) ), this ) ).second ) {
				std::string paramsType = std::cstringToString ( std::type_name < RealParametersType > ( ) );

				std::string classType = std::cstringToString ( std::type_name < Algorithm > ( ) );

				throw::exception::CommonException ( "Callback for " + paramsType + " already registered on " + classType + "." );
			}
		}

	};

	static ReturnType dispatch ( const ParametersType & first, const ParametersType & second ) {
		typename std::map < std::type_index, RegistratorWrapperBase * >::iterator callback = getInstance ( ).registeredFunctions.find ( std::type_index ( typeid ( first ) ) );

		if ( ( callback != getInstance ( ).registeredFunctions.end ( ) ) && callback->second->available ( true, std::type_index ( typeid ( first ) ), std::type_index ( typeid ( second ) ) ) )
			return callback->second->eval ( true, first, second );

		callback = getInstance ( ).registeredFunctions.find ( std::type_index ( typeid ( second ) ) );

		if ( ( callback != getInstance ( ).registeredFunctions.end ( ) ) && callback->second->available ( false, std::type_index ( typeid ( first ) ), std::type_index ( typeid ( second ) ) ) )
			return callback->second->eval ( false, first, second );

		std::string firstType = std::cstringToString ( std::type_name ( typeid ( first ) ) );

		std::string secondType = std::cstringToString ( std::type_name ( typeid ( second ) ) );

		std::string classType = std::cstringToString ( std::type_name < Algorithm > ( ) );

		throw::exception::CommonException ( "Callback for (" + firstType + ", " + secondType + ") (promoting) not registered on " + classType + "." );
	}

};

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

template < class Algorithm, class ReturnType, class FirstParameterType, class SecondParameterType >
class DoubleDispatch {
public:
	class RegistratorWrapperBase {
	public:
		virtual ReturnType eval ( const FirstParameterType &, const SecondParameterType & ) = 0;
	};

private:
	std::map < std::pair < std::type_index, std::type_index >, RegistratorWrapperBase * > registeredFunctions;

	static DoubleDispatch < Algorithm, ReturnType, FirstParameterType, SecondParameterType > & getInstance ( ) {
		static DoubleDispatch < Algorithm, ReturnType, FirstParameterType, SecondParameterType > res;

		return res;
	}

public:
	template < class RealReturnType, class RealFirstParameterType, class RealSecondParameterType >
	class RegistratorWrapper : public RegistratorWrapperBase {
		std::function < RealReturnType ( const RealFirstParameterType &, const RealSecondParameterType & ) > callback;

	public:
		ReturnType eval ( const FirstParameterType & first, const SecondParameterType & second ) {
			return ReturnType ( callback ( ( const RealFirstParameterType & ) first, ( const RealSecondParameterType & ) second ) );
		}

		RegistratorWrapper ( RealReturnType ( * callback ) ( const RealFirstParameterType &, const RealSecondParameterType & ) ) : callback ( callback ) {
			if ( !getInstance ( ).registeredFunctions.insert ( std::make_pair ( std::make_pair ( std::type_index ( typeid ( RealFirstParameterType ) ), std::type_index ( typeid ( RealSecondParameterType ) ) ), this ) ).second ) {
				std::string firstType = std::cstringToString ( std::type_name < RealFirstParameterType > ( ) );

				std::string secondType = std::cstringToString ( std::type_name < RealSecondParameterType > ( ) );

				std::string classType = std::cstringToString ( std::type_name < Algorithm > ( ) );

				throw::exception::CommonException ( "Callback for (" + firstType + ", " + secondType + ") already registered on " + classType + "." );
			}
		}

	};

	static ReturnType dispatch ( const FirstParameterType & first, const SecondParameterType & second ) {
		typename std::map < std::pair < std::type_index, std::type_index >, RegistratorWrapperBase * >::iterator callback = getInstance ( ).registeredFunctions.find ( std::make_pair ( std::type_index ( typeid ( first ) ), std::type_index ( typeid ( second ) ) ) );

		if ( callback == getInstance ( ).registeredFunctions.end ( ) ) {
			std::string firstType = std::cstringToString ( std::type_name ( typeid ( first ) ) );

			std::string secondType = std::cstringToString ( std::type_name ( typeid ( second ) ) );

			std::string classType = std::cstringToString ( std::type_name < Algorithm > ( ) );

			throw::exception::CommonException ( "Callback for (" + firstType + ", " + secondType + ") not registered on " + classType + "." );
		}

		return callback->second->eval ( first, second );
	}

};

template < class Algorithm, class ReturnType, class StaticParamType, class FirstParameterType, class SecondParameterType >
class DoubleDispatchFirstStaticParam {
public:
	class RegistratorWrapperBase {
	public:
		virtual ReturnType eval ( StaticParamType, const FirstParameterType &, const SecondParameterType & ) = 0;
	};

private:
	std::map < std::pair < std::type_index, std::type_index >, RegistratorWrapperBase * > registeredFunctions;

	static DoubleDispatchFirstStaticParam < Algorithm, ReturnType, StaticParamType, FirstParameterType, SecondParameterType > & getInstance ( ) {
		static DoubleDispatchFirstStaticParam < Algorithm, ReturnType, StaticParamType, FirstParameterType, SecondParameterType > res;

		return res;
	}

public:
	template < class RealReturnType, class RealFirstParameterType, class RealSecondParameterType >
	class RegistratorWrapper : public RegistratorWrapperBase {
		std::function < RealReturnType ( StaticParamType, const RealFirstParameterType &, const RealSecondParameterType & ) > callback;

	public:
		ReturnType eval ( StaticParamType res, const FirstParameterType & first, const SecondParameterType & second ) {
			return ReturnType ( callback ( res, ( const RealFirstParameterType & ) first, ( const SecondParameterType & ) second ) );
		}

		RegistratorWrapper ( RealReturnType ( * callback ) ( StaticParamType, const RealFirstParameterType &, const RealSecondParameterType & ) ) : callback ( callback ) {
			if ( !getInstance ( ).registeredFunctions.insert ( std::make_pair ( std::make_pair ( std::type_index ( typeid ( RealFirstParameterType ) ), std::type_index ( typeid ( RealSecondParameterType ) ) ), this ) ).second ) {
				std::string firstType = std::cstringToString ( std::type_name < RealFirstParameterType > ( ) );

				std::string secondType = std::cstringToString ( std::type_name < RealSecondParameterType > ( ) );

				std::string classType = std::cstringToString ( std::type_name < Algorithm > ( ) );

				throw::exception::CommonException ( "Callback for (" + firstType + ", " + secondType + ") already registered on " + classType + "." );
			}
		}

	};

	static ReturnType dispatch ( StaticParamType res, const FirstParameterType & first, const SecondParameterType & second ) {
		typename std::map < std::pair < std::type_index, std::type_index >, RegistratorWrapperBase * >::iterator callback = getInstance ( ).registeredFunctions.find ( std::make_pair ( std::type_index ( typeid ( first ) ), std::type_index ( typeid ( first ) ) ) );

		if ( callback == getInstance ( ).registeredFunctions.end ( ) ) {
			std::string firstType = std::cstringToString ( std::type_name ( typeid ( first ) ) );

			std::string secondType = std::cstringToString ( std::type_name ( typeid ( second ) ) );

			std::string classType = std::cstringToString ( std::type_name < Algorithm > ( ) );

			throw::exception::CommonException ( "Callback for (" + firstType + ", " + secondType + ") not registered on " + classType + "." );
		}

		return callback->second->eval ( res, first, second );
	}

};

} /* namespace std */

#endif /* MULTIPLE_DISPATCH_H_ */
