/*
 * CompactNFA.cpp
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#include "CompactNFA.h"
#include "EpsilonNFA.h"
#include "MultiInitialStateNFA.h"
#include "NFA.h"
#include "DFA.h"
#include "../AutomatonException.h"
#include <ostream>
#include <algorithm>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

namespace automaton {

CompactNFA::CompactNFA ( std::set < State > states, std::set < alphabet::Symbol > inputAlphabet, State initialState, std::set < State > finalStates ) : std::Components2 < CompactNFA, alphabet::Symbol, std::tuple < InputAlphabet >, std::tuple < >, automaton::State, std::tuple < States, FinalStates >, std::tuple < InitialState > > ( std::make_tuple ( std::move ( inputAlphabet ) ), std::tuple < > ( ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::make_tuple ( std::move ( initialState ) ) ) {
}

CompactNFA::CompactNFA ( State initialState ) : CompactNFA ( std::set < State > { initialState }, std::set < alphabet::Symbol > { }, initialState, std::set < State > { } ) {
}

CompactNFA::CompactNFA ( const EpsilonNFA & other ) : CompactNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		if ( transition.first.second.is < string::Epsilon > ( ) ) {
			std::pair < State, string::LinearString > key = std::make_pair ( transition.first.first, string::LinearString ( std::vector < alphabet::Symbol > { } ) );
			transitions[key] = transition.second;
		} else {
			std::pair < State, string::LinearString > key = std::make_pair ( transition.first.first, string::LinearString ( std::vector < alphabet::Symbol > { transition.first.second.get < alphabet::Symbol > ( ) } ) );
			transitions[key] = transition.second;
		}
	}
}

CompactNFA::CompactNFA ( const MultiInitialStateNFA & other ) : CompactNFA ( other.getStates ( ) + std::set < State > { automaton::createUniqueState ( automaton::State ( "q0" ), other.getStates ( ) ) }, other.getInputAlphabet ( ), automaton::createUniqueState ( automaton::State ( "q0" ), other.getStates ( ) ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		std::pair < State, string::LinearString > key = std::make_pair ( transition.first.first, string::LinearString ( std::vector < alphabet::Symbol > { transition.first.second } ) );
		transitions[key] = transition.second;
	}

	std::pair < State, string::LinearString > key = std::make_pair ( this->getInitialState ( ), string::LinearString ( std::vector < alphabet::Symbol > { } ) );
	transitions[key] = other.getInitialStates ( );
}

CompactNFA::CompactNFA ( const NFA & other ) : CompactNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		std::pair < State, string::LinearString > key = std::make_pair ( transition.first.first, string::LinearString ( std::vector < alphabet::Symbol > { transition.first.second } ) );
		transitions[key] = transition.second;
	}
}

CompactNFA::CompactNFA ( const DFA & other ) : CompactNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		std::pair < State, string::LinearString > key = std::make_pair ( transition.first.first, string::LinearString ( std::vector < alphabet::Symbol > { transition.first.second } ) );
		transitions[key].insert ( transition.second );
	}
}

AutomatonBase * CompactNFA::clone ( ) const {
	return new CompactNFA ( * this );
}

AutomatonBase * CompactNFA::plunder ( ) && {
	return new CompactNFA ( std::move ( * this ) );
}

bool CompactNFA::addTransition ( State from, string::LinearString input, State to ) {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ( std::string ) from.getName ( ) + "\" doesn't exist." );

	const std::set < alphabet::Symbol > & inputStringAlphabet = input.getAlphabet ( );

	 // Transition regexp's alphabet must be subset of automaton's alphabet
	if ( !std::includes ( getInputAlphabet ( ).begin ( ), getInputAlphabet ( ).end ( ), inputStringAlphabet.begin ( ), inputStringAlphabet.end ( ) ) )
		throw AutomatonException ( "Input string is over different alphabet than automaton" );

	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ( std::string ) to.getName ( ) + "\" doesn't exist." );

	std::pair < State, string::LinearString > key = std::make_pair ( std::move ( from ), std::move ( input ) );

	return transitions[std::move ( key )].insert ( std::move ( to ) ).second;
}

bool CompactNFA::removeTransition ( const State & from, const string::LinearString & input, const State & to ) {
	std::pair < State, string::LinearString > key = std::make_pair ( from, input );

	return transitions[key].erase ( to );
}

const std::map < std::pair < State, string::LinearString >, std::set < State > > & CompactNFA::getTransitions ( ) const {
	return transitions;
}

std::map < std::pair < State, string::LinearString >, std::set < State > > CompactNFA::getTransitionsFromState ( const State & from ) const {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ( std::string ) from.getName ( ) + "\" doesn't exist" );

	std::map < std::pair < State, string::LinearString >, std::set < State > > transitionsFromState;

	for ( const std::pair < const std::pair < State, string::LinearString >, std::set < State > > & transition : transitions )
		if ( transition.first.first == from )
			transitionsFromState.insert ( make_pair ( transition.first, transition.second ) );

	return transitionsFromState;
}

std::map < std::pair < State, string::LinearString >, std::set < State > > CompactNFA::getTransitionsToState ( const State & to ) const {
	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ( std::string ) to.getName ( ) + "\" doesn't exist" );

	std::map < std::pair < State, string::LinearString >, std::set < State > > transitionsToState;

	for ( const std::pair < const std::pair < State, string::LinearString >, std::set < State > > & transition : transitions )
		if ( transition.second.find ( to ) != transition.second.end ( ) )
			transitionsToState.insert ( make_pair ( transition.first, transition.second ) );

	return transitionsToState;
}

int CompactNFA::compare ( const CompactNFA & other ) const {
	auto first = std::tie ( getStates ( ), getInputAlphabet ( ), getInitialState ( ), getFinalStates ( ), transitions );
	auto second = std::tie ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ), other.getTransitions ( ) );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void CompactNFA::operator >>( std::ostream & out ) const {
	out << "(CompactNFA "
	    << "states = " << getStates ( )
	    << "inputAlphabet = " << getInputAlphabet ( )
	    << "initialState = " << getInitialState ( )
	    << "finalStates = " << getFinalStates ( )
	    << "transitions = " << transitions
	    << ")";
}

CompactNFA::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

CompactNFA CompactNFA::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, CompactNFA::getXmlTagName() );

	std::set < State > states = AutomatonFromXMLParser::parseStates ( input );
	std::set < alphabet::Symbol > inputSymbols = AutomatonFromXMLParser::parseInputAlphabet ( input );
	State initialState = AutomatonFromXMLParser::parseInitialState ( input );
	std::set < State > finalStates = AutomatonFromXMLParser::parseFinalStates ( input );

	CompactNFA automaton ( std::move ( initialState ) );

	automaton.setStates ( std::move ( states ) );
	automaton.setInputAlphabet ( std::move ( inputSymbols ) );
	automaton.setFinalStates ( std::move ( finalStates ) );

	AutomatonFromXMLParser::parseTransitions < CompactNFA > ( input, automaton );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, CompactNFA::getXmlTagName() );
	return automaton;
}

void CompactNFA::parseTransition ( std::deque < sax::Token >::iterator & input, CompactNFA & automaton ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "transition" );
	State from = AutomatonFromXMLParser::parseTransitionFrom ( input );
	string::LinearString inputString = AutomatonFromXMLParser::parseTransitionInputString ( input );
	State to = AutomatonFromXMLParser::parseTransitionTo ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "transition" );

	automaton.addTransition ( std::move ( from ), std::move ( inputString ), std::move ( to ) );
}

void CompactNFA::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( CompactNFA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	AutomatonToXMLComposer::composeStates ( out, this->getStates ( ) );
	AutomatonToXMLComposer::composeInputAlphabet ( out, this->getInputAlphabet ( ) );
	AutomatonToXMLComposer::composeInitialState ( out, this->getInitialState ( ) );
	AutomatonToXMLComposer::composeFinalStates ( out, this->getFinalStates ( ) );
	composeTransitions ( out );

	out.emplace_back ( CompactNFA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

void CompactNFA::composeTransitions ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( "transitions", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & transition : this->getTransitions ( ) )
		for ( const auto & targetState : transition.second ) {
			out.emplace_back ( "transition", sax::Token::TokenType::START_ELEMENT );

			AutomatonToXMLComposer::composeTransitionFrom ( out, transition.first.first );
			AutomatonToXMLComposer::composeTransitionInputString ( out, transition.first.second );
			AutomatonToXMLComposer::composeTransitionTo ( out, targetState );

			out.emplace_back ( "transition", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "transitions", sax::Token::TokenType::END_ELEMENT );
}

} /* namespace automaton */

namespace std {

template < >
class ComponentConstraint2< automaton::CompactNFA, alphabet::Symbol, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::CompactNFA & automaton, const alphabet::Symbol & symbol ) {
		for ( const std::pair < const std::pair < automaton::State, ::string::LinearString >, std::set < automaton::State > > & transition : automaton.getTransitions ( ) )
			if ( transition.first.second.getAlphabet ( ).count ( symbol ) )
				return true;

		return false;
	}

	static bool available ( const automaton::CompactNFA &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const automaton::CompactNFA &, const alphabet::Symbol & ) {
	}
};

template < >
class ComponentConstraint2< automaton::CompactNFA, automaton::State, automaton::States > {
public:
	static bool used ( const automaton::CompactNFA & automaton, const automaton::State & state ) {
		if ( automaton.getInitialState ( ) == state )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for ( const std::pair < const std::pair < automaton::State, ::string::LinearString >, std::set < automaton::State > > & t : automaton.getTransitions ( ) )
			if ( ( t.first.first == state ) || t.second.count ( state ) )
				return true;

		return false;
	}

	static bool available ( const automaton::CompactNFA &, const automaton::State & ) {
		return true;
	}

	static void valid ( const automaton::CompactNFA &, const automaton::State & ) {
	}
};

template < >
class ComponentConstraint2< automaton::CompactNFA, automaton::State, automaton::FinalStates > {
public:
	static bool used ( const automaton::CompactNFA &, const automaton::State & ) {
		return false;
	}

	static bool available ( const automaton::CompactNFA & automaton, const automaton::State & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::CompactNFA &, const automaton::State & ) {
	}
};

template < >
class ElementConstraint2< automaton::CompactNFA, automaton::State, automaton::InitialState > {
public:
	static bool available ( const automaton::CompactNFA & automaton, const automaton::State & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::CompactNFA &, const automaton::State & ) {
	}
};

} /* namespace std */

namespace alib {

auto compactNFAParserRegister = xmlApi < automaton::Automaton >::ParserRegister < automaton::CompactNFA > ( );
auto compactNFAParserRegister2 = xmlApi < alib::Object >::ParserRegister < automaton::CompactNFA > ( );

auto CompactNFAFromDFA = castApi::CastRegister < automaton::CompactNFA, automaton::DFA > ( );
auto CompactNFAFromNFA = castApi::CastRegister < automaton::CompactNFA, automaton::NFA > ( );
auto CompactNFAFromMultiInitialStateNFA = castApi::CastRegister < automaton::CompactNFA, automaton::MultiInitialStateNFA > ( );
auto CompactNFAEpsilonNFA = castApi::CastRegister < automaton::CompactNFA, automaton::EpsilonNFA > ( );
auto CompactNFACastBinder = castApi::CastPoolStringBinder < automaton::CompactNFA > ( automaton::CompactNFA::getXmlTagName() );

} /* namespace alib */
