/*
 * EpsilonNFA.cpp
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#include "EpsilonNFA.h"
#include "MultiInitialStateNFA.h"
#include "NFA.h"
#include "DFA.h"
#include "../AutomatonException.h"
#include "../Automaton.h"
#include <ostream>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

namespace automaton {

EpsilonNFA::EpsilonNFA ( std::set < State > states, std::set < alphabet::Symbol > inputAlphabet, State initialState, std::set < State > finalStates ) : std::Components2 < EpsilonNFA, alphabet::Symbol, std::tuple < InputAlphabet >, std::tuple < >, automaton::State, std::tuple < States, FinalStates >, std::tuple < InitialState > > ( std::make_tuple ( std::move ( inputAlphabet ) ), std::tuple < > ( ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::make_tuple ( std::move ( initialState ) ) ) {
}

EpsilonNFA::EpsilonNFA ( State initialState ) : EpsilonNFA ( std::set < State > { initialState }, std::set < alphabet::Symbol > { }, initialState, std::set < State > { } ) {
}

EpsilonNFA::EpsilonNFA ( const MultiInitialStateNFA & other ) : EpsilonNFA ( other.getStates ( ) + std::set < State > { automaton::createUniqueState ( automaton::State ( "q0" ), other.getStates ( ) ) }, other.getInputAlphabet ( ), automaton::createUniqueState ( automaton::State ( "q0" ), other.getStates ( ) ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > > key = std::make_pair ( transition.first.first, std::variant < string::Epsilon, alphabet::Symbol > ( transition.first.second ) );
		transitions[key] = transition.second;
	}

	std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > > key = std::make_pair ( this->getInitialState ( ), std::variant < string::Epsilon, alphabet::Symbol > ( string::Epsilon::EPSILON ) );
	transitions[key] = other.getInitialStates ( );
}

EpsilonNFA::EpsilonNFA ( const NFA & other ) : EpsilonNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > > key = std::make_pair ( transition.first.first, std::variant < string::Epsilon, alphabet::Symbol > ( transition.first.second ) );
		transitions[key] = transition.second;
	}
}

EpsilonNFA::EpsilonNFA ( const DFA & other ) : EpsilonNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > > key = std::make_pair ( transition.first.first, std::variant < string::Epsilon, alphabet::Symbol > ( transition.first.second ) );
		transitions[key].insert ( transition.second );
	}
}

AutomatonBase * EpsilonNFA::clone ( ) const {
	return new EpsilonNFA ( * this );
}

AutomatonBase * EpsilonNFA::plunder ( ) && {
	return new EpsilonNFA ( std::move ( * this ) );
}

bool EpsilonNFA::addTransition ( State from, std::variant < string::Epsilon, alphabet::Symbol > input, State to ) {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ( std::string ) from.getName ( ) + "\" doesn't exist." );

	if ( input.is < alphabet::Symbol > ( ) && !getInputAlphabet ( ).count ( input.get < alphabet::Symbol > ( ) ) )
		throw AutomatonException ( "Input symbol \"" + ( std::string ) input.get < alphabet::Symbol > ( ) + "\" doesn't exist." );

	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ( std::string ) to.getName ( ) + "\" doesn't exist." );

	std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > > key = std::make_pair ( std::move ( from ), std::move ( input ) );

	return transitions[std::move ( key )].insert ( std::move ( to ) ).second;
}

bool EpsilonNFA::addTransition ( State from, alphabet::Symbol input, State to ) {
	std::variant < string::Epsilon, alphabet::Symbol > inputVariant ( std::move ( input ) );

	return addTransition ( std::move ( from ), std::move ( inputVariant ), std::move ( to ) );
}

bool EpsilonNFA::addTransition ( State from, State to ) {
	std::variant < string::Epsilon, alphabet::Symbol > inputVariant ( string::Epsilon::EPSILON );

	return addTransition ( std::move ( from ), std::move ( inputVariant ), std::move ( to ) );
}

bool EpsilonNFA::removeTransition ( const State & from, const alphabet::Symbol & input, const State & to ) {
	std::variant < string::Epsilon, alphabet::Symbol > inputVariant ( input );
	std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > > key = std::make_pair ( from, inputVariant );

	return transitions[key].erase ( to );
}

bool EpsilonNFA::removeTransition ( const State & from, const State & to ) {
	std::variant < string::Epsilon, alphabet::Symbol > inputVariant ( string::Epsilon::EPSILON );
	std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > > key = std::make_pair ( from, inputVariant );

	return transitions[key].erase ( to );
}

const std::map < std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > >, std::set < State > > & EpsilonNFA::getTransitions ( ) const {
	return transitions;
}

std::map < State, std::set < State > > EpsilonNFA::getEpsilonTransitions ( ) const {
	std::map < State, std::set < State > > result;

	for ( const std::pair < const std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > >, std::set < State > > & transition : transitions )
		if ( transition.first.second.is < string::Epsilon > ( ) )
			result[transition.first.first].insert ( transition.second.begin ( ), transition.second.end ( ) );

	return result;
}

std::map < std::pair < State, alphabet::Symbol >, std::set < State > > EpsilonNFA::getSymbolTransitions ( ) const {
	std::map < std::pair < State, alphabet::Symbol >, std::set < State > > result;

	for ( const std::pair < const std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > >, std::set < State > > & transition : transitions )
		if ( transition.first.second.is < alphabet::Symbol > ( ) )
			result[std::pair < State, alphabet::Symbol > ( transition.first.first, transition.first.second.get < alphabet::Symbol > ( ) )].insert ( transition.second.begin ( ), transition.second.end ( ) );

	return result;
}

std::map < std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > >, std::set < State > > EpsilonNFA::getTransitionsFromState ( const State & from ) const {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ( std::string ) from.getName ( ) + "\" doesn't exist" );

	std::map < std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > >, std::set < State > > transitionsFromState;

	for ( const std::pair < const std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > >, std::set < State > > & transition : transitions )
		if ( transition.first.first == from )
			transitionsFromState[transition.first].insert ( transition.second.begin ( ), transition.second.end ( ) );

	return transitionsFromState;
}

std::map < State, std::set < State > > EpsilonNFA::getEpsilonTransitionsFromState ( const State & from ) const {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ( std::string ) from.getName ( ) + "\" doesn't exist" );

	std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > > key ( from, std::variant < string::Epsilon, alphabet::Symbol > ( string::Epsilon::EPSILON ) );
	std::map < State, std::set < State > > res;
	auto transition = transitions.find ( key );

	if ( transition != transitions.end ( ) )
		res.insert ( std::make_pair ( from, transition->second ) );

	return res;
}

std::map < std::pair < State, alphabet::Symbol >, std::set < State > > EpsilonNFA::getSymbolTransitionsFromState ( const State & from ) const {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ( std::string ) from.getName ( ) + "\" doesn't exist" );

	std::map < std::pair < State, alphabet::Symbol >, std::set < State > > transitionsFromState;

	for ( const std::pair < const std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > >, std::set < State > > & transition : transitions )
		if ( ( transition.first.first == from ) && transition.first.second.is < alphabet::Symbol > ( ) )
			transitionsFromState[std::pair < State, alphabet::Symbol > ( transition.first.first, transition.first.second.get < alphabet::Symbol > ( ) )].insert ( transition.second.begin ( ), transition.second.end ( ) );

	return transitionsFromState;
}

std::map < std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > >, std::set < State > > EpsilonNFA::getTransitionsToState ( const State & to ) const {
	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ( std::string ) to.getName ( ) + "\" doesn't exist" );

	std::map < std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > >, std::set < State > > transitionsToState;

	for ( const std::pair < const std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > >, std::set < State > > & transition : transitions )
		if ( transition.second.find ( to ) != transition.second.end ( ) )
			transitionsToState[transition.first].insert ( transition.second.begin ( ), transition.second.end ( ) );

	return transitionsToState;
}

std::map < State, std::set < State > > EpsilonNFA::getEpsilonTransitionsToState ( const State & to ) const {
	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ( std::string ) to.getName ( ) + "\" doesn't exist" );

	std::map < State, std::set < State > > transitionsToState;

	for ( const std::pair < const std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > >, std::set < State > > & transition : transitions )
		if ( ( transition.second.find ( to ) != transition.second.end ( ) ) && transition.first.second.is < string::Epsilon > ( ) )
			transitionsToState[transition.first.first].insert ( transition.second.begin ( ), transition.second.end ( ) );

	return transitionsToState;
}

std::map < std::pair < State, alphabet::Symbol >, std::set < State > > EpsilonNFA::getSymbolTransitionsToState ( const State & to ) const {
	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ( std::string ) to.getName ( ) + "\" doesn't exist" );

	std::map < std::pair < State, alphabet::Symbol >, std::set < State > > transitionsToState;

	for ( const std::pair < const std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > >, std::set < State > > & transition : transitions )
		if ( ( transition.second.find ( to ) != transition.second.end ( ) ) && transition.first.second.is < alphabet::Symbol > ( ) )
			transitionsToState[std::pair < State, alphabet::Symbol > ( transition.first.first, transition.first.second.get < alphabet::Symbol > ( ) )].insert ( transition.second.begin ( ), transition.second.end ( ) );

	return transitionsToState;
}

bool EpsilonNFA::isEpsilonFree ( ) const {
	for ( const std::pair < const std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > >, std::set < State > > & transition : transitions )
		if ( transition.first.second.is < string::Epsilon > ( ) )
			return false;

	return true;
}

bool EpsilonNFA::isDeterministic ( ) const {
	for ( const std::pair < const std::pair < State, std::variant < string::Epsilon, alphabet::Symbol > >, std::set < State > > & transition : transitions )
		if ( transition.second.size ( ) > 1 )
			return false;

	return isEpsilonFree ( );
}

bool EpsilonNFA::isTotal ( ) const {
	return isDeterministic ( ) && transitions.size ( ) == getInputAlphabet ( ).size ( ) * getStates ( ).size ( );
}

int EpsilonNFA::compare ( const EpsilonNFA & other ) const {
	auto first = std::tie ( getStates ( ), getInputAlphabet ( ), getInitialState ( ), getFinalStates ( ), transitions );
	auto second = std::tie ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ), other.getTransitions ( ) );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void EpsilonNFA::operator >>( std::ostream & out ) const {
	out << "(EpsilonNFA "
	    << "states = " << getStates ( )
	    << "inputAlphabet = " << getInputAlphabet ( )
	    << "initialState = " << getInitialState ( )
	    << "finalStates = " << getFinalStates ( )
	    << "transitions = " << transitions
	    << ")";
}

EpsilonNFA::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

EpsilonNFA EpsilonNFA::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, EpsilonNFA::getXmlTagName() );

	std::set < State > states = AutomatonFromXMLParser::parseStates ( input );
	std::set < alphabet::Symbol > inputSymbols = AutomatonFromXMLParser::parseInputAlphabet ( input );
	State initialState = AutomatonFromXMLParser::parseInitialState ( input );
	std::set < State > finalStates = AutomatonFromXMLParser::parseFinalStates ( input );

	EpsilonNFA automaton ( std::move ( initialState ) );

	automaton.setStates ( std::move ( states ) );
	automaton.setInputAlphabet ( std::move ( inputSymbols ) );
	automaton.setFinalStates ( std::move ( finalStates ) );

	AutomatonFromXMLParser::parseTransitions < EpsilonNFA > ( input, automaton );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, EpsilonNFA::getXmlTagName() );
	return automaton;
}

void EpsilonNFA::parseTransition ( std::deque < sax::Token >::iterator & input, EpsilonNFA & automaton ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "transition" );
	State from = AutomatonFromXMLParser::parseTransitionFrom ( input );
	std::variant < string::Epsilon, alphabet::Symbol > inputVariant = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol ( input );
	State to = AutomatonFromXMLParser::parseTransitionTo ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "transition" );

	automaton.addTransition ( std::move ( from ), std::move ( inputVariant ), std::move ( to ) );
}

void EpsilonNFA::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( EpsilonNFA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	AutomatonToXMLComposer::composeStates ( out, this->getStates ( ) );
	AutomatonToXMLComposer::composeInputAlphabet ( out, this->getInputAlphabet ( ) );
	AutomatonToXMLComposer::composeInitialState ( out, this->getInitialState ( ) );
	AutomatonToXMLComposer::composeFinalStates ( out, this->getFinalStates ( ) );
	composeTransitions ( out );

	out.emplace_back ( EpsilonNFA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

void EpsilonNFA::composeTransitions ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( "transitions", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & transition : this->getTransitions ( ) )
		for ( const auto & targetState : transition.second ) {
			out.emplace_back ( "transition", sax::Token::TokenType::START_ELEMENT );

			AutomatonToXMLComposer::composeTransitionFrom ( out, transition.first.first );
			AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol ( out, transition.first.second );
			AutomatonToXMLComposer::composeTransitionTo ( out, targetState );

			out.emplace_back ( "transition", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "transitions", sax::Token::TokenType::END_ELEMENT );
}

} /* namespace automaton */

namespace std {

template < >
class ComponentConstraint2< automaton::EpsilonNFA, alphabet::Symbol, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::EpsilonNFA & automaton, const alphabet::Symbol & symbol ) {
		for ( const std::pair < const std::pair < automaton::State, std::variant < ::string::Epsilon, alphabet::Symbol > >, std::set < automaton::State > > & transition : automaton.getTransitions ( ) )
			if ( transition.first.second.is < alphabet::Symbol > ( ) && ( transition.first.second.get < alphabet::Symbol > ( ) == symbol ) )
				return true;

		return false;
	}

	static bool available ( const automaton::EpsilonNFA &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const automaton::EpsilonNFA &, const alphabet::Symbol & ) {
	}
};

template < >
class ComponentConstraint2< automaton::EpsilonNFA, automaton::State, automaton::States > {
public:
	static bool used ( const automaton::EpsilonNFA & automaton, const automaton::State & state ) {
		if ( automaton.getInitialState ( ) == state )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for ( const std::pair < const std::pair < automaton::State, std::variant < ::string::Epsilon, alphabet::Symbol > >, std::set < automaton::State > > & transition : automaton.getTransitions ( ) )
			if ( ( transition.first.first == state ) || transition.second.count ( state ) )
				return true;

		return false;
	}

	static bool available ( const automaton::EpsilonNFA &, const automaton::State & ) {
		return true;
	}

	static void valid ( const automaton::EpsilonNFA &, const automaton::State & ) {
	}
};

template < >
class ComponentConstraint2< automaton::EpsilonNFA, automaton::State, automaton::FinalStates > {
public:
	static bool used ( const automaton::EpsilonNFA &, const automaton::State & ) {
		return false;
	}

	static bool available ( const automaton::EpsilonNFA & automaton, const automaton::State & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::EpsilonNFA &, const automaton::State & ) {
	}
};

template < >
class ElementConstraint2< automaton::EpsilonNFA, automaton::State, automaton::InitialState > {
public:
	static bool available ( const automaton::EpsilonNFA & automaton, const automaton::State & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::EpsilonNFA &, const automaton::State & ) {
	}
};

} /* namespace std */

namespace alib {

auto epsilonNFAParserRegister = xmlApi < automaton::Automaton >::ParserRegister < automaton::EpsilonNFA > ( );
auto epsilonNFAParserRegister2 = xmlApi < alib::Object >::ParserRegister < automaton::EpsilonNFA > ( );

auto EpsilonNFAFromDFA = castApi::CastRegister < automaton::EpsilonNFA, automaton::DFA > ( );
auto EpsilonNFAFromNFA = castApi::CastRegister < automaton::EpsilonNFA, automaton::NFA > ( );
auto EpsilonNFAFromMultiInitialStateNFA = castApi::CastRegister < automaton::EpsilonNFA, automaton::MultiInitialStateNFA > ( );
auto EpsilonNFACastBinder = castApi::CastPoolStringBinder < automaton::EpsilonNFA > ( automaton::EpsilonNFA::getXmlTagName() );

} /* namespace alib */
