/*
 * NFA.cpp
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#include "NFA.h"
#include "../AutomatonException.h"
#include <ostream>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

namespace automaton {

NFA::NFA ( std::set < State > states, std::set < alphabet::Symbol > inputAlphabet, State initialState, std::set < State > finalStates ) : std::Components2 < NFA, alphabet::Symbol, std::tuple < InputAlphabet >, std::tuple < >, automaton::State, std::tuple < States, FinalStates >, std::tuple < InitialState > > ( std::make_tuple ( std::move ( inputAlphabet ) ), std::tuple < > ( ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::make_tuple ( std::move ( initialState ) ) ) {
}

NFA::NFA ( State initialState ) : NFA ( std::set < State > { initialState }, std::set < alphabet::Symbol > { }, initialState, std::set < State > { } ) {
}

NFA::NFA ( const DFA & other ) : NFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) )
		transitions[transition.first].insert ( transition.second );
}

AutomatonBase * NFA::clone ( ) const {
	return new NFA ( * this );
}

AutomatonBase * NFA::plunder ( ) && {
	return new NFA ( std::move ( * this ) );
}

bool NFA::addTransition ( State from, alphabet::Symbol input, State to ) {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ( std::string ) from.getName ( ) + "\" doesn't exist." );

	if ( !getInputAlphabet ( ).count ( input ) )
		throw AutomatonException ( "Input symbol \"" + ( std::string ) input + "\" doesn't exist." );

	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ( std::string ) to.getName ( ) + "\" doesn't exist." );

	std::pair < State, alphabet::Symbol > key = std::make_pair ( std::move ( from ), std::move ( input ) );

	return transitions[std::move ( key )].insert ( std::move ( to ) ).second;
}

bool NFA::removeTransition ( const State & from, const alphabet::Symbol & input, const State & to ) {
	std::pair < State, alphabet::Symbol > key = std::make_pair ( from, input );

	return transitions[key].erase ( to );
}

const std::map < std::pair < State, alphabet::Symbol >, std::set < State > > & NFA::getTransitions ( ) const {
	return transitions;
}

std::map < std::pair < State, alphabet::Symbol >, std::set < State > > NFA::getTransitionsFromState ( const State & from ) const {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ( std::string ) from.getName ( ) + "\" doesn't exist" );

	std::map < std::pair < State, alphabet::Symbol >, std::set < State > > transitionsFromState;

	for ( const std::pair < const std::pair < State, alphabet::Symbol >, std::set < State > > & transition : transitions )
		if ( transition.first.first == from )
			transitionsFromState[transition.first].insert ( transition.second.begin ( ), transition.second.end ( ) );

	return transitionsFromState;
}

std::map < std::pair < State, alphabet::Symbol >, std::set < State > > NFA::getTransitionsToState ( const State & to ) const {
	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ( std::string ) to.getName ( ) + "\" doesn't exist" );

	std::map < std::pair < State, alphabet::Symbol >, std::set < State > > transitionsToState;

	for ( const std::pair < const std::pair < State, alphabet::Symbol >, std::set < State > > & transition : transitions )
		if ( transition.second.find ( to ) != transition.second.end ( ) )
			transitionsToState[transition.first].insert ( transition.second.begin ( ), transition.second.end ( ) );

	return transitionsToState;
}

bool NFA::isDeterministic ( ) const {
	for ( const std::pair < const std::pair < State, alphabet::Symbol >, std::set < State > > & transition : transitions )
		if ( transition.second.size ( ) > 1 )
			return false;

	return true;
}

bool NFA::isTotal ( ) const {
	return isDeterministic ( ) && transitionsSize ( ) == getInputAlphabet ( ).size ( ) * getStates ( ).size ( );
}

unsigned NFA::transitionsSize ( ) const {
	int res = 0;

	for ( const auto & transition : transitions )
		res += transition.second.size ( );

	return res;
}

int NFA::compare ( const NFA & other ) const {
	auto first = std::tie ( getStates ( ), getInputAlphabet ( ), getInitialState ( ), getFinalStates ( ), transitions );
	auto second = std::tie ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ), other.getTransitions ( ) );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void NFA::operator >>( std::ostream & out ) const {
	out << "(NFA "
	    << "states = " << getStates ( )
	    << "inputAlphabet = " << getInputAlphabet ( )
	    << "initialState = " << getInitialState ( )
	    << "finalStates = " << getFinalStates ( )
	    << "transitions = " << transitions
	    << ")";
}

NFA::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

NFA NFA::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, NFA::getXmlTagName() );

	std::set < State > states = AutomatonFromXMLParser::parseStates ( input );
	std::set < alphabet::Symbol > inputSymbols = AutomatonFromXMLParser::parseInputAlphabet ( input );
	State initialState = AutomatonFromXMLParser::parseInitialState ( input );
	std::set < State > finalStates = AutomatonFromXMLParser::parseFinalStates ( input );

	NFA automaton ( std::move ( initialState ) );

	automaton.setStates ( std::move ( states ) );
	automaton.setInputAlphabet ( std::move ( inputSymbols ) );
	automaton.setFinalStates ( std::move ( finalStates ) );

	AutomatonFromXMLParser::parseTransitions < NFA > ( input, automaton );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, NFA::getXmlTagName() );
	return automaton;
}

void NFA::parseTransition ( std::deque < sax::Token >::iterator & input, NFA & automaton ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "transition" );
	State from = AutomatonFromXMLParser::parseTransitionFrom ( input );
	alphabet::Symbol inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol ( input );
	State to = AutomatonFromXMLParser::parseTransitionTo ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "transition" );

	automaton.addTransition ( std::move ( from ), std::move ( inputSymbol ), std::move ( to ) );
}

void NFA::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( NFA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	AutomatonToXMLComposer::composeStates ( out, this->getStates ( ) );
	AutomatonToXMLComposer::composeInputAlphabet ( out, this->getInputAlphabet ( ) );
	AutomatonToXMLComposer::composeInitialState ( out, this->getInitialState ( ) );
	AutomatonToXMLComposer::composeFinalStates ( out, this->getFinalStates ( ) );
	composeTransitions ( out );

	out.emplace_back ( NFA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

void NFA::composeTransitions ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( "transitions", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & transition : this->getTransitions ( ) )
		for ( const auto & targetState : transition.second ) {
			out.emplace_back ( "transition", sax::Token::TokenType::START_ELEMENT );

			AutomatonToXMLComposer::composeTransitionFrom ( out, transition.first.first );
			AutomatonToXMLComposer::composeTransitionInputSymbol ( out, transition.first.second );
			AutomatonToXMLComposer::composeTransitionTo ( out, targetState );

			out.emplace_back ( "transition", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "transitions", sax::Token::TokenType::END_ELEMENT );
}

} /* namespace automaton */

namespace std {

template < >
class ComponentConstraint2< automaton::NFA, alphabet::Symbol, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::NFA & automaton, const alphabet::Symbol & symbol ) {
		for ( const std::pair < const std::pair < automaton::State, alphabet::Symbol >, std::set < automaton::State > > & transition : automaton.getTransitions ( ) )
			if ( transition.first.second == symbol )
				return true;

		return false;
	}

	static bool available ( const automaton::NFA &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const automaton::NFA &, const alphabet::Symbol & ) {
	}
};

template < >
class ComponentConstraint2< automaton::NFA, automaton::State, automaton::States > {
public:
	static bool used ( const automaton::NFA & automaton, const automaton::State & state ) {
		if ( automaton.getInitialState ( ) == state )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for ( const std::pair < const std::pair < automaton::State, alphabet::Symbol >, std::set < automaton::State > > & transition : automaton.getTransitions ( ) )
			if ( ( transition.first.first == state ) || ( transition.second.find ( state ) != transition.second.end ( ) ) )
				return true;

		return false;
	}

	static bool available ( const automaton::NFA &, const automaton::State & ) {
		return true;
	}

	static void valid ( const automaton::NFA &, const automaton::State & ) {
	}
};

template < >
class ComponentConstraint2< automaton::NFA, automaton::State, automaton::FinalStates > {
public:
	static bool used ( const automaton::NFA &, const automaton::State & ) {
		return false;
	}

	static bool available ( const automaton::NFA & automaton, const automaton::State & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::NFA &, const automaton::State & ) {
	}
};

template < >
class ElementConstraint2< automaton::NFA, automaton::State, automaton::InitialState > {
public:
	static bool available ( const automaton::NFA & automaton, const automaton::State & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::NFA &, const automaton::State & ) {
	}
};

} /* namespace std */

namespace alib {

auto NFAParserRegister = xmlApi < automaton::Automaton >::ParserRegister < automaton::NFA > ( );
auto NFAParserRegister2 = xmlApi < alib::Object >::ParserRegister < automaton::NFA > ( );

auto NFAFromDFA = castApi::CastRegister < automaton::NFA, automaton::DFA > ( );
auto NFACastBinder = castApi::CastPoolStringBinder < automaton::NFA > ( automaton::NFA::getXmlTagName() );

} /* namespace alib */
