/*
 * VisiblyPushdownDPDA.h
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#ifndef VISIBLY_PUSHDOWN_DPDA_H_
#define VISIBLY_PUSHDOWN_DPDA_H_

#include <map>
#include <vector>
#include <core/components2.hpp>
#include "../AutomatonBase.h"
#include "../common/State.h"
#include "../../alphabet/Symbol.h"

namespace automaton {

class CallAlphabet;
class ReturnAlphabet;
class LocalAlphabet;
class PushdownStoreAlphabet;
class BottomOfTheStackSymbol;
class States;
class FinalStates;
class InitialState;

/**
 * Represents Finite Automaton.
 * Can store nondeterministic finite automaton without epsilon transitions.
 */
class VisiblyPushdownDPDA : public AutomatonBase, public std::Components2 < VisiblyPushdownDPDA, alphabet::Symbol, std::tuple < CallAlphabet, ReturnAlphabet, LocalAlphabet, PushdownStoreAlphabet >, std::tuple < BottomOfTheStackSymbol >, automaton::State, std::tuple < States, FinalStates >, std::tuple < InitialState > > {
protected:
	std::map < std::pair < State, alphabet::Symbol >, std::pair < State, alphabet::Symbol > > callTransitions;
	std::map < std::tuple < State, alphabet::Symbol, alphabet::Symbol >, State > returnTransitions;
	std::map < std::pair < State, alphabet::Symbol >, State > localTransitions;

public:
	explicit VisiblyPushdownDPDA ( std::set < State > states, std::set < alphabet::Symbol > callAlphabet, std::set < alphabet::Symbol > returnAlphabet, std::set < alphabet::Symbol > localAlphabet, std::set < alphabet::Symbol > pushdownStoreSymbol, State initialState, alphabet::Symbol bottomOfTheStackSymbol, std::set < automaton::State > finalStates );
	explicit VisiblyPushdownDPDA ( State initialSymbol, alphabet::Symbol bottomOfTheStackSymbol );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const State & getInitialState ( ) const {
		return accessElement < InitialState > ( ).get ( );
	}

	bool setInitialState ( State state ) {
		return accessElement < InitialState > ( ).set ( std::move ( state ) );
	}

	const std::set < State > & getStates ( ) const {
		return accessComponent < States > ( ).get ( );
	}

	bool addState ( State state ) {
		return accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < State > states ) {
		accessComponent < States > ( ).set ( std::move ( states ) );
	}

	bool removeState ( const State & state ) {
		return accessComponent < States > ( ).remove ( state );
	}

	const std::set < State > & getFinalStates ( ) const {
		return accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( State state ) {
		return accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < State > states ) {
		accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	bool removeFinalState ( const State & state ) {
		return accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < alphabet::Symbol > & getPushdownStoreAlphabet ( ) const {
		return accessComponent < PushdownStoreAlphabet > ( ).get ( );
	}

	bool addPushdownStoreSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addPushdownStoreSymbols ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setPushdownStoreAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < PushdownStoreAlphabet > ( ).set ( std::move ( symbols ) );
	}

	bool removePushdownStoreSymbol ( const alphabet::Symbol & symbol ) {
		return accessComponent < PushdownStoreAlphabet > ( ).remove ( symbol );
	}

	const alphabet::Symbol & getBottomOfTheStackSymbol ( ) const {
		return accessElement < BottomOfTheStackSymbol > ( ).get ( );
	}

	bool setBottomOfTheStackSymbol ( alphabet::Symbol symbol ) {
		return accessElement < BottomOfTheStackSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < alphabet::Symbol > & getCallInputAlphabet ( ) const {
		return accessComponent < CallAlphabet > ( ).get ( );
	}

	bool addCallInputSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < CallAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addCallInputSymbols ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < CallAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setCallInputAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < CallAlphabet > ( ).set ( std::move ( symbols ) );
	}

	bool removeCallInputSymbol ( const alphabet::Symbol & symbol ) {
		return accessComponent < CallAlphabet > ( ).remove ( symbol );
	}

	const std::set < alphabet::Symbol > & getReturnInputAlphabet ( ) const {
		return accessComponent < ReturnAlphabet > ( ).get ( );
	}

	bool addReturnInputSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < ReturnAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addReturnInputSymbols ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < ReturnAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setReturnInputAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < ReturnAlphabet > ( ).set ( std::move ( symbols ) );
	}

	bool removeReturnInputSymbol ( const alphabet::Symbol & symbol ) {
		return accessComponent < ReturnAlphabet > ( ).remove ( symbol );
	}

	const std::set < alphabet::Symbol > & getLocalInputAlphabet ( ) const {
		return accessComponent < LocalAlphabet > ( ).get ( );
	}

	bool addLocalInputSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < LocalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addLocalInputSymbols ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < LocalAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setLocalInputAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < LocalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	bool removeLocalInputSymbol ( const alphabet::Symbol & symbol ) {
		return accessComponent < LocalAlphabet > ( ).remove ( symbol );
	}

	bool removeInputSymbol(const alphabet::Symbol& symbol) {
		return removeCallInputSymbol(symbol) || removeReturnInputSymbol(symbol) || removeLocalInputSymbol(symbol);
	}

	/**
	 * Adds call transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addCallTransition ( State current, alphabet::Symbol input, State next, alphabet::Symbol push );

	/**
	 * Adds return transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addReturnTransition ( State current, alphabet::Symbol input, alphabet::Symbol pop, State next );

	/**
	 * Adds local transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addLocalTransition ( State current, alphabet::Symbol input, State next );

	/**
	 * Removes call transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeCallTransition ( const State & current, const alphabet::Symbol & input, const State & next, const alphabet::Symbol & push );

	/**
	 * Removes return transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeReturnTransition ( const State & current, const alphabet::Symbol & input, const alphabet::Symbol & pop, const State & next );

	/**
	 * Removes transition local from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeLocalTransition ( const State & current, const alphabet::Symbol & input, const State & next );

	const std::map < std::pair < State, alphabet::Symbol >, std::pair < State, alphabet::Symbol > > & getCallTransitions ( ) const;

	const std::map < std::tuple < State, alphabet::Symbol, alphabet::Symbol >, State > & getReturnTransitions ( ) const;

	const std::map < std::pair < State, alphabet::Symbol >, State > & getLocalTransitions ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const VisiblyPushdownDPDA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "VisiblyPushdownDPDA";

		return xmlTagName;
	}

	static VisiblyPushdownDPDA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, VisiblyPushdownDPDA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;
};

} /* namespace automaton */

#endif /* VISIBLY_PUSHDOWN_DPDA_H_ */
