/*
 * DFTA.cpp
 *
 *  Created on: Apr 14, 2015
 *      Author: Stepan Plachy
 */

#include "DFTA.h"
#include "../AutomatonException.h"
#include <ostream>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include <object/Object.h>
#include <XmlApi.hpp>

namespace automaton {

DFTA::DFTA ( std::set < State > states, std::set < alphabet::RankedSymbol > inputAlphabet, std::set < State > finalStates ) : std::Components2 < DFTA, alphabet::RankedSymbol, std::tuple < InputAlphabet >, std::tuple < >, automaton::State, std::tuple < States, FinalStates >, std::tuple < > > ( std::make_tuple ( std::move ( inputAlphabet ) ), std::tuple < > ( ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::tuple < > ( ) ) {
}

DFTA::DFTA() : DFTA ( std::set < State > { }, std::set < alphabet::RankedSymbol > { }, std::set < State > { } ) {
}

AutomatonBase* DFTA::clone() const {
	return new DFTA(*this);
}

AutomatonBase* DFTA::plunder() && {
	return new DFTA(std::move(*this));
}

bool DFTA::addTransition(const alphabet::RankedSymbol & symbol, const std::vector<State> & prevStates, const State & next) {
	if ( prevStates.size() != symbol.getRank().getData())
		throw AutomatonException("Number of states doesn't match rank of the symbol");

	if (! getInputAlphabet().count(symbol))
		throw AutomatonException("Input symbol \"" + (std::string) symbol + "\" doesn't exist.");

	if (! getStates().count(next))
		throw AutomatonException("State \"" + (std::string) next.getName() + "\" doesn't exist.");

	for (const State& it : prevStates) {
		if (! getStates().count(it))
			throw AutomatonException("State \"" + (std::string) it . getName() + "\" doesn't exist.");
	}

	std::pair<alphabet::RankedSymbol, std::vector<State> > key = std::make_pair(symbol, prevStates);
	if ( transitions.find ( key ) != transitions.end ( ) ) {
		if ( transitions.find ( key )->second == next )
			return false;
		else
			throw AutomatonException("Transition already exists");
	}

	transitions.insert(std::make_pair(key, next));
	return true;
}

bool DFTA::removeTransition(const alphabet::RankedSymbol symbol, const std::vector<State> & states, const State & next) {
	std::pair<alphabet::RankedSymbol, std::vector<State> > key = std::make_pair(symbol, states);

	if ( transitions.find ( key ) == transitions.end ( ) )
		return false;

	if ( transitions.find ( key )->second != next )
		throw AutomatonException("Transition does not exist");

	transitions.erase(key);
	return true;
}

int DFTA::compare(const DFTA& other) const {
	auto first = std::tie(getStates(), getInputAlphabet(), getFinalStates(), transitions);
	auto second = std::tie(other.getStates(), other.getInputAlphabet(), other.getFinalStates(), other.transitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void DFTA::operator>>(std::ostream& out) const {
	out << "(DFTA "
		<< " states = " << getStates()
		<< " inputAlphabet = " << getInputAlphabet()
		<< " finalStates = " << getFinalStates()
		<< " transitions = " << transitions
		<< ")";
}

DFTA::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

DFTA DFTA::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, DFTA::getXmlTagName());

	std::set<State> states = AutomatonFromXMLParser::parseStates(input);
	std::set<alphabet::RankedSymbol> inputSymbols = AutomatonFromXMLParser::parseRankedInputAlphabet(input);
	std::set<State> finalStates = AutomatonFromXMLParser::parseFinalStates(input);

	DFTA automaton;
	automaton.setStates(std::move(states));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<DFTA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, DFTA::getXmlTagName());
	return automaton;
}

void DFTA::parseTransition(std::deque<sax::Token>::iterator& input, DFTA& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transition");
	alphabet::RankedSymbol inputSymbol = AutomatonFromXMLParser::parseTransitionInputRankedSymbol(input);
	std::vector<State> from = AutomatonFromXMLParser::parseTransitionFromMultiple(input);
	State to = AutomatonFromXMLParser::parseTransitionTo(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transition");

	automaton.addTransition(std::move(inputSymbol), std::move(from), std::move(to));
}

void DFTA::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(DFTA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeRankedInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(DFTA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

void DFTA::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back(sax::Token("transitions", sax::Token::TokenType::START_ELEMENT));
	for(const auto& transition : this->getTransitions()) {
		out.emplace_back(sax::Token("transition", sax::Token::TokenType::START_ELEMENT));

		AutomatonToXMLComposer::composeTransitionInputSymbol(out, transition.first.first);
		AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.second);
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second);

		out.emplace_back(sax::Token("transition", sax::Token::TokenType::END_ELEMENT));
	}

	out.emplace_back(sax::Token("transitions", sax::Token::TokenType::END_ELEMENT));
}

} /* namespace automaton */

namespace std {

template < >
class ComponentConstraint2< automaton::DFTA, alphabet::RankedSymbol, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::DFTA & automaton, const alphabet::RankedSymbol & symbol ) {
		for (const std::pair<const std::pair<alphabet::RankedSymbol, std::vector<automaton::State> >, automaton::State>& t : automaton.getTransitions())
			if (t.first.first == symbol)
				return true;

		return false;
	}

	static bool available ( const automaton::DFTA &, const alphabet::RankedSymbol & ) {
		return true;
	}

	static void valid ( const automaton::DFTA &, const alphabet::RankedSymbol & ) {
	}
};

template < >
class ComponentConstraint2< automaton::DFTA, automaton::State, automaton::States > {
public:
	static bool used ( const automaton::DFTA & automaton, const automaton::State & state ) {
		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for (const std::pair<const std::pair<alphabet::RankedSymbol, std::vector<automaton::State> >, automaton::State>& t : automaton.getTransitions())
			if(std::contains(t.first.second.begin(), t.first.second.end(), state ) || t.second == state)
				return true;

		return false;
	}

	static bool available ( const automaton::DFTA &, const automaton::State & ) {
		return true;
	}

	static void valid ( const automaton::DFTA &, const automaton::State & ) {
	}
};

template < >
class ComponentConstraint2< automaton::DFTA, automaton::State, automaton::FinalStates > {
public:
	static bool used ( const automaton::DFTA &, const automaton::State & ) {
		return false;
	}

	static bool available ( const automaton::DFTA & automaton, const automaton::State & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::DFTA &, const automaton::State & ) {
	}
};

} /* namespace std */

namespace alib {

auto DFTAParserRegister = xmlApi<automaton::Automaton>::ParserRegister<automaton::DFTA>();
auto DFTAParserRegister2 = xmlApi<alib::Object>::ParserRegister<automaton::DFTA>();

} /* namespace alib */
