/*
 * OneTapeDTM.cpp
 *
 *  Created on: Apr 24, 2013
 *      Author: Martin Zak
 */

#include "OneTapeDTM.h"
#include "../AutomatonException.h"
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include <object/Object.h>
#include <XmlApi.hpp>

namespace automaton {

OneTapeDTM::OneTapeDTM ( std::set < State > states, std::set < alphabet::Symbol > tapeAlphabet, alphabet::Symbol blankSymbol, std::set< alphabet::Symbol > inputAlphabet, State initialState, std::set < State > finalStates ) : std::Components2 < OneTapeDTM, alphabet::Symbol, std::tuple < TapeAlphabet, InputAlphabet >, std::tuple < BlankSymbol >, automaton::State, std::tuple < States, FinalStates >, std::tuple < InitialState > > ( std::make_tuple ( std::move ( tapeAlphabet), std::move ( inputAlphabet ) ), std::make_tuple ( blankSymbol ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::make_tuple ( std::move ( initialState ) ) ) {

}

OneTapeDTM::OneTapeDTM(State initial, alphabet::Symbol blank) : OneTapeDTM(std::set<automaton::State>{initial}, std::set<alphabet::Symbol> {blank}, blank, std::set<alphabet::Symbol>{}, initial, std::set<automaton::State>{}) {

}

AutomatonBase* OneTapeDTM::clone() const {
	return new OneTapeDTM(*this);
}

AutomatonBase* OneTapeDTM::plunder() && {
	return new OneTapeDTM(std::move(*this));
}

bool OneTapeDTM::addTransition(State from, alphabet::Symbol input, State to, alphabet::Symbol output, Shift shift) {
	if (!getStates().count(from)) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (!getTapeAlphabet().count(input)) {
		throw AutomatonException("Tape symbol \"" + (std::string) input + "\" doesn't exist.");
	}

	if (!getStates().count(to)) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	if (!getTapeAlphabet().count(output)) {
		throw AutomatonException("Tape symbol  \"" + (std::string) output + "\" doesn't exist.");
	}

	std::pair<State, alphabet::Symbol> key = std::make_pair(std::move(from), std::move(input));

	std::tuple<State, alphabet::Symbol, Shift > value(std::move(to), std::move(output), shift);

	if (transitions.find(key) != transitions.end()) {
		if(transitions.find(key)->second == value)
			return false;
		else
			throw AutomatonException("Transition (\"" + (std::string) key.first.getName() + "\", \"" + (std::string) key.second + "\") -> ? already exists.");
	}

	transitions.insert(std::make_pair(std::move(key), std::move(value)));
	return true;
}

bool OneTapeDTM::removeTransition(const State& from, const alphabet::Symbol& input, const State& to, const alphabet::Symbol& output, const Shift& shift) {
	std::pair<State, alphabet::Symbol> key = std::make_pair(from, input);

	if (transitions.find(key) == transitions.end())
		return false;

	std::tuple<State, alphabet::Symbol, Shift > value(to, output, shift);
	if(transitions.find(key)->second != value) 
		throw AutomatonException("Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input + "\") -> ? doesn't exists.");

	transitions.erase(key);
	return true;
}

const std::map<std::pair<State, alphabet::Symbol>, std::tuple<State, alphabet::Symbol, Shift> >& OneTapeDTM::getTransitions() const {
	return transitions;
}

int OneTapeDTM::compare(const OneTapeDTM& other) const {
	auto first = std::tie(getStates(), getInputAlphabet(), getInitialState(), getFinalStates(), getTapeAlphabet(), getBlankSymbol(), transitions);
	auto second = std::tie(other.getStates(), other.getInputAlphabet(), other.getInitialState(), other.getFinalStates(), other.getTapeAlphabet(), other.getBlankSymbol(), other.transitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void OneTapeDTM::operator>>(std::ostream& out) const {
	out << "(OneTapeDTM "
			<< "states = " << getStates()
			<< "inputAlphabet = " << getInputAlphabet()
			<< "initialState = " << getInitialState()
			<< "finalStates = " << getFinalStates()
			<< "tapeAlphabet = " << getTapeAlphabet()
			<< "blankSymbol = " << getBlankSymbol()
			<< "transitions = " << transitions
			<< ")";
}

OneTapeDTM::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

OneTapeDTM OneTapeDTM::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, OneTapeDTM::getXmlTagName());

	std::set<State> states = AutomatonFromXMLParser::parseStates(input);
	std::set<alphabet::Symbol> tapeSymbols = AutomatonFromXMLParser::parseTapeAlphabet(input);
	std::set<alphabet::Symbol> inputSymbols = AutomatonFromXMLParser::parseInputAlphabet(input);
	alphabet::Symbol blank = AutomatonFromXMLParser::parseBlankSymbol(input);
	State initialState = AutomatonFromXMLParser::parseInitialState(input);
	std::set<State> finalStates = AutomatonFromXMLParser::parseFinalStates(input);

	OneTapeDTM automaton(std::move(initialState), std::move(blank));
	automaton.setStates(std::move(states));
	automaton.setTapeAlphabet(std::move(tapeSymbols));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<OneTapeDTM>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, OneTapeDTM::getXmlTagName());
	return automaton;
}

void OneTapeDTM::parseTransition(std::deque<sax::Token>::iterator& input, OneTapeDTM& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transition");
	State from = AutomatonFromXMLParser::parseTransitionFrom(input);
	alphabet::Symbol inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol(input);
	State to = AutomatonFromXMLParser::parseTransitionTo(input);
	alphabet::Symbol outputSymbol = AutomatonFromXMLParser::parseTransitionOutputSymbol(input);
	Shift shift = AutomatonFromXMLParser::parseTransitionShift(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transition");

	automaton.addTransition(std::move(from), std::move(inputSymbol), std::move(to), std::move(outputSymbol), shift);
}

void OneTapeDTM::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(OneTapeDTM::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeTapeAlphabet(out, this->getTapeAlphabet());
	AutomatonToXMLComposer::composeInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composeBlankSymbol(out, this->getBlankSymbol());
	AutomatonToXMLComposer::composeInitialState(out, this->getInitialState());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(OneTapeDTM::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

void OneTapeDTM::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getTransitions()) {
		out.emplace_back("transition", sax::Token::TokenType::START_ELEMENT);

		AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
		AutomatonToXMLComposer::composeTransitionInputSymbol(out, transition.first.second);
		AutomatonToXMLComposer::composeTransitionTo(out, std::get<0>(transition.second));
		AutomatonToXMLComposer::composeTransitionOutputSymbol(out, std::get<1>(transition.second));
		AutomatonToXMLComposer::composeTransitionShift(out, std::get<2>(transition.second));

		out.emplace_back("transition", sax::Token::TokenType::END_ELEMENT);
	}

	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace automaton */

namespace std {

template < >
class ComponentConstraint2< automaton::OneTapeDTM, alphabet::Symbol, automaton::TapeAlphabet > {
public:
	static bool used ( const automaton::OneTapeDTM & automaton, const alphabet::Symbol & symbol ) {
		if ( automaton.getBlankSymbol ( ) == symbol )
			return true;

		if ( automaton.getInputAlphabet().count(symbol))
			return true;

		for (const std::pair<const std::pair<automaton::State, alphabet::Symbol>, std::tuple<automaton::State, alphabet::Symbol, automaton::Shift> >& transition : automaton.getTransitions())
			if (symbol == transition.first.second || symbol == std::get<1>(transition.second))
				return true;

		return false;

	}

	static bool available ( const automaton::OneTapeDTM &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const automaton::OneTapeDTM &, const alphabet::Symbol & ) {
	}
};

template < >
class ComponentConstraint2< automaton::OneTapeDTM, alphabet::Symbol, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::OneTapeDTM &, const alphabet::Symbol & ) {
		return false;
	}

	static bool available ( const automaton::OneTapeDTM & automaton, const alphabet::Symbol & symbol ) {
		return automaton.accessComponent < automaton::TapeAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const automaton::OneTapeDTM & automaton, const alphabet::Symbol & symbol ) {
		if (symbol == automaton.getBlankSymbol())
			throw automaton::AutomatonException("Input symbol \"" + (std::string) symbol + "\" cannot be blank symbol.");
	}
};

template < >
class ElementConstraint2< automaton::OneTapeDTM, alphabet::Symbol, automaton::BlankSymbol > {
public:
	static bool available ( const automaton::OneTapeDTM & automaton, const alphabet::Symbol & symbol ) {
		return automaton.accessComponent < automaton::TapeAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const automaton::OneTapeDTM & automaton, const alphabet::Symbol & symbol ) {
		if (automaton.getInputAlphabet().count( symbol ))
			throw automaton::AutomatonException("Blank symbol \"" + (std::string) symbol + "\" cannot be in input alphabet.");
	}
};

template < >
class ComponentConstraint2< automaton::OneTapeDTM, automaton::State, automaton::States > {
public:
	static bool used ( const automaton::OneTapeDTM & automaton, const automaton::State & state ) {
		if ( automaton.getInitialState ( ) == state )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for (const std::pair<const std::pair<automaton::State, alphabet::Symbol>, std::tuple<automaton::State, alphabet::Symbol, automaton::Shift> >& transition : automaton.getTransitions ( ) )
			if ( state == transition.first.first || state == std::get < 0 > ( transition.second ) )
				return true;

		return false;
	}

	static bool available ( const automaton::OneTapeDTM &, const automaton::State & ) {
		return true;
	}

	static void valid ( const automaton::OneTapeDTM &, const automaton::State & ) {
	}
};

template < >
class ComponentConstraint2< automaton::OneTapeDTM, automaton::State, automaton::FinalStates > {
public:
	static bool used ( const automaton::OneTapeDTM &, const automaton::State & ) {
		return false;
	}

	static bool available ( const automaton::OneTapeDTM & automaton, const automaton::State & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::OneTapeDTM &, const automaton::State & ) {
	}
};

template < >
class ElementConstraint2< automaton::OneTapeDTM, automaton::State, automaton::InitialState > {
public:
	static bool available ( const automaton::OneTapeDTM & automaton, const automaton::State & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::OneTapeDTM &, const automaton::State & ) {
	}
};

} /* namespace std */

namespace alib {

auto oneTapeDTMParserRegister = xmlApi<automaton::Automaton>::ParserRegister<automaton::OneTapeDTM>();
auto oneTapeDTMParserRegister2 = xmlApi<alib::Object>::ParserRegister<automaton::OneTapeDTM>();

} /* namespace alib */
