/*
 * XmlDataFactory.hpp
 *
 *  Created on: Jan 1, 2014
 *      Author: Jan Travnicek
 */

#ifndef DATA_FACTORY_H_
#define DATA_FACTORY_H_

#include <string>
#include <deque>
#include <measure>

#include <sax/Token.h>
#include <sax/SaxParseInterface.h>
#include <sax/SaxComposeInterface.h>
#include <XmlApi.hpp>
#include <exception/CommonException.h>

namespace alib {

/**
 * Data builder.
 */
class XmlDataFactory {
public:
	/**
	 * Parses the XML in file and returns the String.
	 * @param filename path to the file
	 * @return String
	 */
	template < class T >
	static T fromFile ( const std::string & filename ) {
		std::deque < sax::Token > tokens;
		sax::SaxParseInterface::parseFile ( filename, tokens );

		return fromTokens < T > ( std::move( tokens ) );
	}

	/**
	 * Parses the XML and returns the String.
	 * @param str string containing the XML
	 * @return String
	 */
	template < class T >
	static T fromString ( const std::string & str ) {
		std::deque < sax::Token > tokens;
		sax::SaxParseInterface::parseMemory ( str, tokens );

		return fromTokens < T > ( std::move ( tokens ) );
	}

	/**
	 * Parses the XML from stdin and returns the String.
	 * @return String
	 */
	template < class T >
	static T fromStdin ( ) {
		std::deque < sax::Token > tokens;
		sax::SaxParseInterface::parseStdin ( tokens );

		return fromTokens < T > ( std::move ( tokens ) );
	}

	/**
	 * Parses the XML from stream and returns the String.
	 * @return String
	 */
	template < class T >
	static T fromStream ( std::istream & in ) {
		std::deque < sax::Token > tokens;
		sax::SaxParseInterface::parseStream ( in, tokens );

		return fromTokens < T > ( std::move ( tokens ) );
	}

	/**
	 * Parses the String from list of tokens.
	 * @param tokens XML represented as list of tokens
	 * @return parsed String
	 */
	template < class T >
	static T fromTokens ( std::deque < sax::Token > && tokens ) {
		xmlApiInputContext context ( tokens.begin ( ) );

		if ( context == tokens.end ( ) ) throw exception::CommonException ( "Empty tokens list" );

		if ( alib::xmlApi < exception::CommonException >::first ( context ) ) throw alib::xmlApi < exception::CommonException >::parse ( context );

		measurements::start ( "XML Parser", measurements::Type::INIT );
		T res = alib::xmlApi < T >::parse ( context );
		measurements::end ( );

		if ( context != tokens.end ( ) ) throw exception::CommonException ( "Unexpeted tokens at the end of the xml" );

		return res;
	}

	/**
	 * Determines whether the input token stream contains given type
	 * @param tokens
	 * @return bool
	 */
	template < class T >
	static bool first ( const std::deque < sax::Token > & tokens ) {
		if( tokens.empty ( ) ) throw exception::CommonException ( "Empty tokens list" );

		return alib::xmlApi < exception::CommonException >::first ( tokens.begin ( ) ) || alib::xmlApi < T >::first ( tokens.begin ( ) );
	}

	/**
	 * Parses the XML in file and returns the String.
	 * @param filename path to the file
	 * @return String
	 */
	template < class T >
	static void toFile ( const T & data, const std::string & filename ) {
		std::deque < sax::Token > tokens = toTokens < T > ( data );
		sax::SaxComposeInterface::printFile ( filename, tokens );
	}

	/**
	 * Parses the XML and returns the String.
	 * @param str string containing the XML
	 * @return String
	 */
	template < class T >
	static std::string toString ( const T & data ) {
		std::deque < sax::Token > tokens = toTokens < T > ( data );
		std::string str;
		sax::SaxComposeInterface::printMemory ( str, tokens );

		return str;
	}

	/**
	 * Parses the XML from stdin and returns the String.
	 * @return String
	 */
	template < class T >
	static void toStdout ( const T & data ) {
		std::deque < sax::Token > tokens = toTokens < T > ( data );
		sax::SaxComposeInterface::printStdout ( tokens );
	}

	/**
	 * Parses the XML from stream and returns the String.
	 * @return String
	 */
	template < class T >
	static void toStream ( const T & data, std::ostream & out ) {
		std::deque < sax::Token > tokens = toTokens < T > ( data );
		sax::SaxComposeInterface::printStream ( out, tokens );
	}

	/**
	 * Parses the String from list of tokens.
	 * @param tokens XML represented as list of tokens
	 * @return parsed String
	 */
	template < class T >
	static std::deque < sax::Token > toTokens ( const T & data ) {
		xmlApiOutputContext context;

		measurements::start ( "XML Composer", measurements::Type::FINALIZE );
		alib::xmlApi < T >::compose ( context, data );
		measurements::end ( );

		return context;
	}

};

template < >
inline bool XmlDataFactory::first < exception::CommonException > ( const std::deque < sax::Token > & tokens ) {
	return alib::xmlApi < exception::CommonException >::first ( tokens.begin ( ) );
}

template < >
inline exception::CommonException XmlDataFactory::fromTokens < exception::CommonException > ( std::deque < sax::Token > && tokens ) {
	std::deque < sax::Token >::iterator iter = tokens.begin ( );

	if ( iter == tokens.end ( ) ) throw exception::CommonException ( "Empty tokens list" );

	exception::CommonException res = alib::xmlApi < exception::CommonException >::parse ( iter );

	if ( iter != tokens.end ( ) ) throw exception::CommonException ( "Unexpeted tokens at the end of the xml" );

	return res;
}

} /* namespace alib */

#endif /* DATA_FACTORY_H_ */
