/*
 * EpsilonFreeCFG.h
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#ifndef EPSILON_FREE_CFG_H_
#define EPSILON_FREE_CFG_H_

#include "../GrammarBase.h"
#include <map>
#include <vector>
#include <core/components.hpp>
#include "../../alphabet/Symbol.h"

namespace grammar {

/**
 * Epsilon free context free grammar. Type 2 in Chomsky hierarchy. Produces context free languages.
 */
class TerminalAlphabet;
class NonterminalAlphabet;
class InitialSymbol;

class EpsilonFreeCFG : public GrammarBase, public std::Components < EpsilonFreeCFG, alphabet::Symbol, std::tuple < TerminalAlphabet, NonterminalAlphabet >, std::tuple < InitialSymbol > > {
	std::map < alphabet::Symbol, std::set < std::vector < alphabet::Symbol > > > rules;
	bool generatesEpsilon;

public:
	explicit EpsilonFreeCFG ( alphabet::Symbol initialSymbol );

	explicit EpsilonFreeCFG ( std::set < alphabet::Symbol > nonTerminalSymbols, std::set < alphabet::Symbol > terminalSymbols, alphabet::Symbol initialSymbol );

	virtual GrammarBase * clone ( ) const;

	virtual GrammarBase * plunder ( ) &&;

	bool addRule ( alphabet::Symbol leftHandSide, std::vector < alphabet::Symbol > rightHandSide );

	const std::map < alphabet::Symbol, std::set < std::vector < alphabet::Symbol > > > & getRules ( ) const;

	bool removeRule ( const alphabet::Symbol & leftHandSide, const std::vector < alphabet::Symbol > & rightHandSide );

	bool addRawRule ( alphabet::Symbol leftHandSide, std::vector < alphabet::Symbol > rightHandSide );

	std::map < alphabet::Symbol, std::set < std::vector < alphabet::Symbol > > > getRawRules ( ) const;

	bool removeRawRule ( const alphabet::Symbol & leftHandSide, const std::vector < alphabet::Symbol > & rightHandSide );

	const alphabet::Symbol & getInitialSymbol ( ) const {
		return accessElement < InitialSymbol > ( ).get ( );
	}

	bool setInitialSymbol ( alphabet::Symbol symbol ) {
		return accessElement < InitialSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < alphabet::Symbol > & getNonterminalAlphabet ( ) const {
		return accessComponent < NonterminalAlphabet > ( ).get ( );
	}

	bool addNonterminalSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < NonterminalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void setNonterminalAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < NonterminalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	const std::set < alphabet::Symbol > & getTerminalAlphabet ( ) const {
		return accessComponent < TerminalAlphabet > ( ).get ( );
	}

	bool addTerminalSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < TerminalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void setTerminalAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < TerminalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void setGeneratesEpsilon ( bool genEps );
	bool getGeneratesEpsilon ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const EpsilonFreeCFG & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "EpsilonFreeCFG";

		return xmlTagName;
	}

	static EpsilonFreeCFG parse ( std::deque < sax::Token >::iterator & input );
	static void parseRule ( std::deque < sax::Token >::iterator & input, EpsilonFreeCFG & grammar );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeRules ( std::deque < sax::Token > & out ) const;
};

} /* namespace grammar */

#endif /* EPSILON_FREE_CFG_H_ */
