/*
 * NonContractingGrammar.cpp
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#include "NonContractingGrammar.h"
#include "../GrammarException.h"
#include <algorithm>
#include <sstream>

#include "../../alphabet/Symbol.h"

#include <sax/FromXMLParserHelper.h>
#include "../common/GrammarFromXMLParser.h"
#include "../common/GrammarToXMLComposer.h"
#include "../Grammar.h"
#include <object/Object.h>
#include <XmlApi.hpp>

namespace grammar {

NonContractingGrammar::NonContractingGrammar ( alphabet::Symbol initialSymbol ) : NonContractingGrammar ( std::set < alphabet::Symbol > { initialSymbol }, std::set < alphabet::Symbol > ( ), initialSymbol ) {
}

NonContractingGrammar::NonContractingGrammar ( std::set < alphabet::Symbol > nonterminalAlphabet, std::set < alphabet::Symbol > terminalAlphabet, alphabet::Symbol initialSymbol ) : std::Components < NonContractingGrammar, alphabet::Symbol, std::tuple < TerminalAlphabet, NonterminalAlphabet >, std::tuple < InitialSymbol > > ( std::make_tuple ( std::move ( terminalAlphabet ), std::move ( nonterminalAlphabet ) ), std::make_tuple ( std::move ( initialSymbol ) ) ), generatesEpsilon ( false ) {
}

GrammarBase * NonContractingGrammar::clone ( ) const {
	return new NonContractingGrammar ( * this );
}

GrammarBase * NonContractingGrammar::plunder ( ) && {
	return new NonContractingGrammar ( std::move ( * this ) );
}

bool NonContractingGrammar::addRule ( std::vector < alphabet::Symbol > leftHandSide, std::vector < alphabet::Symbol > rightHandSide ) {
	int lSize = leftHandSide.size ( );
	int rSize = rightHandSide.size ( );

	if ( lSize > rSize )
		throw GrammarException ( "Invalid size of right hand side of a rule" );

	if ( std::all_of ( leftHandSide.begin ( ), leftHandSide.end ( ), [this] ( const alphabet::Symbol symbol ) {
			return !getNonterminalAlphabet ( ).count ( symbol );
		} ) )
		throw GrammarException ( "Rule must rewrite nonterminal symbol" );

	for ( const alphabet::Symbol & symbol : leftHandSide )
		if ( !getTerminalAlphabet ( ).count ( symbol ) && !getNonterminalAlphabet ( ).count ( symbol ) )
			throw GrammarException ( "Symbol \"" + ( std::string ) symbol + "\" is not neither terminal nor nonterminal symbol" );

	for ( const alphabet::Symbol & symbol : rightHandSide )
		if ( !getTerminalAlphabet ( ).count ( symbol ) && !getNonterminalAlphabet ( ).count ( symbol ) )
			throw GrammarException ( "Symbol \"" + ( std::string ) symbol + "\" is not neither terminal nor nonterminal symbol" );

	return rules[std::move ( leftHandSide )].insert ( std::move ( rightHandSide ) ).second;
}

const std::map < std::vector < alphabet::Symbol >, std::set < std::vector < alphabet::Symbol > > > & NonContractingGrammar::getRules ( ) const {
	return rules;
}

bool NonContractingGrammar::removeRule ( const std::vector < alphabet::Symbol > & leftHandSide, const std::vector < alphabet::Symbol > & rightHandSide ) {
	return rules[leftHandSide].erase ( rightHandSide );
}

void NonContractingGrammar::setGeneratesEpsilon ( bool genEps ) {
	generatesEpsilon = genEps;
}

bool NonContractingGrammar::getGeneratesEpsilon ( ) const {
	return generatesEpsilon;
}

int NonContractingGrammar::compare ( const NonContractingGrammar & other ) const {
	auto first = std::tie ( getTerminalAlphabet ( ), getNonterminalAlphabet ( ), getInitialSymbol ( ), rules );
	auto second = std::tie ( other.getTerminalAlphabet ( ), other.getNonterminalAlphabet ( ), other.getInitialSymbol ( ), other.rules );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void NonContractingGrammar::operator >>( std::ostream & out ) const {
	out << "(NonContractingGrammar "
	    << "nonterminalAlphabet = " << getNonterminalAlphabet ( )
	    << "terminalAlphabet = " << getTerminalAlphabet ( )
	    << "initialSymbol = " << getInitialSymbol ( )
	    << "rules = " << rules
	    << "generatesEpsilon = " << generatesEpsilon << ")";
}

NonContractingGrammar::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

NonContractingGrammar NonContractingGrammar::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, NonContractingGrammar::getXmlTagName() );

	std::set < alphabet::Symbol > nonterminalAlphabet = GrammarFromXMLParser::parseNonterminalAlphabet ( input );
	std::set < alphabet::Symbol > terminalAlphabet = GrammarFromXMLParser::parseTerminalAlphabet ( input );
	alphabet::Symbol initialSymbol = GrammarFromXMLParser::parseInitialSymbol ( input );

	NonContractingGrammar grammar ( std::move ( initialSymbol ) );

	grammar.setNonterminalAlphabet ( std::move ( nonterminalAlphabet ) );
	grammar.setTerminalAlphabet ( std::move ( terminalAlphabet ) );

	GrammarFromXMLParser::parseRules ( input, grammar );

	bool generatesEpsilon = GrammarFromXMLParser::parseGeneratesEpsilon ( input );
	grammar.setGeneratesEpsilon ( generatesEpsilon );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, NonContractingGrammar::getXmlTagName() );
	return grammar;
}

void NonContractingGrammar::parseRule ( std::deque < sax::Token >::iterator & input, NonContractingGrammar & grammar ) {
	std::vector < alphabet::Symbol > lhs = GrammarFromXMLParser::parseRuleLHS ( input );
	std::vector < alphabet::Symbol > rhs = GrammarFromXMLParser::parseRuleRHS ( input );

	grammar.addRule ( std::move ( lhs ), std::move ( rhs ) );
}

void NonContractingGrammar::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( NonContractingGrammar::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	GrammarToXMLComposer::composeNonterminalAlphabet ( out, this->getNonterminalAlphabet ( ) );
	GrammarToXMLComposer::composeTerminalAlphabet ( out, this->getTerminalAlphabet ( ) );
	GrammarToXMLComposer::composeInitialSymbol ( out, this->getInitialSymbol ( ) );
	composeRules ( out );
	GrammarToXMLComposer::composeGeneratesEpsilon ( out, this->getGeneratesEpsilon ( ) );

	out.emplace_back ( NonContractingGrammar::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

void NonContractingGrammar::composeRules ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( "rules", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & rule : this->getRules ( ) )

		for ( const auto & rhs : rule.second ) {
			out.emplace_back ( "rule", sax::Token::TokenType::START_ELEMENT );

			GrammarToXMLComposer::composeRuleLHS ( out, rule.first );
			GrammarToXMLComposer::composeRuleRHS ( out, rhs );

			out.emplace_back ( "rule", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "rules", sax::Token::TokenType::END_ELEMENT );
}

} /* namespace grammar */

namespace std {

template < >
bool grammar::NonContractingGrammar::Component < grammar::NonContractingGrammar, alphabet::Symbol, grammar::TerminalAlphabet >::used ( const alphabet::Symbol & symbol ) const {
	const grammar::NonContractingGrammar * grammar = static_cast < const grammar::NonContractingGrammar * > ( this );

	for ( const std::pair < const std::vector < alphabet::Symbol >, std::set < std::vector < alphabet::Symbol > > > & rule : grammar->getRules ( ) ) {
		if ( std::find ( rule.first.begin ( ), rule.first.end ( ), symbol ) != rule.first.end ( ) )
			return true;

		for ( const std::vector < alphabet::Symbol > & rhs : rule.second )
			if ( std::find ( rhs.begin ( ), rhs.end ( ), symbol ) != rhs.end ( ) )
				return true;

	}

	return false;
}

template < >
bool grammar::NonContractingGrammar::Component < grammar::NonContractingGrammar, alphabet::Symbol, grammar::TerminalAlphabet >::available ( const alphabet::Symbol & ) const {
	return true;
}

template < >
void grammar::NonContractingGrammar::Component < grammar::NonContractingGrammar, alphabet::Symbol, grammar::TerminalAlphabet >::valid ( const alphabet::Symbol & symbol ) const {
	const grammar::NonContractingGrammar * grammar = static_cast < const grammar::NonContractingGrammar * > ( this );

	if ( grammar->accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol ) )
		throw grammar::GrammarException ( "Symbol " + ( std::string ) symbol + "cannot be in terminal alphabet since it is already nonterminal alphabet" );
}

template < >
bool grammar::NonContractingGrammar::Component < grammar::NonContractingGrammar, alphabet::Symbol, grammar::NonterminalAlphabet >::used ( const alphabet::Symbol & symbol ) const {
	const grammar::NonContractingGrammar * grammar = static_cast < const grammar::NonContractingGrammar * > ( this );

	for ( const std::pair < const std::vector < alphabet::Symbol >, std::set < std::vector < alphabet::Symbol > > > & rule : grammar->getRules ( ) ) {
		if ( std::find ( rule.first.begin ( ), rule.first.end ( ), symbol ) != rule.first.end ( ) )
			return true;

		for ( const std::vector < alphabet::Symbol > & rhs : rule.second )
			if ( std::find ( rhs.begin ( ), rhs.end ( ), symbol ) != rhs.end ( ) )
				return true;

	}

	if ( grammar->accessElement < grammar::InitialSymbol > ( ).get ( ) == symbol )
		return true;

	return false;
}

template < >
bool grammar::NonContractingGrammar::Component < grammar::NonContractingGrammar, alphabet::Symbol, grammar::NonterminalAlphabet >::available ( const alphabet::Symbol & ) const {
	return true;
}

template < >
void grammar::NonContractingGrammar::Component < grammar::NonContractingGrammar, alphabet::Symbol, grammar::NonterminalAlphabet >::valid ( const alphabet::Symbol & symbol ) const {
	const grammar::NonContractingGrammar * grammar = static_cast < const grammar::NonContractingGrammar * > ( this );

	if ( grammar->accessComponent < grammar::TerminalAlphabet > ( ).get ( ).count ( symbol ) )
		throw grammar::GrammarException ( "Symbol " + ( std::string ) symbol + "cannot be in nonterminal alphabet since it is already in terminal alphabet" );
}

template < >
bool grammar::NonContractingGrammar::Element < grammar::NonContractingGrammar, alphabet::Symbol, grammar::InitialSymbol >::available ( const alphabet::Symbol & symbol ) const {
	const grammar::NonContractingGrammar * grammar = static_cast < const grammar::NonContractingGrammar * > ( this );

	return grammar->accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol );
}

template < >
void grammar::NonContractingGrammar::Element < grammar::NonContractingGrammar, alphabet::Symbol, grammar::InitialSymbol >::valid ( const alphabet::Symbol & ) const {
}

} /* namespace std */

namespace alib {

auto nonContractingGrammarParserRegister = xmlApi < grammar::Grammar >::ParserRegister < grammar::NonContractingGrammar > ( );
auto nonContractingGrammarParserRegister2 = xmlApi < alib::Object >::ParserRegister < grammar::NonContractingGrammar > ( );

} /* namespace alib */
