/*
 * AdjacencyMatrixUndirectedGraph.cpp
 *
 *   Created on: Jun 19, 2014
 *       Author: David Rosca
 *  Modified by: Jan Broz
 */

#include "AdjacencyMatrixUndirectedGraph.h"

#include <compare>

#include <object/Object.h>
#include <XmlApi.hpp>
#include <sax/FromXMLParserHelper.h>
#include "../common/GraphFromXMLParser.h"
#include "../common/GraphToXMLComposer.h"
#include "../Graph.h"
#include "utils.h"


namespace graph {

AdjacencyMatrixUndirectedGraph::AdjacencyMatrixUndirectedGraph() : UndirectedGraph() {
}

AdjacencyMatrixUndirectedGraph::AdjacencyMatrixUndirectedGraph( const UndirectedGraph & graph ) {
	UndirectedGraph::fromUndirected( graph );
}

AdjacencyMatrixUndirectedGraph::AdjacencyMatrixUndirectedGraph( const DirectedGraph & dgraph ) {
	UndirectedGraph::fromDirected( dgraph );
}

GraphBase * AdjacencyMatrixUndirectedGraph::clone() const {
	return new AdjacencyMatrixUndirectedGraph( *this );
}

GraphBase * AdjacencyMatrixUndirectedGraph::plunder() && {
	return new AdjacencyMatrixUndirectedGraph( std::move( *this ) );
}

REPRESENTATION AdjacencyMatrixUndirectedGraph::getRepresentation() const
{
	return REPRESENTATION::ADJACENCY_MATRIX;
}

std::string AdjacencyMatrixUndirectedGraph::className() const
{
	return "AdjacencyMatrixUndirectedGraph";
}

std::set<Node> AdjacencyMatrixUndirectedGraph::getNodes() const
{
	std::set<Node> out;

	for (auto i : adj) {
		out.insert(i.first);
	}

	return out;
}

std::set<UndirectedEdge> AdjacencyMatrixUndirectedGraph::getEdges() const
{
	return edges;
}

size_t AdjacencyMatrixUndirectedGraph::nodeCount() const
{
	return adj.size();
}

size_t AdjacencyMatrixUndirectedGraph::edgeCount() const
{
	return edges.size();
}

bool AdjacencyMatrixUndirectedGraph::hasEdge( const Node & first, const Node & second ) const
{
	auto search = adj.find(first);
	if (search == adj.end()) {
		return false;
	}
	return search->second.find(second) != search->second.end();
}

bool AdjacencyMatrixUndirectedGraph::hasNode( const Node & node ) const
{
	return adj.find(node) != adj.end();
}

bool AdjacencyMatrixUndirectedGraph::hasEdge( const UndirectedEdge & edge ) const
{
	return edges.find(edge) != edges.end();
}

std::set<UndirectedEdge> AdjacencyMatrixUndirectedGraph::findEdges( const Node & first, const Node & second ) const
{
	std::set<UndirectedEdge> out;

	for (const UndirectedEdge &e : edges) {
		if (edgeMatch(e, first, second)) {
			out.insert(e);
		}
	}

	return out;
}

std::set<Node> AdjacencyMatrixUndirectedGraph::neighbors( const Node & node ) const
{
	std::set<Node> out;

	auto search = adj.find(node);
	if (search == adj.end()) {
		return out;
	}

	for (auto i : search->second) {
		if (i.second > 0) {
			out.insert(i.first);
		}
	}

	return out;
}

std::set<UndirectedEdge> AdjacencyMatrixUndirectedGraph::neighborEdges( const Node & node ) const
{
	std::set<UndirectedEdge> out;
	const std::set<Node> &nghbrs = neighbors(node);

	for (const UndirectedEdge &e : edges) {
		if (edgeMatch(e, node, nghbrs)) {
			out.insert(e);
		}
	}

	return out;
}

bool AdjacencyMatrixUndirectedGraph::addNode( const Node & node )
{
	if (adj.find(node) != adj.end())
		return false;

	adj[node]; // create empty map for this key
	return true;
}

bool AdjacencyMatrixUndirectedGraph::addNode( const Node & node, int value )
{
	if (!addNode( node ))
		return false;

	UndirectedGraph::addNode( node, value );
	return true;
}

bool AdjacencyMatrixUndirectedGraph::removeNode( const Node & node )
{
	auto search = adj.find(node);
	if (search == adj.end()) {
		return false;
	}

	adj.erase(search);
	UndirectedGraph::removeNode(node);

	for (auto& i : edges) {
		if (i.getFirstNode() == node) {
			adj[i.getSecondNode()].erase(node);
			edges.erase(i);
		} else if (i.getSecondNode() == node) {
			adj[i.getFirstNode()].erase(node);
			edges.erase(i);
		}
	}

	return true;
}

bool AdjacencyMatrixUndirectedGraph::addEdge( const UndirectedEdge & edge )
{
	if (edges.find(edge) != edges.end())
		return false;
	if (adj.find(edge.getFirstNode()) == adj.end())
		return false;
	if (adj.find(edge.getSecondNode()) == adj.end())
		return false;

	edges.insert(edge);

	adj[edge.getFirstNode()][edge.getSecondNode()] += 1;
	adj[edge.getSecondNode()][edge.getFirstNode()] += 1;

	return true;
}

bool AdjacencyMatrixUndirectedGraph::addEdge( const UndirectedEdge & edge, int value )
{
	if (!addEdge( edge ))
		return false;

	UndirectedGraph::addEdge( edge, value );
	return true;
}

bool AdjacencyMatrixUndirectedGraph::removeEdge( const UndirectedEdge & edge )
{
	auto search = edges.find(edge);
	if (search == edges.end()) {
		return false;
	}

	edges.erase(search);
	UndirectedGraph::removeEdge(edge);

	auto &map1 = adj[edge.getFirstNode()];
	int &val1 = map1[edge.getSecondNode()];
	if (--val1 <= 0) {
		map1.erase(edge.getSecondNode());
	}

	auto &map2 = adj[edge.getSecondNode()];
	int &val2 = map2[edge.getFirstNode()];
	if (--val2 <= 0) {
		map2.erase(edge.getFirstNode());
	}

	return true;
}

AdjacencyMatrixUndirectedGraph AdjacencyMatrixUndirectedGraph::parse( std::deque<sax::Token>::iterator & input )
{
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, getXmlTagName());

	AdjacencyMatrixUndirectedGraph graph;
	GraphFromXMLParser::parseNodes(input, graph);
	GraphFromXMLParser::parseUndirectedEdges(input, graph);
	GraphFromXMLParser::parseNodeValues(input, graph);
	GraphFromXMLParser::parseUndirectedEdgeValues(input, graph);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, getXmlTagName());
	return graph;
}

void AdjacencyMatrixUndirectedGraph::compose( std::deque<sax::Token> & out ) const
{
	out.emplace_back(getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	GraphToXMLComposer::composeNodes(out, this->getNodes());
	GraphToXMLComposer::composeEdges(out, this->getEdges());
	GraphToXMLComposer::composeNodeValues(out, *this);
	GraphToXMLComposer::composeEdgeValues(out, *this);

	out.emplace_back(getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}


} // namespace graph

namespace alib {

auto adjMatUndirGraphParserRegister  = xmlApi<graph::Graph>::ParserRegister<graph::AdjacencyMatrixUndirectedGraph>();
auto adjMatUndirGraphParserRegister2 = xmlApi<alib::Object>::ParserRegister<graph::AdjacencyMatrixUndirectedGraph>();

} /* namespace alib */
