/*
 * SuffixTrieTerminatingSymbol.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "SuffixTrieTerminatingSymbol.h"
#include <exception/CommonException.h>

#include <iostream>
#include <algorithm>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include "../common/IndexFromXMLParser.h"
#include "../common/IndexToXMLComposer.h"
#include <object/Object.h>
#include <XmlApi.hpp>

namespace indexes {

SuffixTrieTerminatingSymbol::SuffixTrieTerminatingSymbol ( std::set < alphabet::Symbol > alphabet, alphabet::Symbol terminatingSymbol ) : SuffixTrieTerminatingSymbol ( std::move ( alphabet ), std::move ( terminatingSymbol ), SuffixTrieNodeTerminatingSymbol ( { } ) ) {
}

SuffixTrieTerminatingSymbol::SuffixTrieTerminatingSymbol ( std::set < alphabet::Symbol > alphabet, alphabet::Symbol terminatingSymbol, SuffixTrieNodeTerminatingSymbol tree ) : std::Components < SuffixTrieTerminatingSymbol, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < TerminatingSymbol > > ( std::make_tuple ( std::move ( alphabet ) ), std::make_tuple ( std::move ( terminatingSymbol ) ) ), tree ( NULL ) {
	setTree ( std::move ( tree ) );
}

SuffixTrieTerminatingSymbol::SuffixTrieTerminatingSymbol ( alphabet::Symbol terminatingSymbol, SuffixTrieNodeTerminatingSymbol tree ) : SuffixTrieTerminatingSymbol ( tree.computeMinimalAlphabet ( ) + std::set < alphabet::Symbol > { terminatingSymbol }, terminatingSymbol, tree ) {
}

SuffixTrieTerminatingSymbol::SuffixTrieTerminatingSymbol ( const SuffixTrieTerminatingSymbol & other ) : std::Components < SuffixTrieTerminatingSymbol, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < TerminatingSymbol > > ( std::make_tuple ( other.getAlphabet ( ) ), std::make_tuple ( other.getTerminatingSymbol ( ) ) ), tree ( other.tree->clone ( ) ) {
	this->tree->attachTree ( this );
}

SuffixTrieTerminatingSymbol::SuffixTrieTerminatingSymbol ( SuffixTrieTerminatingSymbol && other ) noexcept : std::Components < SuffixTrieTerminatingSymbol, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < TerminatingSymbol > > ( std::make_tuple ( std::move ( other.accessComponent < GeneralAlphabet > ( ).get ( ) ) ), std::make_tuple ( std::move ( other.accessElement < TerminatingSymbol > ( ).get ( ) ) ) ), tree ( other.tree ) {
	this->tree->attachTree ( this );
	other.tree = NULL;
}

alib::ObjectBase * SuffixTrieTerminatingSymbol::clone ( ) const {
	return new SuffixTrieTerminatingSymbol ( * this );
}

alib::ObjectBase * SuffixTrieTerminatingSymbol::plunder ( ) && {
	return new SuffixTrieTerminatingSymbol ( std::move ( * this ) );
}

SuffixTrieTerminatingSymbol & SuffixTrieTerminatingSymbol::operator =( const SuffixTrieTerminatingSymbol & other ) {
	if ( this == & other )
		return * this;

	* this = SuffixTrieTerminatingSymbol ( other );

	return * this;
}

SuffixTrieTerminatingSymbol & SuffixTrieTerminatingSymbol::operator =( SuffixTrieTerminatingSymbol && other ) noexcept {
	std::swap ( this->tree, other.tree );
	std::swap ( accessComponent < GeneralAlphabet > ( ).get ( ), other.accessComponent < GeneralAlphabet > ( ).get ( ) );
	std::swap ( accessElement < TerminatingSymbol > ( ).get ( ), other.accessElement < TerminatingSymbol > ( ).get ( ) );

	return * this;
}

SuffixTrieTerminatingSymbol::~SuffixTrieTerminatingSymbol ( ) noexcept {
	delete tree;
}

const SuffixTrieNodeTerminatingSymbol & SuffixTrieTerminatingSymbol::getRoot ( ) const {
	return * tree;
}

SuffixTrieNodeTerminatingSymbol & SuffixTrieTerminatingSymbol::getRoot ( ) {
	return * tree;
}

void SuffixTrieTerminatingSymbol::setTree ( SuffixTrieNodeTerminatingSymbol tree ) {
	delete this->tree;
	this->tree = std::move ( tree ).plunder ( );

	if ( !this->tree->attachTree ( this ) ) {
		delete this->tree;
		throw exception::CommonException ( "Input symbols not in the alphabet." );
	}
}

void SuffixTrieTerminatingSymbol::operator >>( std::ostream & out ) const {
	out << "(SuffixTrieTerminatingSymbol " << * ( this->tree ) << ")";
}

std::ostream & operator <<( std::ostream & out, const SuffixTrieTerminatingSymbol & instance ) {
	instance >> out;
	return out;
}

int SuffixTrieTerminatingSymbol::compare ( const SuffixTrieTerminatingSymbol & other ) const {
	auto first = std::tie ( * tree, getAlphabet(), getTerminatingSymbol() );
	auto second = std::tie ( * other.tree, other.getAlphabet(), other.getTerminatingSymbol() );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

SuffixTrieTerminatingSymbol::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

SuffixTrieTerminatingSymbol SuffixTrieTerminatingSymbol::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, SuffixTrieTerminatingSymbol::getXmlTagName() );
	std::set < alphabet::Symbol > rankedAlphabet = IndexFromXMLParser::parseAlphabet ( input );
	alphabet::Symbol terminatingSymbol = alib::xmlApi < alphabet::Symbol >::parse ( input );
	SuffixTrieNodeTerminatingSymbol * root = IndexFromXMLParser::parseSuffixTrieNodeTerminatingSymbol ( input );
	SuffixTrieTerminatingSymbol tree ( std::move ( rankedAlphabet ), terminatingSymbol, std::move ( * root ) );

	delete root;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, SuffixTrieTerminatingSymbol::getXmlTagName() );
	return tree;
}

void SuffixTrieTerminatingSymbol::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( SuffixTrieTerminatingSymbol::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	IndexToXMLComposer::composeAlphabet ( out, getAlphabet ( ) );
	alib::xmlApi < alphabet::Symbol >::compose ( out, getTerminatingSymbol ( ) );
	IndexToXMLComposer::composeNode ( out, * tree );
	out.emplace_back ( SuffixTrieTerminatingSymbol::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace indexes */

namespace std {

template < >
bool indexes::SuffixTrieTerminatingSymbol::Component < indexes::SuffixTrieTerminatingSymbol, alphabet::Symbol, indexes::GeneralAlphabet >::used ( const alphabet::Symbol & symbol ) const {
	return static_cast < const indexes::SuffixTrieTerminatingSymbol * > ( this )->getTerminatingSymbol ( ) == symbol || static_cast < const indexes::SuffixTrieTerminatingSymbol * > ( this )->getRoot ( ).testSymbol ( symbol );
}

template < >
bool indexes::SuffixTrieTerminatingSymbol::Component < indexes::SuffixTrieTerminatingSymbol, alphabet::Symbol, indexes::GeneralAlphabet >::available ( const alphabet::Symbol & ) const {
	return true;
}

template < >
bool indexes::SuffixTrieTerminatingSymbol::Element < indexes::SuffixTrieTerminatingSymbol, alphabet::Symbol, indexes::TerminatingSymbol >::available ( const alphabet::Symbol & symbol ) const {
	return static_cast < const indexes::SuffixTrieTerminatingSymbol * > ( this )->getAlphabet ( ).count ( symbol );
}

template < >
void indexes::SuffixTrieTerminatingSymbol::Component < indexes::SuffixTrieTerminatingSymbol, alphabet::Symbol, indexes::GeneralAlphabet >::valid ( const alphabet::Symbol & ) const {
}

template < >
void indexes::SuffixTrieTerminatingSymbol::Element < indexes::SuffixTrieTerminatingSymbol, alphabet::Symbol, indexes::TerminatingSymbol >::valid ( const alphabet::Symbol & ) const {
}

} /* namespace std */

namespace alib {

auto suffixTreeTerminatingSymbolParserRegister = xmlApi < alib::Object >::ParserRegister < indexes::SuffixTrieTerminatingSymbol > ( );

} /* namespace alib */
