/*
 * UnboundedRegExpElement.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef UNBOUNDED_REG_EXP_ELEMENT_H_
#define UNBOUNDED_REG_EXP_ELEMENT_H_

#include <XmlApi.hpp>
#include "../../alphabet/Symbol.h"
#include <set>
#include <tree>

namespace regexp {

class UnboundedRegExpAlternation;
class UnboundedRegExpConcatenation;
class UnboundedRegExpIteration;
class UnboundedRegExpSymbol;
class UnboundedRegExpEmpty;
class UnboundedRegExpEpsilon;

class FormalRegExpElement;

class UnboundedRegExpElementVisitor {
public:
	virtual void Visit ( void *, const UnboundedRegExpAlternation & ) const = 0;
	virtual void Visit ( void *, const UnboundedRegExpConcatenation & ) const = 0;
	virtual void Visit ( void *, const UnboundedRegExpIteration & ) const = 0;
	virtual void Visit ( void *, const UnboundedRegExpSymbol & ) const = 0;
	virtual void Visit ( void *, const UnboundedRegExpEmpty & ) const = 0;
	virtual void Visit ( void *, const UnboundedRegExpEpsilon & ) const = 0;
};

/**
 * Abstract class representing element in the regular expression. Can be operator or symbol.
 */
class UnboundedRegExpElement : public alib::CommonBase < UnboundedRegExpElement >, public std::BaseNode < UnboundedRegExpElement > {
public:
	virtual void Accept ( void * userData, const UnboundedRegExpElementVisitor & visitor ) const = 0;

	/**
	 * Creates copy of the element.
	 * @return copy of the element
	 */
	virtual std::smart_ptr < FormalRegExpElement > asFormal ( ) const = 0;

	/**
	 * Traverses the regexp tree looking if particular Symbol is used in the regexp.
	 *
	 * @param symbol to test if used in regexp element
	 * @return true if symbol is used by the element and its successor
	 */
	virtual bool testSymbol ( const alphabet::Symbol & symbol ) const = 0;

	/**
	 * Traverses the regexp tree computing minimal alphabet needed by regexp
	 *
	 * @param alphabet All alphabet symbols encountered are added into this set
	 * @return true if symbol is used by the element and its successor
	 */
	virtual void computeMinimalAlphabet ( std::set < alphabet::Symbol > & alphabet ) const = 0;

	/**
	 * Traverses the regexp tree and checks whether all symbols in the regexp tree are in the alphabet
	 *
	 * @param alphabet
	 * @return true if symbols in the regexp are in the alphabet, false otherwise
	 */
	virtual bool checkAlphabet ( const std::set < alphabet::Symbol > & alphabet ) const = 0;

	/**
	 * @copydoc RankedNode::computeMinimalAlphabet()
	 */
	std::set < alphabet::Symbol > computeMinimalAlphabet ( ) const;
};

} /* namespace regexp */

namespace alib {

template < >
struct xmlApi < regexp::UnboundedRegExpElement > {
	static std::rvalue_ref < regexp::UnboundedRegExpElement > parse ( std::deque < sax::Token >::iterator & input );
	static bool first ( const std::deque < sax::Token >::const_iterator & input );
	static void compose ( std::deque < sax::Token > & output, const regexp::UnboundedRegExpElement & data );
};

} /* namespace alib */

#endif /* UNBOUNDED_REG_EXP_ELEMENT_H_ */
