#include "FormalRTESubstitution.h"
#include <exception/CommonException.h>
#include <sstream>

// #include "../unbounded/UnboundedRTESubstitution.h"

namespace rte {

FormalRTESubstitution::FormalRTESubstitution ( FormalRTEElement && left, FormalRTEElement && right, FormalRTESymbol substitutionSymbol ) : left ( nullptr ), right ( nullptr ), substitutionSymbol ( nullptr ) {
	setLeftElement( std::move ( left ) );
	setRightElement ( std::move ( right ) );
	setSubstitutionSymbol ( std::move ( substitutionSymbol ) );
}

FormalRTESubstitution::FormalRTESubstitution ( const FormalRTEElement & left, const FormalRTEElement & right, FormalRTESymbol substitutionSymbol ) : FormalRTESubstitution ( std::move_copy ( left ), std::move_copy ( right ), std::move ( substitutionSymbol ) ) {
}

FormalRTESubstitution::FormalRTESubstitution ( const FormalRTESubstitution & other ) : left ( other.left ), right ( other.right ), substitutionSymbol ( other.substitutionSymbol ) {
}

FormalRTESubstitution::FormalRTESubstitution ( FormalRTESubstitution && other ) noexcept : left ( std::move ( other.left ) ), right ( std::move ( other.right ) ), substitutionSymbol ( std::move ( other.substitutionSymbol ) ) {
	left->attachRTE ( NULL );
	right->attachRTE ( NULL );
	substitutionSymbol->attachRTE ( NULL );
}

FormalRTESubstitution & FormalRTESubstitution::operator =( const FormalRTESubstitution & other ) {
	if ( this == & other )
		return * this;

	* this = FormalRTESubstitution ( other );

	return * this;
}

FormalRTESubstitution & FormalRTESubstitution::operator =( FormalRTESubstitution && other ) noexcept {
	std::swap ( this->left, other.left );
	std::swap ( this->right, other.right );
	std::swap ( this->substitutionSymbol, other.substitutionSymbol );

	this->left->attachRTE ( this->parentRTE );
	this->right->attachRTE ( this->parentRTE );

	return * this;
}

FormalRTESubstitution::~FormalRTESubstitution ( ) noexcept {
}

const FormalRTEElement & FormalRTESubstitution::getLeftElement ( ) const {
	return * left;
}

const FormalRTEElement & FormalRTESubstitution::getRightElement ( ) const {
	return * right;
}

const FormalRTESymbol & FormalRTESubstitution::getSubstitutionSymbol ( ) const {
	return * substitutionSymbol;
}

FormalRTEElement & FormalRTESubstitution::getLeftElement ( ) {
	return * left;
}

FormalRTEElement & FormalRTESubstitution::getRightElement ( ) {
	return * right;
}

FormalRTESymbol & FormalRTESubstitution::getSubstitutionSymbol ( ) {
	return * substitutionSymbol;
}

void FormalRTESubstitution::setLeftElement ( const FormalRTEElement & element ) {
	setLeftElement ( std::move_copy ( element ) );
}

void FormalRTESubstitution::setLeftElement ( FormalRTEElement && element ) {
	left = std::smart_ptr < FormalRTEElement > ( std::move ( element ).plunder ( ) );

	if ( this->parentRTE && !left->attachRTE ( this->parentRTE ) )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

void FormalRTESubstitution::setRightElement ( const FormalRTEElement & element ) {
	setRightElement ( std::move_copy ( element ) );
}

void FormalRTESubstitution::setRightElement ( FormalRTEElement && element ) {
	right = std::smart_ptr < FormalRTEElement > ( std::move ( element ).plunder ( ) );

	if ( this->parentRTE && !right->attachRTE ( this->parentRTE ) )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

void FormalRTESubstitution::setSubstitutionSymbol ( FormalRTESymbol symbol ) {
	substitutionSymbol = std::smart_ptr < FormalRTESymbol > ( std::move ( symbol ).plunder ( ) );

	if ( this->parentRTE && !substitutionSymbol->attachRTE ( this->parentRTE ) )
		throw exception::CommonException ( "Input symbols not in the alphabet." );
}

FormalRTESubstitution * FormalRTESubstitution::clone ( ) const {
	return new FormalRTESubstitution ( * this );
}

/*
 * UnboundedRTEElement* FormalRTESubstitution::cloneAsUnbounded() const {
 *  UnboundedRTESubstitution* res = new UnboundedRTESubstitution();
 *
 *  UnboundedRTEElement* leftTmp = left->cloneAsUnbounded();
 *  res->appendElement(std::move(*leftTmp));
 *  delete leftTmp;
 *
 *  UnboundedRTEElement* rightTmp = right->cloneAsUnbounded();
 *  res->appendElement(std::move(*rightTmp));
 *  delete rightTmp;
 *
 *  return res;
 * }
 */
FormalRTESubstitution * FormalRTESubstitution::plunder ( ) && {
	return new FormalRTESubstitution ( std::move ( * this ) );
}

int FormalRTESubstitution::compare ( const FormalRTESubstitution & other ) const {
	auto first = std::tie ( getLeftElement ( ), getRightElement ( ), getSubstitutionSymbol ( ) );
	auto second = std::tie ( other.getLeftElement ( ), other.getRightElement ( ), other.getSubstitutionSymbol ( ) );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void FormalRTESubstitution::operator >>( std::ostream & out ) const {
	out << "(FormalRTESubstitution";
	out << " " << * left;
	out << " " << * substitutionSymbol;
	out << " " << * right;
	out << ")";
}

bool FormalRTESubstitution::testSymbol ( const alphabet::RankedSymbol & symbol ) const {
	if ( left->testSymbol ( symbol ) ) return true;

	if ( right->testSymbol ( symbol ) ) return true;

	if ( substitutionSymbol->getSymbol ( ) == symbol ) return true;

	return false;
}

bool FormalRTESubstitution::attachRTE ( const FormalRTE * regexp ) {
	if ( this->parentRTE == regexp ) return true;

	this->parentRTE = regexp;

	if ( this->parentRTE == regexp ) return true;

	if ( !left->attachRTE ( regexp ) ) return false;

	if ( !right->attachRTE ( regexp ) ) return false;

	return this->parentRTE->getConstantAlphabet ( ).find ( this->substitutionSymbol->getSymbol ( ) ) != this->parentRTE->getConstantAlphabet ( ).end ( );
}

void FormalRTESubstitution::computeMinimalAlphabet ( std::set < alphabet::RankedSymbol > & alphabetF, std::set < alphabet::RankedSymbol > & alphabetK ) const {
	alphabetK.insert ( substitutionSymbol->getSymbol ( ) );
	left->computeMinimalAlphabet ( alphabetF, alphabetK );
	right->computeMinimalAlphabet ( alphabetF, alphabetK );
}

FormalRTESubstitution::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace rte */
