/*
 * RankedPattern.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "RankedPattern.h"
#include "../TreeException.h"

#include <iostream>
#include <algorithm>
#include <sstream>

#include "../unranked/UnrankedPattern.h"

#include <sax/FromXMLParserHelper.h>
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../Tree.h"
#include "../RankedTreeWrapper.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

namespace tree {

RankedPattern::RankedPattern ( alphabet::RankedSymbol subtreeWildcard, std::set < alphabet::RankedSymbol > alphabet, RankedNode pattern ) : std::Components < RankedPattern, alphabet::RankedSymbol, std::tuple < GeneralAlphabet >, std::tuple < SubtreeWildcard > > ( std::make_tuple ( std::move ( alphabet ) ), std::make_tuple ( std::move ( subtreeWildcard ) ) ), pattern ( nullptr ) {
	setTree ( std::move ( pattern ) );
}

RankedPattern::RankedPattern ( alphabet::RankedSymbol subtreeWildcard, RankedNode pattern ) : RankedPattern ( subtreeWildcard, pattern.computeMinimalAlphabet() + std::set < alphabet::RankedSymbol > { subtreeWildcard }, pattern ) {
}

RankedPattern::RankedPattern ( const UnrankedPattern & other ) : RankedPattern ( alphabet::RankedSymbol ( other.getSubtreeWildcard ( ), 0 ), other.getRoot ( ).asRanked ( ) ) {
}

RankedPattern::RankedPattern ( const RankedPattern & other ) : std::Components < RankedPattern, alphabet::RankedSymbol, std::tuple < GeneralAlphabet >, std::tuple < SubtreeWildcard > > ( std::make_tuple ( other.getAlphabet ( ) ), std::make_tuple ( other.getSubtreeWildcard ( ) ) ), pattern ( other.pattern ) {
	this->pattern->attachAlphabet ( & ( this->getAlphabet() ) );
}

RankedPattern::RankedPattern ( RankedPattern && other ) noexcept : std::Components < RankedPattern, alphabet::RankedSymbol, std::tuple < GeneralAlphabet >, std::tuple < SubtreeWildcard > > ( std::make_tuple ( std::move ( other.accessComponent < GeneralAlphabet > ( ).get ( ) ) ), std::make_tuple ( std::move ( other.accessElement < SubtreeWildcard > ( ).get ( ) ) ) ), pattern ( std::move ( other.pattern ) ) {
	this->pattern->attachAlphabet ( & ( this->getAlphabet() ) );
}

RankedTreeBase * RankedPattern::clone ( ) const {
	return new RankedPattern ( * this );
}

RankedTreeBase * RankedPattern::plunder ( ) && {
	return new RankedPattern ( std::move ( * this ) );
}

RankedPattern & RankedPattern::operator =( const RankedPattern & other ) {
	if ( this == & other )
		return * this;

	* this = RankedPattern ( other );

	return * this;
}

RankedPattern & RankedPattern::operator =( RankedPattern && other ) noexcept {
	std::swap ( this->pattern, other.pattern );
	std::swap ( accessComponent < GeneralAlphabet > ( ).get ( ), other.accessComponent < GeneralAlphabet > ( ).get ( ) );
	std::swap ( accessElement < SubtreeWildcard > ( ).get ( ), other.accessElement < SubtreeWildcard > ( ).get ( ) );

	return * this;
}

RankedPattern::~RankedPattern ( ) noexcept {
}

const RankedNode & RankedPattern::getRoot ( ) const {
	return * pattern;
}

RankedNode & RankedPattern::getRoot ( ) {
	return * pattern;
}

void RankedPattern::setTree ( RankedNode pattern ) {
	this->pattern = std::make_smart < RankedNode > ( std::move ( pattern ) );

	if ( !this->pattern->attachAlphabet ( & ( this->getAlphabet ( ) ) ) )
		throw TreeException ( "Input symbols not in the alphabet." );
}

void RankedPattern::operator >>( std::ostream & out ) const {
	out << "(RankedPattern " << * ( this->pattern ) << ")";
}

int RankedPattern::compare ( const RankedPattern & other ) const {
	auto first = std::tie ( * pattern, getAlphabet(), getSubtreeWildcard() );
	auto second = std::tie ( * other.pattern, other.getAlphabet(), other.getSubtreeWildcard() );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void RankedPattern::nicePrint ( std::ostream & os ) const {
	pattern->nicePrint ( os );
}

RankedPattern::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

RankedPattern RankedPattern::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, RankedPattern::getXmlTagName() );
	alphabet::RankedSymbol subtreeWildcardSymbol = TreeFromXMLParser::parseSubtreeWildcardRankedSymbol ( input );
	std::set < alphabet::RankedSymbol > rankedAlphabet = TreeFromXMLParser::parseRankedAlphabet ( input );
	RankedNode root = alib::xmlApi < RankedNode >::parse ( input );
	RankedPattern tree ( std::move ( subtreeWildcardSymbol ), std::move ( rankedAlphabet ), std::move ( root ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, RankedPattern::getXmlTagName() );
	return tree;
}

void RankedPattern::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( RankedPattern::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeSubtreeWildcard ( out, getSubtreeWildcard() );
	TreeToXMLComposer::composeAlphabet ( out, getAlphabet() );
	alib::xmlApi < RankedNode >::compose ( out, getRoot() );
	out.emplace_back ( RankedPattern::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace tree */

namespace std {

template < >
bool tree::RankedPattern::Component < tree::RankedPattern, alphabet::RankedSymbol, tree::GeneralAlphabet >::used ( const alphabet::RankedSymbol & symbol ) const {
	const tree::RankedPattern * pattern = static_cast < const tree::RankedPattern * > ( this );

	return pattern->getRoot ( ).testSymbol ( symbol ) || pattern->accessElement < tree::SubtreeWildcard > ( ).get ( ) == symbol;
}

template < >
bool tree::RankedPattern::Component < tree::RankedPattern, alphabet::RankedSymbol, tree::GeneralAlphabet >::available ( const alphabet::RankedSymbol & ) const {
	return true;
}

template < >
void tree::RankedPattern::Component < tree::RankedPattern, alphabet::RankedSymbol, tree::GeneralAlphabet >::valid ( const alphabet::RankedSymbol & ) const {
}

template < >
bool tree::RankedPattern::Element < tree::RankedPattern, alphabet::RankedSymbol, tree::SubtreeWildcard >::available ( const alphabet::RankedSymbol & symbol ) const {
	return static_cast < const tree::RankedPattern * > ( this )->accessComponent < tree::GeneralAlphabet > ( ).get ( ).count ( symbol );
}

template < >
void tree::RankedPattern::Element < tree::RankedPattern, alphabet::RankedSymbol, tree::SubtreeWildcard >::valid ( const alphabet::RankedSymbol & symbol ) const {
	if( symbol.getRank().getData() != 0 )
		throw tree::TreeException ( "SubtreeWildcard symbol has nonzero arity" );
}

} /* namespace std */

namespace alib {

auto rankedPatternParserRegister = xmlApi < tree::Tree >::ParserRegister < tree::RankedPattern > ();
auto rankedPatternParserRegister2 = xmlApi < tree::RankedTreeWrapper >::ParserRegister < tree::RankedPattern > ();
auto rankedPatternParserRegister3 = xmlApi < alib::Object >::ParserRegister < tree::RankedPattern > ();

auto RankedPatternFromUnrankedPattern = castApi::CastRegister < tree::RankedPattern, tree::UnrankedPattern > ( );
auto RankedPatternCastBinder = castApi::CastPoolStringBinder < tree::RankedPattern > ( tree::RankedPattern::getXmlTagName() );

} /* namespace alib */
