/*
 * UnrankedNode.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#include "UnrankedNode.h"
#include "UnrankedTree.h"
#include "../TreeException.h"
#include "../ranked/RankedNode.h"
#include <sstream>
#include "../../primitive/Unsigned.h"
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"

namespace tree {

UnrankedNode::UnrankedNode ( alphabet::Symbol symbol, std::vector < std::smart_ptr < UnrankedNode > > children ) : symbol ( std::move ( symbol ) ), children ( std::move ( children ) ), parent ( NULL ), alphabet ( NULL ) {
	for ( auto & element : this->children )
		element->parent = this;

	this->attachAlphabet ( NULL );
}

UnrankedNode::UnrankedNode ( const UnrankedNode & other ) : symbol ( other.symbol ), children ( other.children ), parent ( NULL ), alphabet ( NULL ) {
	for ( auto & element : this->children )
		element->parent = this;

	this->attachAlphabet ( NULL );
}

UnrankedNode::UnrankedNode ( UnrankedNode && other ) noexcept : symbol ( std::move ( other.symbol ) ), children ( std::move ( other.children ) ), parent ( NULL ), alphabet ( NULL ) {
	other.children.clear ( );

	for ( auto & element : this->children )
		element->parent = this;

	this->attachAlphabet ( NULL );
}

UnrankedNode & UnrankedNode::operator =( const UnrankedNode & other ) {
	if ( this == & other )
		return * this;

	* this = UnrankedNode ( other );

	return * this;
}

UnrankedNode & UnrankedNode::operator =( UnrankedNode && other ) noexcept {
	std::swap ( this->symbol, other.symbol );
	std::swap ( this->children, other.children );
	std::swap ( this->alphabet, other.alphabet ); // this->alphabet is stored within other.alphabet and it is reattached on the next line

	for ( auto & element : this->children )
		element->parent = this;

	this->attachAlphabet ( other.alphabet );

	return * this;
}

UnrankedNode::~UnrankedNode ( ) noexcept {
}

const std::vector < std::smart_ptr < const UnrankedNode > > & UnrankedNode::getChildren ( ) const {
	return reinterpret_cast < const std::vector < std::smart_ptr < const UnrankedNode > > & > ( children );
}

const std::vector < std::smart_ptr < UnrankedNode > > & UnrankedNode::getChildren ( ) {
	return children;
}

UnrankedNode * UnrankedNode::getParent ( ) {
	return parent;
}

const UnrankedNode * UnrankedNode::getParent ( ) const {
	return parent;
}

void UnrankedNode::setSymbol ( alphabet::Symbol symbol ) {
	this->symbol = std::move ( symbol );

	if ( ( this->alphabet != NULL ) && ( this->alphabet->count ( this->symbol ) == 0 ) ) throw TreeException ( "Symbol is not in the alphabet" );
}

void UnrankedNode::pushBackChild ( UnrankedNode node ) {
	node.attachAlphabet ( this->alphabet );
	this->children.emplace_back ( std::make_smart < UnrankedNode > ( node ) );
}

void UnrankedNode::swap ( UnrankedNode & other ) {
	const std::set < alphabet::Symbol > * thisParentTree = this->alphabet;
	const std::set < alphabet::Symbol > * otherParentTree = other.alphabet;

	UnrankedNode tmp = std::move ( other );

	other = std::move ( * this );
	* this = std::move ( tmp );

	this->attachAlphabet ( thisParentTree );
	other.attachAlphabet ( otherParentTree );
}

RankedNode UnrankedNode::asRanked ( ) const {
	std::vector < std::smart_ptr < RankedNode > > rankedChildren;

	for ( const std::smart_ptr < const UnrankedNode > & child : this->getChildren() )
		rankedChildren.emplace_back ( std::make_smart < RankedNode > ( child->asRanked ( ) ) );

	primitive::Unsigned rank( rankedChildren.size ( ) );
	return RankedNode ( alphabet::RankedSymbol ( this->symbol, std::move ( rank ) ), std::move ( rankedChildren ) );
}

int UnrankedNode::compare ( const UnrankedNode & other ) const {
	auto first = std::tie ( symbol, getChildren() );
	auto second = std::tie ( other.symbol, other.getChildren() );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void UnrankedNode::operator >>( std::ostream & out ) const {
	out << "(UnrankedNode " << " symbol = " << this->symbol << " children = " << this->children << "})";
}

std::ostream & operator <<( std::ostream & out, const UnrankedNode & node ) {
	node >> out;
	return out;
}

bool UnrankedNode::testSymbol ( const alphabet::Symbol & symbol ) const {
	if ( symbol == this->symbol ) return true;

	for ( const std::smart_ptr < const UnrankedNode > & child : this->getChildren() )
		if ( child->testSymbol ( symbol ) ) return true;

	return false;
}

bool UnrankedNode::attachAlphabet ( const std::set < alphabet::Symbol > * alphabet ) {
	if ( this->alphabet == alphabet ) return true;

	this->alphabet = alphabet;

	for ( std::smart_ptr < UnrankedNode > & child : this->children )
		if ( !child->attachAlphabet ( alphabet ) ) return false;

	if ( ( this->alphabet != NULL ) && ( this->alphabet->count ( this->symbol ) == 0 ) ) return false;

	return true;
}

std::set < alphabet::Symbol > UnrankedNode::computeMinimalAlphabet ( ) const {
	std::set < alphabet::Symbol > res;
	computeMinimalAlphabet ( res );
	return res;
}

void UnrankedNode::computeMinimalAlphabet ( std::set < alphabet::Symbol > & alphabet ) const {
	alphabet.insert ( this->symbol );

	for ( const std::smart_ptr < const UnrankedNode > & child : this->getChildren() )
		child->computeMinimalAlphabet ( alphabet );
}

const alphabet::Symbol & UnrankedNode::getSymbol ( ) const {
	return this->symbol;
}

UnrankedNode::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

void UnrankedNode::nicePrint ( std::ostream & os, const std::string & prefix, const bool last ) const {
	os << prefix;

	std::string nextPrefix ( prefix );

	if ( last ) {
		os << "\\-";
		nextPrefix += "  ";
	} else {
		os << "|-";
		nextPrefix += "| ";
	}

	os << ( std::string ) symbol << std::endl;

	for ( unsigned int i = 0; i < children.size ( ); ++i ) {
		os << nextPrefix << "|" << std::endl;
		children[i]->nicePrint ( os, nextPrefix, i == children.size ( ) - 1 );
	}
}

} /* namespace tree */

namespace alib {

tree::UnrankedNode xmlApi < tree::UnrankedNode >::parse ( std::deque < sax::Token >::iterator & input ) {
	return tree::TreeFromXMLParser::parseUnrankedNode ( input );
}

bool xmlApi < tree::UnrankedNode >::first ( const std::deque < sax::Token >::const_iterator & ) {
	throw tree::TreeException ( "Unimplemented" );
}

void xmlApi < tree::UnrankedNode >::compose ( std::deque < sax::Token > & output, const tree::UnrankedNode & data ) {
	tree::TreeToXMLComposer::composeNode ( output, data );
}

} /* namespace alib */
