/*
 * UnrankedNode.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#ifndef UNRANKED_NODE_H_
#define UNRANKED_NODE_H_

#include "../../alphabet/Symbol.h"
#include <vector>
#include <set>
#include <XmlApi.hpp>

namespace tree {

class RankedNode;
class UnrankedTree;
class UnrankedPattern;
class NonlinearUnrankedPattern;

/**
 * Represents a node in the ranked tree. Contains name of the symbol.
 */
class UnrankedNode {
protected:
	alphabet::Symbol symbol;

	std::vector < std::smart_ptr < UnrankedNode > > children;

	UnrankedNode * parent;

	/**
	 * Parent tree containing this instance of RegExpElement
	 */
	const std::set < alphabet::Symbol > * alphabet;

	/**
	 * @copydoc UnrankedNode::attachTree()
	 */
	bool attachAlphabet ( const std::set < alphabet::Symbol > * tree );

public:
	explicit UnrankedNode ( alphabet::Symbol symbol, std::vector < std::smart_ptr < UnrankedNode > > children );

	UnrankedNode ( const UnrankedNode & other );
	UnrankedNode ( UnrankedNode && other ) noexcept;
	UnrankedNode & operator =( const UnrankedNode & other );
	UnrankedNode & operator =( UnrankedNode && other ) noexcept;
	~UnrankedNode ( ) noexcept;

	/**
	 * @copydoc UnrankedNode::cloneAsUnranked() const
	 */
	RankedNode asRanked ( ) const;

	/**
	 * @copydoc UnrankedNode::testSymbol() const
	 */
	bool testSymbol ( const alphabet::Symbol & symbol ) const;

	/**
	 * @copydoc UnrankedNode::computeMinimalAlphabet()
	 */
	void computeMinimalAlphabet ( std::set < alphabet::Symbol > & alphabet ) const;

	/**
	 * @copydoc RankedNode::computeMinimalAlphabet()
	 */
	std::set < alphabet::Symbol > computeMinimalAlphabet ( ) const;

	/**
	 * @return children
	 */
	const std::vector < std::smart_ptr < const UnrankedNode > > & getChildren ( ) const;

	/**
	 * @return children
	 */
	const std::vector < std::smart_ptr < UnrankedNode > > & getChildren ( );

	UnrankedNode * getParent ( );

	const UnrankedNode * getParent ( ) const;

	void setSymbol ( alphabet::Symbol symbol );

	void pushBackChild ( UnrankedNode node );

	void swap ( UnrankedNode & other );

	int compare ( const UnrankedNode & ) const;

	/**
	 * @copydoc UnrankedNode::operator>>() const
	 */
	void operator >>( std::ostream & out ) const;

	friend std::ostream & operator <<( std::ostream &, const UnrankedNode & node );

	/**
	 * Returns the string representation of Node's Symbol.
	 */
	const alphabet::Symbol & getSymbol ( ) const;

	explicit operator std::string ( ) const;

	void nicePrint ( std::ostream &, const std::string & = "", const bool = true ) const;

	friend class UnrankedTree;
	friend class UnrankedPattern;
	friend class UnrankedNonlinearPattern;
};

} /* namespace tree */

namespace alib {

template < >
struct xmlApi < tree::UnrankedNode > {
	static tree::UnrankedNode parse ( std::deque < sax::Token >::iterator & input );
	static bool first ( const std::deque < sax::Token >::const_iterator & input );
	static void compose ( std::deque < sax::Token > & output, const tree::UnrankedNode & data );
};

} /* namespace alib */

namespace std {

template < >
struct compare < tree::UnrankedNode > {
	int operator ()( const tree::UnrankedNode & first, const tree::UnrankedNode & second ) const {
		return first.compare ( second );
	}

};

} /* namespace std */

#endif /* UNRANKED_NODE_H_ */
