/*
 * Author: Radovan Cerveny
 */

#include "MeasurementProcessorOutput.hpp"
#include "MeasurementProcessorViews.hpp"
#include "sax/SaxComposeInterface.h"
#include <set>
#include <iostream>
#include <exception/CommonException.h>

namespace measurements {

void MeasurementProcessorOutput::outputXml ( std::ostream & os, const MeasurementProvisionerResults & mpr, const MeasurementProcessorOutput & ) {
	os << mpr;
}

void MeasurementProcessorOutput::outputStats ( std::ostream & os, const MeasurementProvisionerResults & mpr, const MeasurementProcessorOutput & ) {

	std::vector < int > errorExitCodes;
	std::set < std::string > measuredPipelines;

	for ( const MPRInputResult & mprir : mpr.inputResults )
		for ( const MPRPipelineResult & mprpr : mprir.pipelineResults ) {
			if ( mprpr.pipelineStatus.exitCode != 0 )
				errorExitCodes.push_back ( mprpr.pipelineStatus.exitCode );

			std::string pipeline;

			for ( auto it = mprpr.commandResults.cbegin ( ); it != mprpr.commandResults.cend ( ); ++it ) {
				if ( it != mprpr.commandResults.cbegin ( ) )
					pipeline += " | ";

				pipeline += it->command;
			}

			measuredPipelines.insert ( pipeline );
		}

	os << "Total measurements: " << mpr.inputResults.size ( ) << std::endl;
	os << "Errors: ";

	if ( errorExitCodes.size ( ) == 0 )
		os << "none" << std::endl;
	else
		os << errorExitCodes << std::endl;

	os << "Measured pipelines:" << std::endl;

	for ( const std::string & pipeline : measuredPipelines )
		os << "  " << pipeline << std::endl;
}

void MeasurementProcessorOutput::outputCsv ( std::ostream & os, const MeasurementProvisionerResults & mpr, const MeasurementProcessorOutput & mpo ) {
	for ( const Table & table : MeasurementProcessorOutput::getTableRepresentation ( mpr, mpo ) ) {
		os << "# " << table.description << std::endl;

		for ( auto it = table.header.begin ( ); it != table.header.end ( ); ++it ) {
			if ( it != table.header.begin ( ) ) os << ",";

			os << * it;
		}

		os << std::endl;

		for ( const Row & row : table.rows ) {
			os << row.input;

			for ( const std::string & datum : row.data )
				os << "," << datum;

			os << std::endl;
		}
	}
}

void MeasurementProcessorOutput::outputHtml ( std::ostream & os, const MeasurementProvisionerResults & mpr, const MeasurementProcessorOutput & mpo ) {
	std::deque < sax::Token > tokens;

	tokens.emplace_back ( "html", sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( "head", sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( "style", sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( "table { border-collapse: collapse; background-color: #aaa } th, td { padding: 0.25rem; text-align: left; border: 1px solid #ccc; } tbody tr:nth-child(odd) { background: #ddd; }", sax::Token::TokenType::CHARACTER );
	tokens.emplace_back ( "style", sax::Token::TokenType::END_ELEMENT );
	tokens.emplace_back ( "title", sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( "Measurement results", sax::Token::TokenType::CHARACTER );
	tokens.emplace_back ( "title", sax::Token::TokenType::END_ELEMENT );
	tokens.emplace_back ( "head", sax::Token::TokenType::END_ELEMENT );
	tokens.emplace_back ( "body", sax::Token::TokenType::START_ELEMENT );

	for ( const Table & table : MeasurementProcessorOutput::getTableRepresentation ( mpr, mpo ) ) {

		tokens.emplace_back ( "div", sax::Token::TokenType::START_ELEMENT );
		tokens.emplace_back ( "h3", sax::Token::TokenType::START_ELEMENT );
		tokens.emplace_back ( table.description, sax::Token::TokenType::CHARACTER );
		tokens.emplace_back ( "h3", sax::Token::TokenType::END_ELEMENT );

		tokens.emplace_back ( "table", sax::Token::TokenType::START_ELEMENT );

		tokens.emplace_back ( "tr", sax::Token::TokenType::START_ELEMENT );

		for ( const std::string & head : table.header ) {
			tokens.emplace_back ( "th", sax::Token::TokenType::START_ELEMENT );
			tokens.emplace_back ( head, sax::Token::TokenType::CHARACTER );
			tokens.emplace_back ( "th", sax::Token::TokenType::END_ELEMENT );
		}

		tokens.emplace_back ( "tr", sax::Token::TokenType::END_ELEMENT );

		for ( const Row & row : table.rows ) {
			tokens.emplace_back ( "tr", sax::Token::TokenType::START_ELEMENT );

			tokens.emplace_back ( "td", sax::Token::TokenType::START_ELEMENT );
			tokens.emplace_back ( row.input, sax::Token::TokenType::CHARACTER );
			tokens.emplace_back ( "td", sax::Token::TokenType::END_ELEMENT );

			for ( const std::string & datum : row.data ) {
				tokens.emplace_back ( "td", sax::Token::TokenType::START_ELEMENT );
				tokens.emplace_back ( datum, sax::Token::TokenType::CHARACTER );
				tokens.emplace_back ( "td", sax::Token::TokenType::END_ELEMENT );
			}

			tokens.emplace_back ( "tr", sax::Token::TokenType::END_ELEMENT );
		}

		tokens.emplace_back ( "table", sax::Token::TokenType::END_ELEMENT );
		tokens.emplace_back ( "div", sax::Token::TokenType::END_ELEMENT );
	}

	tokens.emplace_back ( "body", sax::Token::TokenType::END_ELEMENT );
	tokens.emplace_back ( "html", sax::Token::TokenType::END_ELEMENT );

	sax::SaxComposeInterface::printStream ( os, tokens );
}

std::vector < MeasurementProcessorOutput::Table > MeasurementProcessorOutput::getTableRepresentation ( const MeasurementProvisionerResults & mpr, const MeasurementProcessorOutput & mpo ) {
	 // this starts to get ugly
	std::vector < std::string > masterHeader {
		"input"
	};

	for ( const MPRPipelineResult & mprpr : mpr.inputResults[0].pipelineResults )
		for ( const MPRPipelineCommandResult & mprpcr : mprpr.commandResults )
			masterHeader.push_back ( mprpcr.command );

	 // this is definitely ugly
	const auto & masterFrames = mpr.inputResults[0].pipelineResults[0].commandResults[0].measurementResults.frames;

	std::vector < Table > tables ( masterFrames.size ( ) - 1 );

	 // start from 1 = skip ROOT
	for ( size_t frameIdx = 1; frameIdx < masterFrames.size ( ); ++frameIdx ) {
		Table & table = tables[frameIdx - 1];

		table.description = to_string ( masterFrames[frameIdx].name ) + " " + to_string ( masterFrames[frameIdx].type ) + " :";

		for ( const std::string & attr : mpo.outputEngineAttr )
			table.description += " " + attr;

		table.header = masterHeader;

		for ( const MPRInputResult & mprir : mpr.inputResults ) {
			Row row;

			for ( const auto & input : mprir.inputs )
				row.input += input.second + " ";

			for ( const MPRPipelineResult & mprpr : mprir.pipelineResults )
				for ( const MPRPipelineCommandResult & mprpcr : mprpr.commandResults )
					row.data.push_back ( MeasurementProcessorOutput::getData ( mprpcr.measurementResults.frames[frameIdx], mpo ) );

			table.rows.push_back ( std::move ( row ) );
		}
	}

	return tables;
}

std::string MeasurementProcessorOutput::getData ( const MeasurementFrame & frame, const MeasurementProcessorOutput & mpo ) {
	switch ( mpo.outputEngine ) {
	case OutputEngine::TIME:
		return MeasurementProcessorViews::viewTimeDataFrame ( frame, mpo.outputEngineAttr );

	case OutputEngine::MEMORY:
		return MeasurementProcessorViews::viewMemoryDataFrame ( frame, mpo.outputEngineAttr );

	case OutputEngine::COUNTER:
		return MeasurementProcessorViews::viewCounterDataFrame ( frame, mpo.outputEngineAttr );
	}

	throw::exception::CommonException ( "Missing case." );
}

}
