/*
 * AutomatonToStringComposer.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "AutomatonToStringComposer.h"
#include "exception/CommonException.h"
#include "automaton/FSM/DFA.h"
#include "automaton/FSM/NFA.h"
#include "automaton/FSM/MultiInitialStateNFA.h"
#include "automaton/FSM/EpsilonNFA.h"

#include "../StringApi.hpp"

namespace automaton {

void AutomatonToStringComposer::composeTransitionsFromState(std::ostream& out, const DFA& automaton, const State& from) {
	std::map<std::pair<State, alphabet::Symbol>, State> symbolTransitionsFromState = automaton.getTransitionsFromState(from);

	for(const alphabet::Symbol& inputSymbol : automaton.getInputAlphabet()) {
		const std::map<std::pair<State, alphabet::Symbol>, State>::iterator toStates = symbolTransitionsFromState.find(std::make_pair(from, inputSymbol));
		if(toStates == symbolTransitionsFromState.end()) {
			out << " -";
		} else {
			out << " ";
			alib::stringApi<label::Label>::compose(out, toStates->second.getName());
		}
	}
}

void AutomatonToStringComposer::composeTransitionsFromState(std::ostream& out, const NFA& automaton, const State& from) {
	std::map<std::pair<State, alphabet::Symbol>, std::set<State> > symbolTransitionsFromState = automaton.getTransitionsFromState(from);

	for(const alphabet::Symbol& inputSymbol : automaton.getInputAlphabet()) {
		const std::map<std::pair<State, alphabet::Symbol>, std::set<State> >::iterator toStates = symbolTransitionsFromState.find(std::make_pair(from, inputSymbol));
		if(toStates == symbolTransitionsFromState.end() || toStates->second.size() == 0) {
			out << " -";
		} else {
			bool sign = false;
			for(const State& to : toStates->second) {
				out << (sign ? "|" : " ");
				alib::stringApi<label::Label>::compose(out, to.getName());
				sign = true;
			}
		}
	}
}

void AutomatonToStringComposer::composeTransitionsFromState(std::ostream& out, const MultiInitialStateNFA& automaton, const State& from) {
	std::map<std::pair<State, alphabet::Symbol>, std::set<State> > symbolTransitionsFromState = automaton.getTransitionsFromState(from);

	for(const alphabet::Symbol& inputSymbol : automaton.getInputAlphabet()) {
		const std::map<std::pair<State, alphabet::Symbol>, std::set<State> >::iterator toStates = symbolTransitionsFromState.find(std::make_pair(from, inputSymbol));
		if(toStates == symbolTransitionsFromState.end() || toStates->second.size() == 0) {
			out << " -";
		} else {
			bool sign = false;
			for(const State& to : toStates->second) {
				out << (sign ? "|" : " ");
				alib::stringApi<label::Label>::compose(out, to.getName());
				sign = true;
			}
		}
	}
}

void AutomatonToStringComposer::composeTransitionsFromState(std::ostream& out, const EpsilonNFA& automaton, const State& from) {
	std::map<std::pair<State, alphabet::Symbol>, std::set<State> > symbolTransitionsFromState = automaton.getSymbolTransitionsFromState(from);

	for(const alphabet::Symbol& inputSymbol : automaton.getInputAlphabet()) {
		const std::map<std::pair<State, alphabet::Symbol>, std::set<State> >::iterator toStates = symbolTransitionsFromState.find(std::make_pair(from, inputSymbol));
		if(toStates == symbolTransitionsFromState.end() || toStates->second.size() == 0) {
			out << " -";
		} else {
			bool sign = false;
			for(const State& to : toStates->second) {
				out << (sign ? "|" : " ");
				alib::stringApi<label::Label>::compose(out, to.getName());
				sign = true;
			}
		}
	}

	std::map<State, std::set<State> > epsilonTransitionsFromState = automaton.getEpsilonTransitionsFromState(from);
	if(epsilonTransitionsFromState.size() == 0 || epsilonTransitionsFromState.begin()->second.size() == 0) {
		out << " -";
	} else {
		bool sign = false;
		for(const State& to : epsilonTransitionsFromState.begin()->second) {
			out << (sign ? "|" : " ");
			alib::stringApi<label::Label>::compose(out, to.getName());
			sign = true;
		}
	}
}

void AutomatonToStringComposer::compose(std::ostream& out, const DFA& automaton) {
	out << "DFA";
	for(const auto& symbol : automaton.getInputAlphabet()) {
		out << " ";
		alib::stringApi<alphabet::Symbol>::compose(out, symbol);
	}

	out << std::endl;

	for(const auto& state : automaton.getStates()) {
		if(automaton.getInitialState() == state) {
			out << ">";
		}
		if(automaton.getFinalStates().find(state) != automaton.getFinalStates().end()) {
			out << "<";
		}
		alib::stringApi<label::Label>::compose(out, state.getName());

		composeTransitionsFromState(out, automaton, state);

		out << std::endl;
	}
}

AutomatonToStringComposer::RegistratorWrapper<void, DFA> AutomatonToStringComposerDFA = AutomatonToStringComposer::RegistratorWrapper<void, DFA>(AutomatonToStringComposer::compose);

void AutomatonToStringComposer::compose(std::ostream& out, const NFA& automaton) {
	out << "NFA";
	for(const auto& symbol : automaton.getInputAlphabet()) {
		out << " ";
		alib::stringApi<alphabet::Symbol>::compose(out, symbol);
	}

	out << std::endl;

	for(const auto& state : automaton.getStates()) {
		if(automaton.getInitialState() == state) {
			out << ">";
		}
		if(automaton.getFinalStates().find(state) != automaton.getFinalStates().end()) {
			out << "<";
		}
		alib::stringApi<label::Label>::compose(out, state.getName());

		composeTransitionsFromState(out, automaton, state);

		out << std::endl;
	}
}

AutomatonToStringComposer::RegistratorWrapper<void, NFA> AutomatonToStringComposerNFA = AutomatonToStringComposer::RegistratorWrapper<void, NFA>(AutomatonToStringComposer::compose);

void AutomatonToStringComposer::compose(std::ostream& out, const MultiInitialStateNFA& automaton) {
	out << "MISNFA";
	for(const auto& symbol : automaton.getInputAlphabet()) {
		out << " ";
		alib::stringApi<alphabet::Symbol>::compose(out, symbol);
	}

	out << std::endl;

	for(const auto& state : automaton.getStates()) {
		if(automaton.getInitialStates().find(state) != automaton.getInitialStates().end()) {
			out << ">";
		}
		if(automaton.getFinalStates().find(state) != automaton.getFinalStates().end()) {
			out << "<";
		}
		alib::stringApi<label::Label>::compose(out, state.getName());

		composeTransitionsFromState(out, automaton, state);

		out << std::endl;
	}
}

AutomatonToStringComposer::RegistratorWrapper<void, MultiInitialStateNFA> AutomatonToStringComposerMultiInitialStateNFA = AutomatonToStringComposer::RegistratorWrapper<void, MultiInitialStateNFA>(AutomatonToStringComposer::compose);

void AutomatonToStringComposer::compose(std::ostream& out, const EpsilonNFA& automaton) {
	out << "ENFA";
	for(const auto& symbol : automaton.getInputAlphabet()) {
		out << " ";
		alib::stringApi<alphabet::Symbol>::compose(out, symbol);
	}

	out << " #E";
	out << std::endl;

	for(const auto& state : automaton.getStates()) {
		if(automaton.getInitialState() == state) {
			out << ">";
		}
		if(automaton.getFinalStates().find(state) != automaton.getFinalStates().end()) {
			out << "<";
		}
		alib::stringApi<label::Label>::compose(out, state.getName());

		composeTransitionsFromState(out, automaton, state);

		out << std::endl;
	}
}

AutomatonToStringComposer::RegistratorWrapper<void, EpsilonNFA> AutomatonToStringComposerEpsilonNFA = AutomatonToStringComposer::RegistratorWrapper<void, EpsilonNFA>(AutomatonToStringComposer::compose);

void AutomatonToStringComposer::compose(std::ostream& output, const Automaton& automaton) {
	dispatch(output, automaton.getData());
}

} /* namespace automaton */
