/*
 * TreeFromStringParser.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "TreeFromStringParser.h"
#include "exception/CommonException.h"
#include "tree/TreeClasses.h"

#include "../StringApi.hpp"
#include "alphabet/SubtreeWildcardSymbol.h"
#include "alphabet/NonlinearVariableSymbol.h"

namespace tree {

Tree TreeFromStringParser::parseTree ( std::istream & input ) const {
	return parseTree ( input, std::set < FEATURES > ( { FEATURES::RANKED_TREE, FEATURES::RANKED_PATTERN, FEATURES::RANKED_NONLINEAR_PATTERN, FEATURES::UNRANKED_TREE, FEATURES::UNRANKED_PATTERN, FEATURES::UNRANKED_NONLINEAR_PATTERN } ) );
}

Tree TreeFromStringParser::parseTree ( std::istream & input, const std::set < FEATURES > & features ) const {
	TreeFromStringLexer::Token token = m_TreeLexer.next ( input );

	if ( token.type == TreeFromStringLexer::TokenType::SUBTREE_WILDCARD ) {
		token = m_TreeLexer.next ( input );

		if ( token.type == TreeFromStringLexer::TokenType::BAR ) {
			alphabet::Symbol subtreeWildcard ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD );
			return Tree ( UnrankedPattern ( subtreeWildcard, UnrankedNode ( subtreeWildcard, { } ) ) );
		} else {
			alphabet::RankedSymbol subtreeWildcard ( alphabet::Symbol ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD ), 0 );
			return Tree ( RankedPattern ( subtreeWildcard, RankedNode ( subtreeWildcard, { } ) ) );
		}
	} else if ( token.type == TreeFromStringLexer::TokenType::NONLINEAR_VARIABLE ) {
		token = m_TreeLexer.next ( input );

		if ( token.type == TreeFromStringLexer::TokenType::BAR ) {
			alphabet::Symbol subtreeWildcard ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD );
			alphabet::Symbol nonlinearVariable ( alphabet::NonlinearVariableSymbol ( alphabet::symbolFrom ( token.value ) ) );
			return Tree ( UnrankedNonlinearPattern ( subtreeWildcard, { nonlinearVariable }, UnrankedNode ( subtreeWildcard, { } ) ) );
		} else {
			alphabet::RankedSymbol subtreeWildcard ( alphabet::Symbol ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD ), 0 );
			alphabet::RankedSymbol nonlinearVariable ( alphabet::Symbol ( alphabet::NonlinearVariableSymbol ( alphabet::symbolFrom ( token.value ) ) ), 0 );
			return Tree ( RankedNonlinearPattern ( subtreeWildcard, { nonlinearVariable }, RankedNode ( nonlinearVariable, { } ) ) );
		}
	} else {
		m_TreeLexer.putback ( input, token );
		alphabet::Symbol symbol = alib::stringApi < alphabet::Symbol >::parse ( input );

		unsigned rank = 0;

		token = m_TreeLexer.next ( input );

		if ( token.type == TreeFromStringLexer::TokenType::RANK ) {
			std::set < alphabet::RankedSymbol > nonlinearVariables;

			rank = std::stou ( token.value );

			std::vector < std::smart_ptr < tree::RankedNode > > childs;
			bool isPattern = false;

			for ( unsigned i = 0; i < rank; i++ )
				childs.emplace_back ( std::make_smart < RankedNode > ( parseRankedContent ( input, isPattern, nonlinearVariables ) ) );

			if ( isPattern && nonlinearVariables.size ( ) ) {
				alphabet::RankedSymbol subtreeWildcard ( alphabet::Symbol ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD ), 0 );
				RankedNonlinearPattern tree ( std::move ( subtreeWildcard ), std::move ( nonlinearVariables ), RankedNode ( alphabet::RankedSymbol ( std::move ( symbol ), rank ), std::move ( childs ) ) );

				if ( features.count ( FEATURES::RANKED_NONLINEAR_PATTERN ) ) return Tree ( std::move ( tree ) );
			} else if ( isPattern ) {
				alphabet::RankedSymbol subtreeWildcard ( alphabet::Symbol ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD ), 0 );
				RankedPattern tree ( std::move ( subtreeWildcard ), RankedNode ( alphabet::RankedSymbol ( std::move ( symbol ), rank ), std::move ( childs ) ) );

				if ( features.count ( FEATURES::RANKED_PATTERN ) ) return Tree ( std::move ( tree ) );
			} else {
				RankedTree tree ( RankedNode ( alphabet::RankedSymbol ( std::move ( symbol ), rank ), std::move ( childs ) ) );

				if ( features.count ( FEATURES::RANKED_TREE ) ) return Tree ( std::move ( tree ) );
			}
		} else {
			std::set < alphabet::Symbol > nonlinearVariables;

			std::vector < std::smart_ptr < tree::UnrankedNode > > childs;
			bool isPattern = false;

			while ( token.type != TreeFromStringLexer::TokenType::BAR ) {
				m_TreeLexer.putback ( input, token );
				childs.emplace_back ( std::make_smart < UnrankedNode > ( parseUnrankedContent ( input, isPattern, nonlinearVariables ) ) );
				token = m_TreeLexer.next ( input );
			}

			if ( token.type != TreeFromStringLexer::TokenType::BAR )
				throw exception::CommonException ( "Missing bar" );

			if ( isPattern && nonlinearVariables.size ( ) ) {
				alphabet::Symbol subtreeWildcard ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD );
				UnrankedNonlinearPattern tree ( std::move ( subtreeWildcard ), std::move ( nonlinearVariables ), UnrankedNode ( std::move ( symbol ), std::move ( childs ) ) );

				if ( features.count ( FEATURES::UNRANKED_NONLINEAR_PATTERN ) ) return Tree ( std::move ( tree ) );
			} else if ( isPattern ) {
				alphabet::Symbol subtreeWildcard ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD );
				UnrankedPattern tree ( std::move ( subtreeWildcard ), UnrankedNode ( std::move ( symbol ), std::move ( childs ) ) );

				if ( features.count ( FEATURES::UNRANKED_PATTERN ) ) return Tree ( std::move ( tree ) );
			} else {
				UnrankedTree tree ( UnrankedNode ( std::move ( symbol ), std::move ( childs ) ) );

				if ( features.count ( FEATURES::UNRANKED_TREE ) ) return Tree ( std::move ( tree ) );
			}
		}

		throw exception::CommonException ( "Invalid input" );
	}
}

tree::RankedNode TreeFromStringParser::parseRankedContent ( std::istream & input, bool & isPattern, std::set < alphabet::RankedSymbol > & nonlinearVariables ) const {
	TreeFromStringLexer::Token token = m_TreeLexer.next ( input );

	if ( token.type == TreeFromStringLexer::TokenType::SUBTREE_WILDCARD ) {
		isPattern = true;
		alphabet::RankedSymbol subtreeWildcard ( alphabet::Symbol ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD ), 0 );
		return RankedNode ( std::move ( subtreeWildcard ), { } );
	} else if ( token.type == TreeFromStringLexer::TokenType::NONLINEAR_VARIABLE ) {
		isPattern = true;
		alphabet::RankedSymbol nonlinearVariable ( alphabet::Symbol ( alphabet::NonlinearVariableSymbol ( alphabet::symbolFrom ( token.value ) ) ), 0 );
		nonlinearVariables.insert ( nonlinearVariable );
		return RankedNode ( std::move ( nonlinearVariable ), { } );
	} else {
		m_TreeLexer.putback ( input, token );
		alphabet::Symbol symbol = alib::stringApi < alphabet::Symbol >::parse ( input );

		unsigned rank = 0;
		std::vector < std::smart_ptr < tree::RankedNode > > childs;

		TreeFromStringLexer::Token token = m_TreeLexer.next ( input );

		if ( token.type == TreeFromStringLexer::TokenType::RANK )
			rank = std::stou ( token.value );
		else
			throw exception::CommonException ( "Missing rank" );

		for ( unsigned i = 0; i < rank; i++ )
			childs.emplace_back ( std::make_smart < RankedNode > ( parseRankedContent ( input, isPattern, nonlinearVariables ) ) );

		return RankedNode ( alphabet::RankedSymbol ( std::move ( symbol ), rank ), std::move ( childs ) );
	}
}

tree::UnrankedNode TreeFromStringParser::parseUnrankedContent ( std::istream & input, bool & isPattern, std::set < alphabet::Symbol > & nonlinearVariables ) const {
	TreeFromStringLexer::Token token = m_TreeLexer.next ( input );

	if ( token.type == TreeFromStringLexer::TokenType::SUBTREE_WILDCARD ) {
		token = m_TreeLexer.next ( input );

		if ( token.type != TreeFromStringLexer::TokenType::BAR )
			throw exception::CommonException ( "Missing bar" );

		isPattern = true;
		alphabet::Symbol subtreeWildcard ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD );
		return UnrankedNode ( std::move ( subtreeWildcard ), { } );
	} else if ( token.type == TreeFromStringLexer::TokenType::NONLINEAR_VARIABLE ) {
		token = m_TreeLexer.next ( input );

		if ( token.type != TreeFromStringLexer::TokenType::BAR )
			throw exception::CommonException ( "Missing bar" );

		isPattern = true;
		alphabet::Symbol nonlinearVariable ( alphabet::NonlinearVariableSymbol ( alphabet::symbolFrom ( token.value ) ) );
		nonlinearVariables.insert ( nonlinearVariable );
		return UnrankedNode ( std::move ( nonlinearVariable ), { } );
	} else {
		m_TreeLexer.putback ( input, token );
		alphabet::Symbol symbol = alib::stringApi < alphabet::Symbol >::parse ( input );

		std::vector < std::smart_ptr < tree::UnrankedNode > > childs;

		TreeFromStringLexer::Token token = m_TreeLexer.next ( input );

		while ( token.type != TreeFromStringLexer::TokenType::BAR ) {
			m_TreeLexer.putback ( input, token );
			childs.emplace_back ( std::make_smart < UnrankedNode > ( parseUnrankedContent ( input, isPattern, nonlinearVariables ) ) );
			token = m_TreeLexer.next ( input );
		}

		if ( token.type != TreeFromStringLexer::TokenType::BAR )
			throw exception::CommonException ( "Missing bar" );

		return UnrankedNode ( std::move ( symbol ), std::move ( childs ) );
	}
}

} /* namespace tree */
